#include "win.h"

#include <QtGui>
#include "math/primitives/polygon.h"
#include "ui/chainplug.h"
#include "ui/exportdialog.h"
#include "ui/renderview.h"
#include "ui/settingsplug.h"
#include "ui/sizedialog.h"
#include "ui/stripview.h"
#include "ui/ui_defs.h"
#include "model/activeproject.h"

#include "ui/valuepopup.h"

/* WIN
 ******************************************************************/
Win::Win(const Environment& env, const ArcList& arcs,
		 const ColorSrcList& clrs, ActiveProject& prj,
		 QWidget *parent)
	: QWidget(parent)
	, mEnv(env)
	, mArcs(arcs)
	, mClrs(clrs)
	, mPrj(prj)
	, mLayer(0)
	, mStrip0(NULL)
	, mStrip1(NULL)
	, mSettingsStrip(NULL)
{
	// XXX I don't think I can rely on font info in the constructor.
	// It's usually fine, but occasionally reports garbage values.
	// I'm thinking the guarantee is after QWidget::ensurePolish()
	// has been called, which I don't believe can be done in the constructor.
	// See that method description for how to handle code after it's
	// called, although for now I'm going through the app, which I'm
	// hoping works fine.
//	const QFontMetrics		fm = fontMetrics();
	const QFontMetrics		fm = qApp->fontMetrics();
	mLook.mFh = fm.height();
	mLook.mFa = fm.ascent();
	mLook.mFd = fm.descent();

	// Layout:
	// FX strip
	// parameter strip
	// workspace
	// command strip

	mStrip0 = new StripView(prj, arcs, mLook);
	mStrip1 = new StripView(prj, arcs, mLook);
	mRenderView = new RenderView(prj, mLook);
	mSettingsStrip = new StripView(prj, arcs, mLook);

	QObject::connect(&prj, SIGNAL(projectChanged(const int)),
					 mStrip0, SLOT(onProjectChanged(const int)));
	QObject::connect(&prj, SIGNAL(projectChanged(const int)),
					 mStrip1, SLOT(onProjectChanged(const int)));
	QObject::connect(&prj, SIGNAL(projectChanged(const int)),
					 mRenderView, SLOT(onProjectChanged(const int)));
	QObject::connect(&prj, SIGNAL(paramBatchFinished()),
					 mRenderView, SLOT(onParamBatchFinished()));

	qRegisterMetaType<Param>("Param");
	QObject::connect(&prj, SIGNAL(paramChanged(const Param&)),
					 mStrip1, SLOT(onParamChanged(const Param&)));
	QObject::connect(&prj, SIGNAL(paramChanged(const Param&)),
					 mRenderView, SLOT(onParamChanged(const Param&)));

	QGridLayout *mainLayout = new QGridLayout;
	if (mStrip0 != NULL) {
		mStrip0->setPlug(new ChainPlug(prj, mNavigation));
		mainLayout->addWidget(mStrip0, 0, 0);
	}
	if (mStrip1 != NULL) {
		mainLayout->addWidget(mStrip1, 1, 0);
	}
	mainLayout->addWidget(mRenderView, 2, 0);

	if (mSettingsStrip != NULL) {
		mSettingsStrip->setPlug(new SettingsPlug(prj, mNavigation, mSettingsStrip));
		mainLayout->addWidget(mSettingsStrip, 3, 0);
	}

// testing a popup control
// mainLayout->addWidget(new ValuePopupButton(mLook, this), 4, 0);

	setLayout(mainLayout);
	setWindowTitle(tr("arclight"));

	QObject::connect(&mNavigation, SIGNAL(onSelectFx(const int)),
					 this, SLOT(onSelectFx(const int)));
	QObject::connect(&mNavigation, SIGNAL(onCommand(const QString)),
					 this, SLOT(onCommand(const QString)));
	QObject::connect(&mNavigation, SIGNAL(onLoadProject()),
					 this, SLOT(onRequestLoadProject()));
	QObject::connect(&mNavigation, SIGNAL(onSaveProject()),
					 this, SLOT(onRequestSaveProject()));

	QObject::connect(&mNavigation, SIGNAL(onAddFx(const QPoint)),
					 this, SLOT(onAddFx(const QPoint)));
	QObject::connect(&mNavigation, SIGNAL(onRemoveFx(const int)),
					 this, SLOT(onRemoveFx(const int)));
}

void Win::onRemoveFx(const int fxId)
{
	fx::EditOp	edit;
	edit.setToRemove(fxId);
	mPrj.edit(edit);

	if (fxId == mNavigation.mSelectedFx) mNavigation.selectFx(-1);
}

void Win::onCommand(const QString cmd)
{
	if (cmd == EXPORT_CMD) {
		ExportDialog	d(mPrj, this);
		if (d.exec() == QDialog::Accepted) {
			if (mRenderView != NULL) mRenderView->writeExports(mPrj);
		}
	} else if (cmd == SIZE_CMD) {
		QSize			size(mPrj.getCanvasSize());
		SizeDialog		d(size, this);
		if (d.exec() == QDialog::Accepted && d.getSize(size)) {
			mPrj.setCanvasSize(size);
		}
	}
}

void Win::onRequestLoadProject()
{
	const QString	fn = QFileDialog::getOpenFileName(this, tr("Open Project"), tr("/home"), tr("Project Files (*.json)"));
	if (fn.length() > 0) {
		QFile		f(fn);
		if (f.open(QFile::ReadOnly)) {
			mPrj.readJson(&f);
		}
	}
}

void Win::onRequestSaveProject()
{
	const QString	fn = QFileDialog::getSaveFileName(this, tr("Save Project"), tr("/home"), tr("Project Files (*.json)"));
	if (fn.length() > 0) {
		QFile		f(fn);
		if (f.open(QFile::WriteOnly | QFile::Truncate)) {
			json::PrettyFormat	fmt;
			json::Writer		w(&f, fmt);
			mPrj.writeJson(w);
		}
	}
}

/* NEW-FX-UI.  Given an FX id and
 * a Create, create a new UI.
 ******************************************************************/
class Win::NewFxUi : public fx::Visitor {
public:
	const int		mFxId;
	FxUi::Create	mCreate;
	FxUi*			mUi;

	NewFxUi(const int fxId, const Environment& env, ActiveProject& prj,
			const ArcList& arcs, const ColorSrcList& clrs, const int mode)
		: mFxId(fxId)
		, mCreate(env, prj, arcs, clrs, mode)
		, mUi(NULL)
	{
	}

	virtual int		visit(Fx& fx, fx::VisitParams&)
	{
		if (fx.getId() != mFxId) return fx::CONTINUE;
		mUi = fx.newUi(mCreate);
		return fx::STOP;
	}
};

void Win::onSelectFx(const int fxId)
{
	NewFxUi		p(fxId, mEnv, mPrj, mArcs, mClrs, FxUi::EDIT);
	mPrj.forEach(p);

	if (mStrip0 != NULL) mStrip0->update();
	if (mStrip1 != NULL) mStrip1->setPlug(p.mUi);
	if (mRenderView != NULL) mRenderView->onSelectFx(mNavigation.mSelectedFx);
}

/* WIN::FX-MENU
 ******************************************************************/
class Win::FxMenu : public fx::FactoryVisitor {
public:
	QMenu&			mMenu;

	FxMenu(QMenu& m) : mMenu(m)
	{
	}

	virtual int		visit(const FxFactory& f)
	{
		const QString	label = f.label(), key = f.key();
		if (label.length() > 0 && key.length() > 0) {
			QAction*	a = mMenu.addAction(label);
			if (a != NULL) a->setData(QVariant(key));
		}
		return fx::CONTINUE;
	}
};

void Win::onAddFx(const QPoint pos)
{
	QMenu				m(this);
	FxMenu				visitor(m);
	FxFactory::forEach(visitor);
	QAction*			a = m.exec(pos);
	if (a != NULL) {
		const QString	key = a->data().toString();
		Fx*				fx = FxFactory::create(key);
		if (fx != NULL) {
			fx->init(mPrj);
			fx::EditOp	edit;
			edit.setToAdd(mPrj.getFxId(mLayer), fx);
			mPrj.edit(edit);

			if (edit.success()) mNavigation.selectFx(fx->getId());
		}
	}
}

