#include "arcrender.h"

#include <QPixmap>
#include <QPainter>

#include "engine/arc.h"
#include "engine/archead.h"
#include "math/geo.h"
#include "math/primitives/point3.h"
#include "math/primitives/range.h"

static inline void makeColor(	int r, int g, int b,
								const uchar srcA, const float scale,
								QColor&);

/* ARC-RENDER
 ******************************************************************/
ArcRender::ArcRender()
{
}

QPixmap* ArcRender::onRender(ArcHead& ah, const QSize& size)
{
	QPixmap*		ans = new QPixmap(size);
	if (ans == NULL) return NULL;

	{
		// Clear the background.
		QPainter	painter(ans);
		QPainter::CompositionMode	mode = painter.compositionMode();
		painter.setCompositionMode(QPainter::CompositionMode_Source);
		painter.fillRect(0, 0, size.width(), size.height(), QColor(0, 0, 0, 0));
		painter.setCompositionMode(mode);
	}
	if (!onRender(ah, *ans)) {
		delete ans;
		return NULL;
	}
	return ans;
}

bool ArcRender::onRender(ArcHead& ah, QPixmap& pm)
{
	Arc*					arc = ah.getArc();
	if (arc == NULL) return false;
	const QSize				qs = pm.size();
	const int				w = qs.width(), h = qs.height();
	if (!mArcP.begin(&ah, w*h, arc::HAS_DISTANCE_SRC | arc::HAS_NORMAL_SRC)) return false;
	if (!makeSource(qs)) return false;
	float*					rowV = mRowV.get(mArcP.mRowSize);
	float*					rowMask = mRowMask.get(mArcP.mRowSize);
	if (rowV == NULL || rowMask == NULL) return false;

	set_to_1(rowV, mArcP.mRowSize);
	arc->work(mArcP, rowV);

	QPainter				painter(&pm);
	QColor					clr;
	int						pix = 0;
	const int				r = 180, g = 180, b = 180;
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			int				a = rowMask[pix] * 255;
			if (a < 0) a = 0; else if (a > 255) a = 255;
			if (a > 0) {
				float		v = rowV[pix];
				if (v < -1) v = -1;
				else if (v > 1) v = 1;
				makeColor(r, g, b, a, v, clr);

				painter.setPen(clr);
				painter.drawPoint(x, y);
			}
			pix++;
		}
	}
	return true;
}

bool ArcRender::makeSource(const QSize& s)
{
	if (mSize == s) return true;
	const int		size = s.width() * s.height();
	if (mArcP.mRowSize != size) return false;

	float*				rowMask = mRowMask.get(size);
	if (rowMask == NULL) return false;

	set_to_0(mArcP.mDistanceRow, size);
	set_to_0(mArcP.mNormalRow, size);
	set_to_0(rowMask, size);

	const int				w = s.width(), h = s.height();
	const Point3			cen((w-1) * 0.5f, (h-1) * 0.5f, 0);
	Point3					cur;
	const float				maxD = (cen.x > cen.y ? cen.y : cen.x);
	float					d;
	int						pix = 0;
	const Range				degSrcL(0, 180), degSrcH(180, 360),
							degDstL(-1, 1), degDstH(1, -1);
	for (int y=0; y<h; y++) {
		cur.y = y;
		for (int x=0; x<w; x++) {
			cur.x = x;
			d = cur.dist2(cen);
			if (d < maxD) {
				mArcP.mDistanceRow[pix] = 1 - (d / maxD);
				float		deg = geo::degree(cur.x - cen.x, cen.y - cur.y);
				if (deg > degSrcL.mMax) deg = degSrcH.translateTo(deg, degDstH);
				else deg = degSrcL.translateTo(deg, degDstL);
				mArcP.mNormalRow[pix] = deg;
				// This is the antialias
				if (d <= maxD-1) rowMask[pix] = 1;
				else rowMask[pix] = maxD - d;
			}

			pix++;
		}
	}

	mSize = s;
	return true;
}

/* misc
 ******************************************************************/
static inline void makeColor(	int r, int g, int b,
								const uchar a, const float scale,
								QColor& clr) {
	r = (int)(r + (255*scale));
	if (r < 0) r = 0; else if (r > 255) r = 255;
	g = (int)(g + (255*scale));
	if (g < 0) g = 0; else if (g > 255) g = 255;
	b = (int)(b + (255*scale));
	if (b < 0) b = 0; else if (b > 255) b = 255;
	clr.setRgb(r, g, b, a);
}
