#ifndef MANAGEDLIST_H
#define MANAGEDLIST_H

/* MANAGED LIST
 * A templatized set of classes implementing a singly-linked list.
 * It sucks to have a custom collection class, but it's often
 * useful to have a list that does not release memory (a performance
 * killer in realtime environments), and there are never guarantees
 * about how other libraries have implemented their collections.
 ******************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <exception>
#include "collections/recyclelist.h"

#pragma once

/* MANAGED-FACTORY
 * The basic memory management behaviour for a managed list.
 * implementing newObject() and deleteObject() is essential,
 * and most clients will want setObject(), but the remaining
 * functions depend on the subclass and use.
 ******************************************************************/
template <class T>
class ManagedFactory
{
public:
	// If sort is true, then clients must implement sort()
	explicit ManagedFactory(const bool sort = false) : mSort(sort) { }
	virtual ~ManagedFactory() { }

	const bool		mSort;

	virtual T		newObject() const = 0;
	virtual void	deleteObject(T) const = 0;
	// Set the object to its initial, empty state
	virtual void	initializeObject(T) const = 0;
	// Set dst to const src.  Note that src can be NULL.
	virtual void	setObject(T dst, T const src) const = 0;
	// Only necessary if mSort is set to true. -1 for <, 0 for =, 1 for >
	virtual int		sortObject(T, T) const { return -1; }
};

/* MANAGED-COMBINER
 * Utility for the removeDuplicates() function.
 ******************************************************************/
template <class T>
class ManagedCombiner
{
public:
	virtual ~ManagedCombiner() { }
	// If the two objects are equal, combine them into the first
	// (the second will be tossed) and return true.
	virtual bool	combine(T, T) = 0;
};

/* MANAGED-LIST
 * A simple list that doesn't free memory when it shrinks, instead
 * saving the entries to be reused later.  Clients supply an object
 * for allocating and deallocating template memory.
 * NOTE:  This class MUST operate on pointer types.  I should
 * enforce that, but first I want to get things working.
 ******************************************************************/
template <class T>
class ManagedList
{
// Exceptions
public:
	class is_empty : public std::exception { };
	class bad_index : public std::exception { };

public:
	// The factory handles all memory management for this class.
	// Note that this class DOES NOT OWN the factory.  The client
	// is responsible for deleting it.  This is because factories
	// have no state, so you can create a static object and reuse.
	ManagedList(const ManagedFactory<T>&);
	virtual ~ManagedList();

	bool						isEmpty() const;
	int							size() const;

	void						clear();

	T							front();
	T							at(const int index);

	// Increase the size of the list with a new object at the front.
	T							pushFront();
	// Increase the size of the list with a new object at the back.
	T							pushBack();
	// Same as above, but call ManagedFactory::set()
	T							pushBack(T const);
	// This is the required variant if sorting in the factory is true.
	// You need to provide a source to copy from, which is pretty annoying.
	void						pushSorted(T const);

	void						popFront();

	void						swap(ManagedList<T>&);

	// Utility for removing duplicates, as designated by the combine() function
	// on the factory.
	void						removeDuplicates(ManagedCombiner<T>&);

private:
	typedef prvlist::_node<T>	Node;

	Node*						mHead;
	Node*						mRecycle;

	const ManagedFactory<T>&	mFactory;

	// Answer a new node, either popped from the recycle list
	// or allocated.
	Node*						newNode();
	void						deleteAll(Node*);

// internal for the iterator
public:
	const Node*					getHead() const;
};

/* MANAGED-LIST-ITERATOR
 * No, not STL compatible at the moment.
 ******************************************************************/
template <class T>
class ManagedListIterator
{
public:
	ManagedListIterator(const ManagedListIterator<T>&);
	ManagedListIterator(const ManagedList<T>&);

	void						operator++();
	ManagedListIterator<T>&		operator=(const ManagedListIterator<T>&);

	bool						hasValue() const;
	const T						operator*();

	bool						hasNext() const;
	const T						peekNext() const;

private:
	ManagedListIterator();
	typedef prvlist::_node<T>	Node;

	const Node*					mN;
};

/* MANAGED-LIST IMPLEMENTATION
 ******************************************************************/
template <class T>
ManagedList<T>::ManagedList(const ManagedFactory<T>& f)
	: mHead(NULL)
	, mRecycle(NULL)
	, mFactory(f)
{
}

template <class T>
ManagedList<T>::~ManagedList()
{
	deleteAll(mHead);
	deleteAll(mRecycle);
}

template <class T>
bool ManagedList<T>::isEmpty() const
{
	return mHead == NULL;
}

template <class T>
int ManagedList<T>::size() const
{
	int						ans = 0;
	prvlist::_node_base*	n = mHead;
	while (n != NULL) {
		ans++;
		n = n->mNext;
	}
	return ans;
}

template <class T>
void ManagedList<T>::clear()
{
	if (mHead == NULL) return;
	if (mRecycle == NULL) mRecycle = mHead;
	else mRecycle->tail()->mNext = mHead;
	mHead = NULL;
}

template <class T>
T ManagedList<T>::front()
{
	if (mHead == NULL) throw is_empty();
	return mHead->mData;
}

template <class T>
T ManagedList<T>::at(const int index)
{
	int			idx = index;
	Node*		n = mHead;
	while (n != NULL) {
		if (idx == 0) return n->mData;
		idx--;
		n = (Node*)(n->mNext);
	}
	throw bad_index();
}

template <class T>
T ManagedList<T>::pushFront()
{
	assert(!mFactory.mSort);

	Node*			n = newNode();
	n->mNext = mHead;
	mHead = n;

	mFactory.initializeObject(n->mData);
	return n->mData;
}

template <class T>
T ManagedList<T>::pushBack()
{
	assert(!mFactory.mSort);

	Node*			n = newNode();
	if (mHead == NULL) mHead = n;
	else mHead->tail()->mNext = n;

	mFactory.initializeObject(n->mData);
	return n->mData;
}

template <class T>
T ManagedList<T>::pushBack(T const src)
{
	T				ans = pushBack();
	mFactory.setObject(ans, src);
	return ans;
}

template <class T>
void ManagedList<T>::pushSorted(T const src)
{
	assert(mFactory.mSort);

	Node*			n = newNode();
	mFactory.setObject(n->mData, src);

	if (mHead == NULL || mFactory.sortObject(n->mData, mHead->mData) < 0) {
		n->mNext = mHead;
		mHead = n;
	} else {
		Node*		prv = mHead;
		while (prv->mNext != NULL) {
			Node*	nxt = (Node*)(prv->mNext);
			if (mFactory.sortObject(n->mData, nxt->mData) < 0) {
				n->mNext = nxt;
				prv->mNext = n;
				return;
			}
			prv = nxt;
		}
		prv->mNext = n;
	}
}

template <class T>
void ManagedList<T>::popFront()
{
	if (mHead == NULL) throw is_empty();
	Node*		ans = mHead;
	mHead = (Node*)(mHead->mNext);
	ans->mNext = mRecycle;
	mRecycle = ans;
}

template <class T>
void ManagedList<T>::swap(ManagedList<T>& o)
{
	if (this != &o) {
		Node*		n = mHead; mHead = o.mHead; o.mHead = n;
	}
}

template <class T>
void ManagedList<T>::removeDuplicates(ManagedCombiner<T>& cmb)
{
	// We always retain the previous.
	Node*			prv = mHead;
	while (prv != NULL && prv->mNext != NULL) {
		Node*		pop = (Node*)(prv->mNext);
		while (pop != NULL && cmb.combine(prv->mData, pop->mData)) {
			prv->mNext = pop->mNext;
			pop->mNext = mRecycle;
			mRecycle = pop;
			pop = (Node*)(prv->mNext);
		}

		prv = (Node*)(prv->mNext);
	}
}

template <class T>
prvlist::_node<T>* ManagedList<T>::newNode()
{
	Node*			n = NULL;
	if (mRecycle != NULL) {
		n = mRecycle;
		mRecycle = (Node*)(mRecycle->mNext);
		n->mNext = NULL;
	} else {
		n = new Node();
		if (n == NULL) throw is_empty();
		n->mData = mFactory.newObject();
	}
	return n;
}

template <class T>
void ManagedList<T>::deleteAll(Node* n)
{
	while (n != NULL) {
		Node*		next = (Node*)(n->mNext);
		n->mNext = NULL;
		mFactory.deleteObject(n->mData);
		delete n;
		n = next;
	}
}

template <class T>
const prvlist::_node<T>* ManagedList<T>::getHead() const
{
	return mHead;
}

/* MANAGED-LIST-ITERATOR IMPLEMENTATION
 ******************************************************************/
template <class T>
ManagedListIterator<T>::ManagedListIterator(const ManagedListIterator<T>& o)
	: mN(o.mN)
{
}

template <class T>
ManagedListIterator<T>::ManagedListIterator(const ManagedList<T>& l)
	: mN(l.getHead())
{
}

template <class T>
ManagedListIterator<T>& ManagedListIterator<T>::operator=(const ManagedListIterator<T>& o)
{
	if (this != &o) mN = o.mN;
	return *this;
}

template <class T>
void ManagedListIterator<T>::operator++()
{
	if (mN != NULL) mN = (Node*)(mN->mNext);
}

template <class T>
bool ManagedListIterator<T>::hasValue() const
{
	return mN != NULL;
}

template <class T>
const T ManagedListIterator<T>::operator*()
{
	if (mN == NULL) return NULL;
	return mN->mData;
}

template <class T>
bool ManagedListIterator<T>::hasNext() const
{
	return mN != NULL && mN->mNext != NULL;
}

template <class T>
const T ManagedListIterator<T>::peekNext() const
{
	if (!hasNext()) return NULL;
	const Node*		nxt = (Node*)(mN->mNext);
	return nxt->mData;
}

#endif // MANAGEDLIST_H
