#ifndef RECYCLELIST_H
#define RECYCLELIST_H

#include <exception>
#include <stdlib.h>

#pragma once

namespace prvlist {

/* NODE
 **************************************************************/
struct _node_base {
	_node_base*		mNext;

	_node_base*		tail() {
		_node_base*	n = this;
		while (n->mNext != NULL) n = n->mNext;
		return n;
	}
};

template <class T>
struct _node : public _node_base {
	T			mData;
};

} // prvlist

/* RECYCLE-LIST
 * A simple list that doesn't free memory when it shrinks, instead
 * saving the entries to be reused later.
 * I'd really like to just use STL list or something in the platform,
 * but as far as I know, there are never any guarantees as to how
 * other collections will be implemented, and I specifically want
 * something that avoids memory allocation and deallocation.
 ******************************************************************/
template <class T>
class RecycleList
{
// Exceptions
public:
	class is_empty : public std::exception { };
	class bad_index : public std::exception { };

public:
    RecycleList();
	virtual ~RecycleList();

	bool						isEmpty() const;
	int							size() const;
	void						clear();

	T&							front();

	bool						pushFront(const T&);
	bool						pushBack(const T&);
	T&							popFront();

	// Take all data nodes and move the data to the supplied list.
	void						moveTo(RecycleList<T>&);

private:
	typedef prvlist::_node<T>	Node;

	Node*						mHead;
	Node*						mRecycle;

	void						deleteAll(Node*);

// internal for the iterator
public:
	const Node*					getHead() const;
};

/* RECYCLE-LIST-ITERATOR.  No, not STL compatible at the moment.
 ******************************************************************/
template <class T>
class RecycleListIterator
{
public:
	RecycleListIterator(const RecycleList<T>&);

	void						operator++();

	bool						hasValue() const;
	const T						operator*();

private:
	RecycleListIterator();
	typedef prvlist::_node<T>	Node;

	const Node*					mN;
};

/* RECYCLE-LIST IMPLEMENTATION
 ******************************************************************/
template <class T>
RecycleList<T>::RecycleList()
	: mHead(NULL)
	, mRecycle(NULL)
{
}

template <class T>
RecycleList<T>::~RecycleList()
{
	deleteAll(mHead);
	deleteAll(mRecycle);
}

template <class T>
bool RecycleList<T>::isEmpty() const
{
	return mHead == NULL;
}

template <class T>
int RecycleList<T>::size() const
{
	int				ans = 0;
	Node*			n = mHead;
	while (n != NULL) {
		ans++;
		n = n->mNext;
	}
	return ans;
}

template <class T>
void RecycleList<T>::clear()
{
	if (mHead != NULL) {
		if (mRecycle == NULL) mRecycle = mHead;
		else mRecycle->tail()->mNext = mHead;

		mHead = NULL;
	}
}

template <class T>
T& RecycleList<T>::front()
{
	if (mHead == NULL) throw is_empty();
	return mHead->mData;
}

template <class T>
bool RecycleList<T>::pushFront(const T& push)
{
	Node*			n = NULL;
	if (mRecycle != NULL) {
		n = mRecycle;
		mRecycle = (Node*)(mRecycle->mNext);
		n->mNext = NULL;
	}
	if (n == NULL) n = new Node();
	if (n == NULL) return false;

	n->mData = push;
	n->mNext = mHead;
	mHead = n;
	return true;
}

template <class T>
bool RecycleList<T>::pushBack(const T& push)
{
	Node*			n = NULL;
	if (mRecycle != NULL) {
		n = mRecycle;
		mRecycle = (Node*)(mRecycle->mNext);
		n->mNext = NULL;
	}
	if (n == NULL) n = new Node();
	if (n == NULL) return false;

	n->mData = push;
	if (mHead == NULL) mHead = n;
	else {
		Node*			tail = mHead;
		while (tail->mNext != NULL) tail = (Node*)(tail->mNext);
		tail->mNext = n;
	}
	return true;
}

template <class T>
T& RecycleList<T>::popFront()
{
	if (mHead == NULL) throw is_empty();
	Node*		ans = mHead;
	mHead = (Node*)(mHead->mNext);
	ans->mNext = mRecycle;
	mRecycle = ans;
	return ans->mData;
}

template <class T>
void RecycleList<T>::moveTo(RecycleList<T>& o)
{
	if (mHead == NULL) return;

	try {
		Node*			n = mHead;
		while (n != NULL) {
			o.pushFront(n->mData);
			n = (Node*)(n->mNext);
		}
	} catch (std::exception&) {
	}

	clear();
}

template <class T>
void RecycleList<T>::deleteAll(Node* n)
{
	while (n != NULL) {
		Node*		next = (Node*)(n->mNext);
		n->mNext = NULL;
		delete n;
		n = next;
	}
}

template <class T>
const prvlist::_node<T>* RecycleList<T>::getHead() const
{
	return mHead;
}

/* RECYCLE-LIST-ITERATOR IMPLEMENTATION
 ******************************************************************/
template <class T>
RecycleListIterator<T>::RecycleListIterator(const RecycleList<T>& l)
	: mN(l.getHead())
{
}

template <class T>
void RecycleListIterator<T>::operator++()
{
	if (mN != NULL) mN = (Node*)(mN->mNext);
}

template <class T>
bool RecycleListIterator<T>::hasValue() const
{
	return mN != NULL;
}

template <class T>
const T RecycleListIterator<T>::operator*()
{
	if (mN == NULL) return NULL;
	return mN->mData;
}

#endif // RECYCLELIST_H
