#include "colorsrc2.h"

#include <QColor>

#include "engine/engineparam.h"
#include "engine/fxdefs.h"

static const QString		KEY("cs2");
const QString&				COLOR_SRC_2_KEY_SZ = KEY;

static const QString		NAME("Color");
// clr0
static const QString		R1_NAME("R1");
static const int			R1_ID = 0;
static const QString		G1_NAME("G1");
static const int			G1_ID = 1;
static const QString		B1_NAME("B1");
static const int			B1_ID = 2;
static const QString		A1_NAME("A1");
static const int			A1_ID = 3;
// clr1
static const QString		R2_NAME("R2");
static const int			R2_ID = 4;
static const QString		G2_NAME("G2");
static const int			G2_ID = 5;
static const QString		B2_NAME("B2");
static const int			B2_ID = 6;
static const QString		A2_NAME("A2");
static const int			A2_ID = 7;

static const QString		C0_SZ("0");
static const QString		C1_SZ("1");

static void assign(param::SetArgs& sa, float& dst);

/* COLOR-SRC-2
 ******************************************************************/
ColorSrc2::ColorSrc2()
{
	mClr0.setRgba(0, 0, 0, 255);
	mClr1.setRgba(255, 255, 255, 255);
	premultiply();
}

ColorSrc2::ColorSrc2(const ColorSrc2& o)
	: ColorSrc(o)
	, mClr0(o.mClr0)
	, mClr1(o.mClr1)
	, mClr0_Pm(o.mClr0_Pm)
	, mClr1_Pm(o.mClr1_Pm)
{
}

ColorSrc2* ColorSrc2::copy() const
{
	return new ColorSrc2(*this);
}

void ColorSrc2::get(const float v, color_f& output) const
{
	const float			left = 1-v;
	output.r = (mClr0_Pm.r*left) + (mClr1_Pm.r*v);
	output.g = (mClr0_Pm.g*left) + (mClr1_Pm.g*v);
	output.b = (mClr0_Pm.b*left) + (mClr1_Pm.b*v);
	output.a = (mClr0_Pm.a*left) + (mClr1_Pm.a*v);
}

void ColorSrc2::scaleAlpha(const float a)
{
	mClr0.a *= a;
	mClr1.a *= a;
	premultiply();
}

void ColorSrc2::getParams(param::List& list, param::GetArgs& get) const
{
	// Convert my premultipled color into 8-bit unmatted values
	// (although I still communicate with floats of 0-1).
	QColor			clr;
	mClr0.getColor(clr);
	list.addParam(&NAME, R1_NAME, R1_ID, clr.redF(), get);
	list.addParam(&NAME, G1_NAME, G1_ID, clr.greenF(), get);
	list.addParam(&NAME, B1_NAME, B1_ID, clr.blueF(), get);
	list.addParam(&NAME, A1_NAME, A1_ID, mClr0.a, get);
	mClr1.getColor(clr);
	list.addParam(&NAME, R2_NAME, R2_ID, clr.redF(), get);
	list.addParam(&NAME, G2_NAME, G2_ID, clr.greenF(), get);
	list.addParam(&NAME, B2_NAME, B2_ID, clr.blueF(), get);
	list.addParam(&NAME, A2_NAME, A2_ID, mClr1.a, get);
}

void ColorSrc2::setParam(const param::UrlIterator& it, param::SetArgs& sa)
{
	if (it.matches(R1_ID))			assign(sa, mClr0.r);
	else if (it.matches(G1_ID))		assign(sa, mClr0.g);
	else if (it.matches(B1_ID))		assign(sa, mClr0.b);
	else if (it.matches(A1_ID))		assign(sa, mClr0.a);
	else if (it.matches(R2_ID))		assign(sa, mClr1.r);
	else if (it.matches(G2_ID))		assign(sa, mClr1.g);
	else if (it.matches(B2_ID))		assign(sa, mClr1.b);
	else if (it.matches(A2_ID))		assign(sa, mClr1.a);

	if (sa.mChanged) premultiply();
}

void ColorSrc2::premultiply()
{
	mClr0_Pm.setAndPremultiply(mClr0);
	mClr1_Pm.setAndPremultiply(mClr1);
}

/* IO
 ******************************************************************/
void ColorSrc2::writeJsonObject(json::Writer& w) const
{
	w.key(CLR_SZ);
	w.stringValue(KEY);

	w.key(C0_SZ);
	w.stringValue(mClr0.asHex());
	w.key(C1_SZ);
	w.stringValue(mClr1.asHex());
}

void ColorSrc2::jsonString(json::Stack&, const QString& k, const QString& v)
{
	if (k == C0_SZ) mClr0.fromHex(v);
	if (k == C1_SZ) mClr1.fromHex(v);
	else return;
	premultiply();
}

/* Misc
 ******************************************************************/
static void assign(param::SetArgs& sa, float& dst)
{
	float			c = sa.mV;
	if (c < 0) c = 0; else if (c > 1) c = 1;	// limit
	if (dst != c) {
		dst = c;
		sa.mChanged = true;
	}
}
