#include "arcdefs.h"

#include <assert.h>
#include <QDebug>

#include "engine/arc.h"

using namespace arc;

static const QString	ARC_SZ_S("arc");
const QString&			arc::ARC_SZ = ARC_SZ_S;

/* WORK-PARAMS
 ******************************************************************/
WorkParams::WorkParams()
	: mDistanceRow(NULL)
	, mNormalRow(NULL)
	, mRowSize(0)
	, mArc(NULL)
	, mFlags(0)
{
	// Technically we can optimize by turning off sources that
	// aren't in use, but I'm not sure if it's worth all the
	// conditionals that results in
	for (int k=0; k<SRC_SIZE; k++) mSrc[k].setIsOn(true);
}

WorkParams::~WorkParams()
{
	while (!mEval.isEmpty()) delete mEval.popFront();
}

bool WorkParams::begin(ArcHead* h, const int rowSize, const int flags)
{
	if (h == NULL || rowSize < 1) return false;
	mArc = h;
	mDistanceRow = mSrc[DISTANCE_SRC].row(rowSize);
	mNormalRow = mSrc[NORMAL_SRC].row(rowSize);
	mRowSize = rowSize;
	mFlags = flags;
	return true;
}

bool WorkParams::hasFlag(const int f) const
{
	return (mFlags&f) != 0;
}

ModEval* WorkParams::acquireEval(	Arc* arc, const int idx,
									float* v0, float* v1,
									float* v2, float* v3)
{
	assert(arc != NULL);
	if (mArc == NULL) return NULL;
	// Acquire a new modEval
	ModEval*		ans = (mEval.isEmpty() ? new ModEval() : mEval.popFront());
	if (ans == NULL) return NULL;
	// Load it up
	mDst[0] = v0; mDst[1] = v1; mDst[2] = v2; mDst[3] = v3;
	setTargets(ans, arc->getId(), idx, mDst);
	if (ans->mSize > 0) return ans;

	releaseEval(ans);
	return NULL;
}

void WorkParams::releaseEval(ModEval* eval)
{
	if (eval == NULL) return;
	if (!mEval.pushFront(eval)) delete eval;
}

Mask* WorkParams::acquireScratch()
{
	if (mScratch.isEmpty()) return new Mask();
	return mScratch.popFront();
}

void WorkParams::releaseScratch(Mask* m)
{
	if (m == NULL) return;
	if (!mScratch.pushFront(m)) delete m;
}

void WorkParams::setTargets(	ModEval* eval, const int arcId,
								const int startIdx, float** v)
{
	eval->mSize = 0;

	const ModTable&			tbl = mArc->mLinks;
	for (int k=0; k<tbl.mLink.size(); k++) {
		const ModLink&		link = tbl.mLink[k];
		// Source is within bounds
		if (link.mSrc < 0 || link.mSrc >= SRC_SIZE) continue;
		// Link is for this arc
		if (link.mDstArc != arcId) continue;
		// Dst param is within the list provided
		const int			idx = link.mDstParam - startIdx;
		if (idx < 0 || idx >= 4 || v[idx] == NULL) continue;

		float*				srcRow = mSrc[link.mSrc].row(mRowSize);
		if (srcRow != NULL) {
			eval->mFlags[eval->mSize] = link.mFlags;
			eval->mSrcRow[eval->mSize] = srcRow;
			eval->mDst[eval->mSize] = v[idx];
			eval->mOrig[eval->mSize] = *(v[idx]);
			eval->mR[eval->mSize] = link.mR;

			eval->mSize++;
			if (eval->mSize >= MAX_EVAL) return;
		}
	}
}

/* MOD-EVAL
 ******************************************************************/
ModEval::ModEval()
{
}

void ModEval::moveTo(const int rowIdx) {
	// Reset each mod dest first, since a dest might occur
	// multiple times, and this operation is cumulative.
	for (int k=0; k<mSize; k++) {
		*(mDst[k]) = mOrig[k];
	}

	// Accum the mod sources
	for (int k=0; k<mSize; k++) {
		float		v = mSrcRow[k][rowIdx];
		if ((mFlags[k]&ModLink::RANGE_ON_F) != 0) {
			v = RANGE_UNI.translateTo(v, mR[k]);
		}
		*(mDst[k]) += v;
	}
}
