#include "colorf.h"

#include <QColor>
#include <QDebug>

static const QChar		SPACE('0');

static int intFromFloat(const float cf);
static float floatFromHex(const QChar& h0, const QChar& h1);

// Macros for accessing colour elements in the dst array during compositing
#define PIXR(pix)		(pix)
#define PIXG(pix)		(pix+1)
#define PIXB(pix)		(pix+2)
#define PIXA(pix)		(pix+3)

/* COLOR-F
 ******************************************************************/
color_f::color_f()
	: r(0)
	, g(0)
	, b(0)
	, a(0)
{
}

color_f::color_f(const color_f& o)
	: r(o.r)
	, g(o.g)
	, b(o.b)
	, a(o.a)
{
}

color_f& color_f::operator=(const color_f& o)
{
	if (this != &o) {
		r = o.r;
		g = o.g;
		b = o.b;
		a = o.a;
	}
	return *this;
}

void color_f::setRgba(const int ru, const int gu, const int bu, const int au)
{
	a = (float)au / 255.0f;
	r = (float)ru / 255.0f;
	g = (float)gu / 255.0f;
	b = (float)bu / 255.0f;
}

void color_f::setAndPremultiply(const color_f& o)
{
	a = o.a;
	r = o.r * o.a;
	g = o.g * o.a;
	b = o.b * o.a;
}

void color_f::getColor(QColor& clr) const
{
	clr.setRgb(intFromFloat(r), intFromFloat(g), intFromFloat(b), intFromFloat(a));
}

void color_f::premultiply(const float au)
{
	a = au;
	r *= au;
	g *= au;
	b *= au;
}

void color_f::unPremultiply_Pm()
{
	const float		s = 1 / a;
	r *= s;
	g *= s;
	b *= s;
}

void color_f::setA_Pm(const float au)
{
	const float		ratio = (au / a);
	r *= ratio;
	g *= ratio;
	b *= ratio;
	a = au;
}

QString color_f::asHex() const
{
	return QString("%1%2%3%4")
			.arg(intFromFloat(r), 2, 16, SPACE)
			.arg(intFromFloat(g), 2, 16, SPACE)
			.arg(intFromFloat(b), 2, 16, SPACE)
			.arg(intFromFloat(a), 2, 16, SPACE);
}

bool color_f::fromHex(const QString& hex)
{
	if (hex.size() < 8) return false;
	r = floatFromHex(hex.at(0), hex.at(1));
	g = floatFromHex(hex.at(2), hex.at(3));
	b = floatFromHex(hex.at(4), hex.at(5));
	a = floatFromHex(hex.at(6), hex.at(7));
	return true;
}

/* misc
 ******************************************************************/
static inline int intFromFloat(const float cf)
{
	int			ci = (int)(cf * 255);
	if (ci <= 0) return 0;
	if (ci >= 255) return 255;
	return ci;
}

static int intFromHex(const QChar& h)
{
	const char		c = h.toAscii();
	if (c >= '0' && c <= '9') return c-'0';
	if (c >= 'a' && c <= 'f') return 10 + (c-'a');
	if (c >= 'A' && c <= 'F') return 10 + (c-'A');
	return 0;
}

static float floatFromHex(const QChar& h0, const QChar& h1)
{
	int				v = (intFromHex(h0) << 4) | (intFromHex(h1));
	if (v < 0) v = 0; else if (v > 255) v = 255;
	return float(v) / 255.0f;
}

QDebug& operator<<(QDebug& d, const color_f& o)
{
	(d << "r=").nospace();
	(d << o.r).nospace();
	(d << "g=").nospace();
	(d << o.g).nospace();
	(d << "b=").nospace();
	(d << o.b).nospace();
	(d << "a=").nospace();
	(d << o.a).nospace();
	return d;
}

/* COMPOSITING
 ******************************************************************/
void srcOver(const color_f& src, float* dst, int pix)
{
	const float		am = (1-src.a);
	pix *= 4;
	dst[PIXR(pix)] = src.r + (am*dst[PIXR(pix)]);
	dst[PIXG(pix)] = src.g + (am*dst[PIXG(pix)]);
	dst[PIXB(pix)] = src.b + (am*dst[PIXB(pix)]);
	dst[PIXA(pix)] = src.a + (am*dst[PIXA(pix)]);
}

#undef PIXR
#undef PIXG
#undef PIXB
#undef PIXA
