#include "colorsrchead.h"

#include "stdlib.h"
#include <QDebug>
#include "compositing/colorsrc1.h"
#include "compositing/colorsrclist.h"
#include "engine/engineparam.h"
#include "engine/fxdefs.h"
#include "io/json_reader.h"

// Instead of a registry, hardcoded references
#include "compositing/colorsrc1.h"
#include "compositing/colorsrc2.h"

/* COLOR-SRC-HEAD
 ******************************************************************/
ColorSrcHead::ColorSrcHead()
	: mSrc(NULL)
{
}

ColorSrcHead::ColorSrcHead(const ColorSrcHead& o)
	: Parent(o)
	, json::Reader(o)
	, ColorSrcTarget(o)
	, mKey(o.mKey)
	, mSrc(o.mSrc != NULL ? o.mSrc->copy() : NULL)
{
}

ColorSrcHead::ColorSrcHead(const QString& key, ColorSrc* src)
	: mKey(key)
	, mSrc(src)
{
}

ColorSrcHead::~ColorSrcHead()
{
	setSrc(NULL);
}

ColorSrcHead& ColorSrcHead::operator=(const ColorSrcHead& o)
{
	if (this != &o) {
		mKey = o.mKey;
		setSrc(o.mSrc != NULL ? o.mSrc->copy() : NULL);
	}
	return *this;
}

bool ColorSrcHead::isEmpty() const
{
	return mSrc == NULL;
}

void ColorSrcHead::setId(const int id)
{
	mId = id;
}

QString ColorSrcHead::getKey() const
{
	return mKey;
}

const ColorSrc* ColorSrcHead::getSrc() const
{
	return mSrc;
}

void ColorSrcHead::scaleAlpha(const float a)
{
	if (mSrc != NULL) mSrc->scaleAlpha(a);
}

void ColorSrcHead::setSrc(ColorSrc* src)
{
	if (mSrc == src) return;
	delete mSrc;
	mSrc = src;
}

param::Parent* ColorSrcHead::findParent(param::UrlIterator& it)
{
	// Just a passthrough
	if (mSrc == NULL) return NULL;
	// The color sources aren't composites and don't have IDs.
//	++it;
	return mSrc->findParent(it);
}

void ColorSrcHead::getParams(param::List& list, param::GetArgs& get) const
{
	if (mSrc == NULL) return;

	param::List			l;
	mSrc->getParams(l, get);
	l.prepend(mId);
	list.takeAll(l);
}

void ColorSrcHead::setParam(const param::UrlIterator&, param::SetArgs&)
{
}

void ColorSrcHead::addTypeParam(	param::List& list, param::GetArgs& get,
									const int bit, const int flags) const
{
	const float				exp = param::valueFromFlag(flags, bit);
	param::List				group;
	getParams(group, get);
	list.addTypeParam(fx::COLOR_SZ, mId, param::COLOR_BROWSER_TYPE, get, exp, bit, &mKey, &group);
}

void ColorSrcHead::swap(ColorSrcHead& o)
{
	if (this != &o) {
		QString		k = mKey;	mKey = o.mKey;	o.mKey = k;
		ColorSrc*	s = mSrc;	mSrc = o.mSrc;	o.mSrc = s;
	}
}

/* COLOR-SRC-MATCH
 ******************************************************************/
ColorSrcMatch::ColorSrcMatch()
{
}

ColorSrcMatch::~ColorSrcMatch()
{
}

/* COLOR-SRC-HEAD::MATCH-DEFAULT
 ******************************************************************/
class ColorSrcHead::MatchDefault : public ColorSrcMatch {
public:
	MatchDefault(const QString& key)	 : mKey(key), mDefault(0, 0, 0, 255)	{ }

	const QString			mKey;
	const ColorSrc1			mDefault;

	virtual float			quality(const ColorSrcHead& h) const
	{
		if (h.getKey() == mKey) return 2;
		return 0;
	}

	virtual void			getDefault(ColorSrcHead& h) const
	{
		ColorSrc*			c = mDefault.copy();
		if (c != NULL) h.setSrc(c);
	}
};

/* COLOR-SRC-MATCH defaults
 ******************************************************************/
static const ColorSrcHead::MatchDefault	SOLO(COLOR_SRC_DEFAULT_SOLO);
const ColorSrcMatch&						COLOR_SRC_SOLO = SOLO;

static const ColorSrcHead::MatchDefault	DUO(COLOR_SRC_DEFAULT_DUO);
const ColorSrcMatch&						COLOR_SRC_DUO = DUO;


/* COLOR-SRC-HEAD::READER
 * A generic class clients can instantiate to read in arcs.
 ******************************************************************/
class ColorSrcHead::Reader : public json::Reader
{
public:
	explicit Reader(ColorSrcTarget& t) : mT(t) { }

	virtual void		jsonString(json::Stack& s, const QString& k, const QString& v) {
		if (k == CLR_SZ) {
			ColorSrc*			cs = NULL;
			if (v == COLOR_SRC_1_KEY_SZ) cs = new ColorSrc1(0, 0, 0, 255);
			else if (v == COLOR_SRC_2_KEY_SZ) cs = new ColorSrc2();
			if (cs != NULL) {
				mT.parseClr(cs);
				s.push(cs);
			}
		}

	}

	virtual void		jsonPop() { delete this; }

private:
	ColorSrcTarget&		mT;
};

/* IO
 * Sort of out of order so it can use the reader.
 ******************************************************************/
void ColorSrcHead::writeJsonObject(json::Writer& w, const QString& key) const
{
	w.key(key);
	w.beginObject();
	writeJson(w);
	w.endObject();
}

void ColorSrcHead::writeJson(json::Writer& w) const
{
	if (mKey.length() < 1) {
		qDebug() << "ERROR ColorSrcHead::writeJsonObject() missing key";
		return;
	}

	w.key(fx::KEY_SZ);
	w.stringValue(mKey);

	if (mSrc != NULL) {
		w.key(fx::DATA_SZ);
		w.beginObject();
		mSrc->writeJsonObject(w);
		w.endObject();
	}
}

void ColorSrcHead::startObject(json::Stack& s, const QString& k)
{
	if (k == fx::DATA_SZ) s.push(new Reader(*this));
}

void ColorSrcHead::jsonString(json::Stack&, const QString& k, const QString& v)
{
	if (k == fx::KEY_SZ) mKey = v;
}

void ColorSrcHead::parseClr(ColorSrc* cs)
{
	delete mSrc;
	mSrc = cs;
}
