#include "fx.h"

#include <QMutex>
#include <QMutexLocker>
#include "collections/recyclelist.h"
#include "fxui/parameditor.h"
#include "param/param.h"
#include "math/geo.h"

/* FX
 ******************************************************************/
Fx::Fx(const int uniqueId)
	: mFlags(0)
{
	mId = uniqueId;
}

Fx::Fx(const Fx& o)
	: param::Parent(o)
	, Reader(o)
	, mFlags(o.mFlags)
{
}

Fx::~Fx()
{
}

void Fx::init(fx::Setup& setup)
{
	if (mId < 0) mId = setup.nextFxUniqueId();
}

bool Fx::setFlag(const int bit, const bool on)
{
	const int			prev = mFlags;
	if (on) mFlags |= bit;
	else mFlags &= ~bit;
	return mFlags != prev;
}

int Fx::forEach(fx::Visitor& v, fx::VisitParams& vp)
{
	return v.visit(*this, vp);
}

int Fx::edit(fx::EditOp&)
{
	return fx::CONTINUE;
}

FxUi* Fx::newUi(FxUi::Create& c) const
{
	param::Url		url;
	Parent::makeUrl(this, url);
	return new fxui::ParamEditor(c, url);
}

void Fx::jsonFloat(json::Stack&, const QString& k, const float v)
{
	if (k == fx::ID_SZ) mId = geo::iroundf(v);
}

void Fx::writeJsonHeader(json::Writer& w, const QString& key) const
{
	w.beginObject();
	w.key(fx::FX_SZ);
	w.stringValue(key);
	w.key(fx::ID_SZ);
	w.intValue(mId);
}

void Fx::writeJsonFooter(json::Writer& w) const
{
	w.endObject();
}

/* FX-FACTORY
 ******************************************************************/
FxFactory::FxFactory()
{
}

FxFactory::~FxFactory()
{
}

/* FX-FACTORY static
 ******************************************************************/
static RecycleList<FxFactory*>		FACTORY;
// Make it thread safe although currently all access should be
// via the UI thread.  Still, it's a global, so we have no real control.
static QMutex						FACTORY_LOCK;

class FxFactory::CreateVisitor : public fx::FactoryVisitor {
public:
	CreateVisitor(const QString& key) : mKey(key), mOut(NULL) { }

	const QString		mKey;
	Fx*					mOut;

	virtual int			visit(const FxFactory& f) {
		if (mOut != NULL) return fx::STOP;
		if (f.match(mKey) > 0.95f) {
			mOut = f.create();
			return fx::STOP;
		}
		return fx::CONTINUE;
	}
};

void FxFactory::add(FxFactory* fx)
{
	if (fx == NULL) return;
	QMutexLocker		lock(&FACTORY_LOCK);
	try {
		// XXX Actually, I think the default behaviour returns false, not an exception
		FACTORY.pushBack(fx);
	} catch (std::exception&) {
		delete fx;
	}
}

void FxFactory::shutdown()
{
	QMutexLocker		lock(&FACTORY_LOCK);
	while (!FACTORY.isEmpty()) delete FACTORY.popFront();
}

Fx* FxFactory::create(const QString& key)
{
	if (key.length() < 1) return NULL;
	CreateVisitor		v(key);
	forEach(v);
	return v.mOut;
}

void FxFactory::forEach(fx::FactoryVisitor& v)
{
	QMutexLocker		lock(&FACTORY_LOCK);
	RecycleListIterator<FxFactory*>		it(FACTORY);
	while (it.hasValue()) {
		const FxFactory*		e = *it;
		if (e != NULL) {
			if (v.visit(*e) <= fx::STOP) return;
		}
		++it;
	}
}
