#ifndef FX_H
#define FX_H

#include <QList>
#include "fxui.h"
#include "fxdefs.h"
#include "engine/engineparam.h"
#include "io/json.h"

class Frame;
class FxUi;

/* FX
 * Abstract processing object.  All it does is manipulate
 * frame data.  The main processing call is work().  FX
 * objects do not need to be thread safe -- a copy exists
 * in the rendering thread.
 ******************************************************************/
class Fx : public param::Parent
		 , public json::Reader
{
public:
	virtual ~Fx();

	virtual Fx*				copy() const = 0;

	// Clients should initialize to their default state.
	// ALWAYS CALL THE PARENT, which sets up the ID.
	// Only initialize invalid values, since this is
	// also called after file IO.
	virtual void			init(fx::Setup&);

	virtual QString			getLabel() const = 0;

	// Follow the URL to find the parent at the given param.
	virtual param::Parent*	findParent(param::UrlIterator&) = 0;

	virtual bool			setFlag(const int bit, const bool on);

	virtual int				forEach(fx::Visitor&, fx::VisitParams&);

	// Subclasses need to implement the edit operation.  Mostly for chains.
	virtual int				edit(fx::EditOp&);

	// The main processing entry point.  The buffers allocated
	// in the args will always be large enough to accomodate
	// the output.
	virtual void			work(fx::WorkArgs&, fx::WorkOutput&) = 0;

	// A default UI editing the parameters is provided, otherwise
	// subclasses override for a custom UI.  Note that there is
	// no default browser -- asking for it will just flip to edit mode.
	virtual FxUi*			newUi(FxUi::Create&) const;

	// IO
	virtual void			writeJson(json::Writer&) const = 0;
	// Default behaviour for reading in the ID.
	virtual void			jsonFloat(json::Stack&, const QString&, const float);

protected:
	Fx(const int uniqueId);
	Fx(const Fx&);

	// Subclasses can use this to store persistent bit data such as which
	// parameter groups are expanded.
	int						mFlags;

	// A convenience since all clients should have the same header / footer
	void					writeJsonHeader(json::Writer&, const QString& key) const;
	void					writeJsonFooter(json::Writer&) const;

private:
	Fx();
};

/* FX-FACTORY
 * Each FX needs to create and install a factory to register
 * with the system.
 ******************************************************************/
class FxFactory
{
public:		// static
	// Clients setup/tear down
	static void					add(FxFactory*);
	static void					shutdown();

	// Client accessing
	static Fx*					create(const QString& key);
	static void					forEach(fx::FactoryVisitor&);

public:
	virtual ~FxFactory();

	virtual QString				label() const = 0;
	virtual QString				key() const = 0;

	// Answer a match quality, between 0 (no match) and 1 (exact).
	virtual float				match(const QString& key) const = 0;

	virtual Fx*					create() const = 0;

protected:
	FxFactory();

private:
	FxFactory(const FxFactory&);

	class CreateVisitor;
};

#endif // FX_H
