#include "fxdefs.h"

#include <math.h>
#include <QDateTime>
#include "fx.h"
#include "math/primitives/segment3.h"
#include "math/cppolygon.h"

using namespace fx;

static const QString	ARC_SZ_S("arc");
const QString&			fx::ARC_SZ = ARC_SZ_S;

static const QString	COLOR_SZ_S("color");
const QString&			fx::COLOR_SZ = COLOR_SZ_S;

static const QString	DATA_SZ_S("data");
const QString&			fx::DATA_SZ = DATA_SZ_S;

static const QString	FX_SZ_S("fx");
const QString&			fx::FX_SZ = FX_SZ_S;

static const QString	ID_SZ_S("id");
const QString&			fx::ID_SZ = ID_SZ_S;

static const QString	KEY_SZ_S("key");
const QString&			fx::KEY_SZ = KEY_SZ_S;

static const QString	SHAPE_SZ_S("shape");
const QString&			fx::SHAPE_SZ = SHAPE_SZ_S;

/* FX-DEFS::VISIT-PARAMS
 ******************************************************************/
VisitParams::VisitParams()
	: mDepth(0)
{
}

/* CELL-GENERATOR
 ******************************************************************/
CellGenerator::CellGenerator()
{
}

CellGenerator::~CellGenerator()
{
}

/* FX::EDIT-OP
 ******************************************************************/
EditOp::EditOp()
	: mSuccess(false)
	, mMode(0)
	, mFxId(-1)
	, mFx(NULL)
{
}

EditOp::~EditOp()
{
	deleteFx();
}

bool EditOp::success() const
{
	return mSuccess;
}

void EditOp::setToSuccess()
{
	mSuccess = true;
}

const int EditOp::mode() const
{
	return mMode;
}

const int EditOp::fxId() const
{
	return mFxId;
}

Fx* EditOp::popFx()
{
	Fx*		ans = mFx;
	mFx = NULL;
	return ans;
}

void EditOp::setToAdd(const int parentFxId, Fx* fxToAdd)
{
	deleteFx();
	mSuccess = false;
	mMode = ADD_MODE;
	mFxId = parentFxId;
	mFx = fxToAdd;
}

void EditOp::setToRemove(const int fxIdToRemove)
{
	deleteFx();
	mSuccess = false;
	mMode = REMOVE_MODE;
	mFxId = fxIdToRemove;
}

void EditOp::deleteFx()
{
	delete mFx;
	mFx = NULL;
}

/* WORK-ARGS
 ******************************************************************/
WorkArgs::WorkArgs()
	: mSize(0, 0)
{
	uncacheAll();
}

WorkArgs::~WorkArgs()
{
}

const QPolygonF& WorkArgs::viewShape() const
{
	return mShape;
}

QPolygonF& WorkArgs::editShape()
{
	mShapeStamp = QDateTime::currentMSecsSinceEpoch();
	return mShape;
}

bool WorkArgs::begin(const QSize& s)
{
	const int			cellSize = s.width() * s.height();
	const int			imgSize = cellSize*4;
	if (mSize != s) {
		if (mCell.get(cellSize) == NULL) return false;
		if (mImg.get(imgSize) == NULL) return false;
		mSize = s;
		uncacheAll();
	}

	float*				img = mImg.get(imgSize);
	if (img == NULL) return false;

	set_to_0(img, imgSize);
	return true;
}

Cell* WorkArgs::getCell()
{
	return mCell.get(mSize.width() * mSize.height());
}

float* WorkArgs::getImg()
{
	return mImg.get(mSize.width() * mSize.height() * 4);
}

void WorkArgs::distanceCached(const float minD, const float maxD)
{
	// Special rule -- distance is locked to shape
	mDistanceStamp = mShapeStamp;
	mMinD = minD;
	mMaxD = maxD;
}

void WorkArgs::getDistanceRange(float* minD, float* maxD) const
{
	if (minD != NULL) *minD = mMinD;
	if (maxD != NULL) *maxD = mMaxD;
}

void WorkArgs::cache(CellGenerator& gen)
{
	const int			w = mSize.width(), h = mSize.height();
	const int			size = w*h;
	Cell*				cell = mCell.get(size);
	if (cell == NULL) return;

	gen.cache(cell, mSize, *this);
}

// Convert an alpha premultiplied colour to the local colorspace,
// a packed 32 bit uint.
static inline uint unmatte(const float rm, const float gm, const float bm, const float am)
{
	uint		r = 0, g = 0, b = 0, a = (uint)(am*255.0f);
	if (am <= 0) {
		r = (uint)(rm*255.0f);
		g = (uint)(gm*255.0f);
		b = (uint)(bm*255.0f);
	} else {
		r = (uint)((rm*255.0f) / am);
		g = (uint)((gm*255.0f) / am);
		b = (uint)((bm*255.0f) / am);
	}
	// Limit the components
	if (r > 255) r = 255;
	if (g > 255) g = 255;
	if (b > 255) b = 255;
	if (a > 255) a = 255;
	// Pack
	return (a << 24) | (r << 16) | (g << 8) | b;
}

void WorkArgs::finish(QImage& img)
{
	const int				w = mSize.width(), h = mSize.height();
	int						spix = 0;
	const float*			src = mImg.get(w*h*4);
	if (src == NULL) return;
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			img.setPixel(x, y, unmatte(src[spix], src[spix+1], src[spix+2], src[spix+3]));
			spix += 4;
		}
	}
}

Mask* WorkArgs::acquireMask(const int size)
{
	Mask*		ans = NULL;
	if (!mScratch.isEmpty()) ans = mScratch.popFront();
	if (ans == NULL) ans = new Mask();
	if (ans != NULL && ans->get(size) != NULL) return ans;
	delete ans;
	return NULL;
}

void WorkArgs::releaseMask(Mask* m)
{
	if (m == NULL) return;
	mScratch.pushFront(m);
}

void WorkArgs::uncacheAll()
{
	mNormalCached = 0;
	mShapeStamp = 0;
	mDistanceStamp = 0;
	mNormalStamp = 0;
	mMinD = mMaxD = 0;
}

/* FX::WORK-OUTPUT
 ******************************************************************/
WorkOutput::WorkOutput(const QSize& size)
	: mImg(size, QImage::Format_ARGB32)
{
}

/* FX::FIND-VISITOR
 ******************************************************************/
FindVisitor::FindVisitor(const int fxId)
{
	setId(fxId);
}

void FindVisitor::setId(const int fxId)
{
	mId = fxId;
	mFx = NULL;
}

int FindVisitor::visit(Fx& fx, VisitParams&)
{
	if (fx.getId() != mId) return fx::CONTINUE;
	mFx = &fx;
	return fx::STOP;
}
