#ifndef FXDEFS_H
#define FXDEFS_H

#include <QImage>
#include <QPolygonF>

#include "collections/recyclelist.h"
#include "engine/archead.h"
#include "engine/cell.h"
#include "param/paramurl.h"

class ColorSrcHead;
class ColorSrcMatch;
class Fx;
class FxFactory;

namespace fx {

class WorkArgs;

// For visting
static const int		CONTINUE = 1;
static const int		STOP = 0;

// Constants
extern const QString&	ARC_SZ;
extern const QString&	COLOR_SZ;
extern const QString&	DATA_SZ;
extern const QString&	FX_SZ;
extern const QString&	ID_SZ;
extern const QString&	KEY_SZ;
extern const QString&	SHAPE_SZ;

/* VISIT-PARAMS
 ******************************************************************/
struct VisitParams {
	int					mDepth;

	// The depth that contains the layer FX
	static const int	LAYER_DEPTH = 0;
	// The depth of each layer's top chain
	static const int	ROOT_DEPTH = LAYER_DEPTH + 1;

	VisitParams();
};

/* VISITOR
 * For walking the Fx graph.
 ******************************************************************/
class Visitor {
public:
	virtual ~Visitor()	{ }

	// Answer CONTINUE, etc.
	virtual int			visit(Fx&, VisitParams&) = 0;

protected:
	Visitor()			{ }
};

/* FACTORY-VISITOR
 ******************************************************************/
class FactoryVisitor {
public:
	virtual ~FactoryVisitor()	{ }

	// Answer CONTINUE, etc.
	virtual int			visit(const FxFactory&) = 0;

protected:
	FactoryVisitor()	{ }
};

/* CELL-GENERATOR.  Abstract interface for clients that
 * want to cache Cell info.  This is a work in progress.
 ******************************************************************/
class CellGenerator {
public:
	virtual ~CellGenerator();

	// Clients cache any info they need to onto the cell.  They're
	// provided the work args, to check and manipulate.
	virtual void			cache(Cell*, const QSize&, WorkArgs&) = 0;

protected:
	CellGenerator();
};

/* FX::SETUP
 * Setup/initialization behaviour for the fx.
 ******************************************************************/
class Setup
{
public:
	virtual int				nextFxUniqueId() = 0;

	virtual QSize			getCanvasSize() const = 0;

	// Passthrough to Environment
	virtual QString			getInstallPath(const QString& sub) const = 0;

	// Copy the arc at the index in master list to the supplied arg.
	virtual bool			getArc(ArcHead&, const int idx) const = 0;

	// Get the default colorsrc
	virtual bool			getColorSrc(ColorSrcHead&, const ColorSrcMatch&) const = 0;
};

/* FX::EDIT-OP
 * An edit operation on an FX.  This is just a package of
 * information, the clients (which generally means a Chain) needs
 * to deal with it.
 ******************************************************************/
class EditOp {
public:
	static const int	ADD_MODE = (1);
	static const int	REMOVE_MODE = (2);

public:
	EditOp();
	~EditOp();

	// Set if the edit was successful
	bool				success() const;
	void				setToSuccess();

	const int			mode() const;
	const int			fxId() const;
	Fx*					popFx();

	void				setToAdd(const int parentFxId, Fx* fxToAdd);
	void				setToRemove(const int fxIdToRemove);

private:
	bool				mSuccess;
	int					mMode;
	int					mFxId;
	// I own the FX; if anyone wants it, they have to pop it.
	Fx*					mFx;

	void				deleteFx();
};

/* FX::WORK-ARGS.  Arguments to work().
 ******************************************************************/
class WorkArgs {
public:
	// Hacked in persistent cache info.  This is being used so
	// that when multiple bevels are stacked in a chain, each
	// one knows whether or not it needs to cache the normals.
	int					mNormalCached;

public:
	WorkArgs();
	~WorkArgs();

	// View the shape and the last time it was edited
	const QPolygonF&	viewShape() const;
	// Edit the shape.  Clients need to clear() if that's what they want.
	QPolygonF&			editShape();

	// Time stamps for last edits of the data I cache
	qint64				shapeStamp() const;
	qint64				distanceStamp() const;
	qint64				normalStamp() const;

	// Make sure buffers are allocated and cleared.
	bool				begin(const QSize&);

	// Get cached geometry info
	Cell*				getCell();
	// Get the current working image.  Image data stored as floats
	// with values of 0-1 in RGBA order, alpha premultiplied.
	float*				getImg();

	// Notification that the distance has been cached, and these
	// are the min and max values
	void				distanceCached(const float minD, const float maxD);
	void				getDistanceRange(float* minD, float* maxD) const;

	// Supply a client for caching data.
	void				cache(CellGenerator&);

	void				finish(QImage&);

	// Clients can request buffers of data for operating on.  Note
	// that if a Mask is returned, it's been verified to be the correct
	// size, so clients don't need to double check.
	Mask*				acquireMask(const int size);
	void				releaseMask(Mask*);

private:
	QPolygonF			mShape;
	QSize				mSize;
	CellMask			mCell;
	Mask				mImg;
	qint64				mShapeStamp, mDistanceStamp, mNormalStamp;
	// Distance info
	float				mMinD, mMaxD;

	RecycleList<Mask*>	mScratch;

	void				uncacheAll();
};

/* FX::WORK-OUTPUT
 * Store the result of work processing.
 ******************************************************************/
class WorkOutput {
public:
	WorkOutput(const QSize&);

	// A buffer to store our final pixels
	QImage				mImg;
};

/* FX::FIND-VISITOR
 * Utility to find the FX with the given ID.
 ******************************************************************/
class FindVisitor : public Visitor {
public:
	FindVisitor(const int fxId);

	Fx*					mFx;

	void				setId(const int fxId);

	virtual int			visit(Fx&, VisitParams&);

private:
	int					mId;
};

/* inline
 ******************************************************************/
inline qint64 WorkArgs::shapeStamp() const { return mShapeStamp; }
inline qint64 WorkArgs::distanceStamp() const { return mDistanceStamp; }
inline qint64 WorkArgs::normalStamp() const { return mNormalStamp; }

} // fx

#endif // FXDEFS_H
