#ifndef FXUI_H
#define FXUI_H

#include <QMouseEvent>
#include <QRect>

#include "engine/fxdefs.h"
#include "engine/fxuidefs.h"

class ArcList;
class ColorSrcList;
class Context;
class Environment;
class Look;

/* FX-UI.  Superclass for FX units to supply a UI.  This is pretty
 * tightly defined right now, with a UI composed of a series of
 * entries.
 ******************************************************************/
class FxUi
{
public:
	static const int		EDIT = 0;
	static const int		BROWSE = 1;

	/* PARENT.  The owner of the UI.
	 **************************************************************/
	public:
		class Entry;			// forward declaration

		class Parent {
		public:
			virtual ~Parent();

			virtual int			entryH() const = 0;
			// I do not become the owner of the entry.  The
			// fxui must guarantee the entry exists for as long
			// as the plug is my child.
			virtual void		add(Entry*) = 0;
			virtual void		addBrowser(fxui::BrowserModel*, const int flags, const bool expanded) = 0;

		protected:
			Parent();
		};

	/* MOUSE-ARGS.  Parameters to mouse
	 * methods.
	 **************************************************************/
	public:
		struct mouse_args {
			QMouseEvent*		mEvent;
			Entry*				mEntry;
			QRect				mFrame;
			// Hack to let the browser UI communicate that an update is needed
			bool				mNeedsUpdate;

			mouse_args();
		private:
			mouse_args(const mouse_args&);
			void				operator=(const mouse_args&);
		};

	/* PAINT-ARGS.  Parameters to paint
	 * method.
	 **************************************************************/
	public:
		struct paint_args {
			QPainter*			mPainter;
			const Look*			mLook;
			QRect				mFrame;

			paint_args();
		private:
			paint_args(const paint_args&);
			void				operator=(const paint_args&);
		};

	/* ENTRY.  A single ui object.
	 **************************************************************/
	public:
		class Entry {
		public:
			static const int	NO_PARAMS = -1;
			static const int	WRONG_PARAM = 0;
			static const int	PARAM = 1;

			static const int	DRAW_ON_TOP_F = (1<<0);
				// Tells container to draw after all other entries have been drawn.
			static const int	CONNECT_LEFT = (1<<1);
				// No pad on the left side
			static const int	CONNECT_RIGHT = (1<<2);
				// No pad on the right side

		public:
			virtual ~Entry();

			virtual int			flags() const			{ return 0; }
			virtual QSize		size() const = 0;
			// Quick hack for arc browsing experiment
			virtual bool		getParam(Param&) const { return false; }
			virtual void		onPaintBg(paint_args&);
			virtual void		onPaint(paint_args&) = 0;
			virtual void		mousePressEvent(const mouse_args& down) = 0;
			virtual void		mouseMoveEvent(const mouse_args& down) = 0;
			// For the release, the mEntry is set to NULL if the user releases
			// on a different entry then they pressed down on.  But maybe I should
			// change that, and have clients just check to see if the down
			// entry is the same as the current.
			virtual void		mouseReleaseEvent(const mouse_args& down) = 0;
			// Ugly, but clients need to respond with either NO_PARAMS to PARAM.
			virtual int			onParamChanged(const Param&) { return NO_PARAMS; }

		protected:
			Entry();
		};

	/* CREATE.  Params passed in on object creation.  This object
	 * is transitory, but anything it contains will exist for the
	 * life of the UI.
	 **************************************************************/
	public:
		struct Create {
			const Environment&			mEnv;
			Context&					mPrj;
			const ArcList&				mArcs;
			const ColorSrcList&			mClrs;
			int							mMode;

			Create(const Environment&, Context&, const ArcList&,
				   const ColorSrcList&, const int mode);

		private:
			Create(const Create&);
			void		operator=(const Create&);
		};

/* FX-UI API
 ******************************************************************/
public:
    FxUi();
	virtual ~FxUi();

	virtual void			setParent(Parent*, const Look&);

	// Subclasses answer with a code indicating what needs to
	// happen after a project change, either there's some kind
	// of error (most likely, the source has been removed), they
	// are unaffected, or they need a recache.
	static const int		UI_ERROR = -1;
	static const int		UI_UNCHANGED = 0;
	static const int		UI_CHANGED = 1;
	virtual int				onProjectChanged() = 0;

protected:
	Parent*					mParent;
};

#endif // FXUI_H
