#include "bevel.h"

#include <QtDebug>
#include <QList>
#include <QPainter>
#include <QPolygonF>
#include <QTransform>

#include <math.h>
#include "io/svgreader.h"
#include "math/geo.h"
#include "math/report.h"
#include "param/param.h"
#include "fx/light.h"

#include "experimental/exp_defs.h"

using namespace fx;

static const bool			WIREFRAME = true;

static const QString		FX_NAME("bevel");
static const QString		AMOUNT_NAME("amount");
static const int			AMOUNT_ID = 0;
static const QString		LIGHT_ANGLE_NAME("light");
static const int			LIGHT_ANGLE_ID = 20;
// Should be part of light
static const int			CLR_ID = 100;		// XXX Shared with light.cpp
static const int			ARC_ID = 101;		// XXX Shared with light.cpp
// For testing...
static const QString		WIRE_NAME("Wire");
static const int			WIRE_ID = 21;

// Hardcoded limit on depth amount right now
static const float			DEPTH_MAX = 200;

// FLAGS
static const int			CLR_EXP = (1<<0);
static const int			ARC_EXP = (1<<1);

static void get_circle_edge(	const Point3& cen, const float angle, const float dist,
								Point3& out);

/* FX::BEVEL
 ******************************************************************/
Bevel::Bevel()
	: Fx(-1)
//	, mDepth(124)
//	, mDepth(50)
	, mDepth(136)
	, mLightAngle(0.35f)
	, mWire(0)
	, mLight()
{
	mLight.mClr.setId(CLR_ID);
	mLight.mArc.setId(ARC_ID);
	mCn.setDepth(mDepth);
}

Bevel::Bevel(const Bevel& o)
	: Fx(o)
	, mDepth(o.mDepth)
	, mLightAngle(o.mLightAngle)
	, mWire(o.mWire)
	, mLight(o.mLight)
{
	mLight.mClr.setId(CLR_ID);
	mLight.mArc.setId(ARC_ID);
	mCn.setDepth(mDepth);
}

Bevel::~Bevel()
{
}

Bevel* Bevel::copy() const
{
	return new Bevel(*this);
}

void Bevel::init(fx::Setup& setup)
{
	Fx::init(setup);
	mLight.init(setup);
}

QString Bevel::getLabel() const
{
	return FX_NAME;
}

void Bevel::getParams(param::List& list, param::GetArgs& get) const
{
	list.addParam(&FX_NAME, AMOUNT_NAME, AMOUNT_ID, mDepth / DEPTH_MAX, get);
	list.addParam(&FX_NAME, LIGHT_ANGLE_NAME, LIGHT_ANGLE_ID, mLightAngle, get);
	// Wire doesn't work under the new rendering engine, and really I'm not
	// sure what the best way to support it again would be, so off for now.
//	list.addParam(&FX_NAME, WIRE_NAME, WIRE_ID, mWire, get);

	if (get.supports(param::COLOR_FEATURE)) {
		param::List::Grouper	g(list, COLOR_SZ);
		mLight.mClr.addTypeParam(list, get, CLR_EXP, mFlags);
	}
	if (get.supports(param::ARC_FEATURE)) {
		param::List::Grouper	g(list, ARC_SZ);
		mLight.mArc.addTypeParam(list, get, ARC_EXP, mFlags);
	}
}

param::Parent* Bevel::findParent(param::UrlIterator& it)
{
	// If I hit a composite object, there are two possibilities:
	// If the URL is a leaf, then someone is trying to replace
	// the entire composite.  Otherwise they're looking to set a param.
	if (it.matches(CLR_ID)) {
		if (it.isLeaf()) return this;
		++it;
		return mLight.mClr.findParent(it);
	} else if (it.matches(ARC_ID)) {
		if (it.isLeaf()) return this;
		++it;
		return mLight.mArc.findParent(it);
	}
	return this;
}

void Bevel::setParam(const param::UrlIterator& it, param::SetArgs& sa)
{
	if (it.matches(AMOUNT_ID)) {
		float		trans = sa.mV * DEPTH_MAX;
		if (trans < 0) trans = 0;
		else if (trans > DEPTH_MAX) trans = DEPTH_MAX;
		if (mDepth == trans) return;		// no change
		mDepth = trans;
		mCn.setDepth(mDepth);
	} else if (it.matches(LIGHT_ANGLE_ID)) {
		if (mLightAngle == sa.mV) return;	// no change
		mLightAngle = sa.mV;
	} else if (it.matches(WIRE_ID)) {
		if (mWire == sa.mV) return;			// no change
		mWire = sa.mV;
	// If I reached here, then I am a direct set.  Any parameters
	// inside the arc would go directly there
	} else if (it.matches(CLR_ID)) {
		mLight.mClr.swap(sa.mClr);
	} else if (it.matches(ARC_ID)) {
		mLight.mArc.setContentTo(sa.mArc);
	} else {
		return;
	}
	sa.mChanged = true;
}

void Bevel::work(fx::WorkArgs& arg, fx::WorkOutput& out)
{
//	qDebug() << "BEVEL DEPTH=" << mDepth << " input size=" << f.mShape.size();
	if (arg.viewShape().size() < 1) return;

	// Cache the normal info.
//	mCn.setDepth(mDepth);
	arg.cache(mCn);

	// Setup the light -- it's intrinsically linked to the bevel and
	// probably shouldn't be a separate FX.
	const QSize			imgSize = out.mImg.size();
	// Set the global light location based on my angle and size.
	const Point3		cen(float(imgSize.width())/2, float(imgSize.height())/2, 0);
	Point3				lightLoc;
	get_circle_edge(cen, mLightAngle, (cen.x+cen.y)*1.5f, lightLoc);
	// get the min and max for normalizing
	float				lightMin = 0, lightMax = 1;
	mCn.getLightRange(lightLoc, lightMin, lightMax);
	mLight.setLight(mDepth, lightLoc, lightMin, lightMax);

	mLight.work(arg, out);
}

/* IO
 ******************************************************************/
void Bevel::writeJson(json::Writer& w) const
{
	writeJsonHeader(w, FX_NAME);

	w.key(AMOUNT_NAME);
	w.floatValue(mDepth);
	w.key(LIGHT_ANGLE_NAME);
	w.floatValue(mLightAngle);

	mLight.mClr.writeJsonObject(w, COLOR_SZ);
	mLight.mArc.writeJsonObject(w, ARC_SZ);

	writeJsonFooter(w);
}

void Bevel::startObject(json::Stack& s, const QString& k)
{
	if (k == COLOR_SZ) s.push(&(mLight.mClr));
	else if (k == ARC_SZ) s.push(&(mLight.mArc));
}

void Bevel::jsonFloat(json::Stack& s, const QString& k, const float v)
{
	if (k == AMOUNT_NAME) mDepth = v;
	else if (k == LIGHT_ANGLE_NAME) mLightAngle = v;
	else Fx::jsonFloat(s, k, v);
}

/* FX::BEVEL-FACTORY
 ******************************************************************/
BevelFactory::BevelFactory()
{
}

QString BevelFactory::label() const
{
	return FX_NAME;
}

QString BevelFactory::key() const
{
	return FX_NAME;
}

float BevelFactory::match(const QString& key) const
{
	if (key == FX_NAME) return 1;
	return 0;
}

Fx* BevelFactory::create() const
{
	return new Bevel();
}

/* misc
 ******************************************************************/
static void get_circle_edge(	const Point3& cen, const float angle, const float dist,
								Point3& out)
{
	float				degree = angle * 360;
	if (degree < 0) degree += 360;
	else if (degree > 1) degree -= 360;
	degree *= geo::DTOR;
	out.x = (cen.x + (cos(degree) * dist));
	out.y = (cen.y - (sin(degree) * dist));
	// Vary this for some nice effects
	out.z = 0;
}
