#include "bevel_cellnormal.h"

#include <QAtomicInt>
#include <QDateTime>
#include "bevel_clip.h"
#include "bevel_extrude.h"
#include "bevel_source.h"
#include "bevel_surface.h"

using namespace bevel;

static QAtomicInt		CREATED(0);

static const Surface* findSurface(	const Point3&, const SurfaceList& sl,
									const Surface* prv, const float d);

/* BEVEL::CELL-NORMAL
 ******************************************************************/
CellNormal::CellNormal()
	: mShapeStamp(0)
	, mNeedsCache(true)
{
	mCreated = CREATED.fetchAndAddRelaxed(1);
}

void CellNormal::setDepth(const float depth)
{
	mDepth = depth;
	mNeedsCache = true;
}

void CellNormal::cache(Cell* cell, const QSize& size, fx::WorkArgs& wa)
{
	if (mShapeStamp == wa.shapeStamp() && !mNeedsCache && wa.mNormalCached == mCreated) return;

	mSources.clear();
	init(cell, size);
	if (mDepth < 1) return;
	if (!mShape.set(wa.viewShape())) return;

//	if (lowQuality(cell, size)) {
	if (highQuality(cell, size)) {
		mShapeStamp = wa.shapeStamp();
		mNeedsCache = false;
		wa.mNormalCached = mCreated;
	}
}

void CellNormal::getLightRange(const Point3& loc, float& min, float& max) const
{
	const float				locLength = loc.length();
	min = 1;
	max = -1;
	SourceListIterator		slit(mSources);
	while (slit.hasValue()) {
		const Source&		s = **slit;
		++slit;

		const float			v = bevel::lightFrom(s.mNormal, loc, locLength);
		if (v > max) max = v;
		if (v < min) min = v;
	}
}

void CellNormal::init(Cell* cell, const QSize& size) const
{
	const int			w = size.width(), h = size.height();
	int					pix = 0;
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			cell[pix].mNormalX = 0;
			cell[pix].mNormalY = 0;
			cell[pix].mNormalZ = 0;
			pix++;
		}
	}
}

/* LOW QUALITY
 ******************************************************************/
bool CellNormal::lowQuality(Cell* cell, const QSize& size)
{
	Extrude			extrude;
	if (!extrude.on(mShape, mDepth, mSources)) return false;

	SourceListIterator		slit(mSources);
	math::Normal			nrm;
	while (slit.hasValue()) {
		Source&				s = **slit;
		++slit;

		nrm.setOn(s.mPoly, s.mNormal);
		s.mPoly.getFrame(s.mFrame);
	}

	setPixelNormals(cell, size, mSources);
	return true;
}

void CellNormal::setPixelNormals(Cell* cell, const QSize& size, const SourceList& sl)
{
	const int			w = size.width(), h = size.height();
	if (w < 1 || h < 1) return;

	SourceNormal		sn;
	sn.setMaxDistance(mDepth);
sn.mPrint = false;

	int					pix = 0;
	Point3				loc, nrm;
	for (int y=0; y<h; y++) {
		loc.y = (float)y;
		for (int x=0; x<w; x++) {
			loc.x = (float)x;
sn.mPrint = (x == 110 && y == 28);
			if (sn.getNormal(sl, loc, nrm)) {
				cell[pix].mNormalX = nrm.x;
				cell[pix].mNormalY = nrm.y;
				cell[pix].mNormalZ = nrm.z;
			}
			pix++;
		}
	}
}

/* HIGH QUALITY
 ******************************************************************/
bool CellNormal::highQuality(Cell* cell, const QSize& size)
{
	Extrude			extrude;
	if (!extrude.on(mShape, mDepth, mSources)) return false;

	SurfaceList		surfaces;
	Clip			clip;
	clip.setExtension(Point3(0, 0, mDepth*10), mDepth);

	SourceListIterator		slit(mSources);
	math::Normal			nrm;
int		c = -1;
	while (slit.hasValue()) {
c++;
		Source&				s = **slit;
		++slit;
// qDebug() << "SOURCE=" << c;
//		if (c < 1 || c > 2) continue;
//		if (c != 9) continue;
clip.mPrint = false;
if (clip.mPrint) qDebug() << "CLIP source " << c;
		clip.on(mShape, s, mSources, surfaces);
		nrm.setOn(s.mPoly, s.mNormal);
	}

	buildEdges(surfaces);
	setPixelNormals(cell, size, surfaces);
	return true;
}

void CellNormal::buildEdges(SurfaceList& allA)
{
	SurfaceListIterator		slit(allA);
int		c = -1;
	while (slit.hasValue()) {
c++;
		Surface&			s = **slit;
		++slit;
//		if (c < 1 || c > 2) continue;

//qDebug() << "\nCellNormal::build for " << c << " nrm " << s->mSource->mNormal;
//qDebug() << "poly size=" << s->mPoly.getSize() << " " << s->mPoly;
		EdgeBuilder		b(mEbState, s);
		b.add(allA);
	}
}

void CellNormal::setPixelNormals(Cell* cell, const QSize& size, const SurfaceList& sl)
{
	const int			w = size.width(), h = size.height();
	if (w < 1 || h < 1) return;
	const Surface**		row = mRow.get(w);
	if (row == NULL) return;
	for (int x=0; x<w; x++) row[x] = NULL;

	mSrfNrm.setMaxDistance(mDepth);
mSrfNrm.mPrint = false;

	int					pix = 0;
	Point3				loc0, loc1, nrm;
	for (int y=0; y<h; y++) {
		loc0.y = (float)y;
		for (int x=0; x<w; x++) {
			loc0.x = (float)x;
			row[x] = findSurface(loc0, sl, row[x], cell[pix].mD);
			if (row[x] != NULL) {
				// Find the z value of the point on the surface plane.
				loc1.set(loc0);
				loc0.z = -10000;
				loc1.z = 10000;
				mSP.on(loc0, loc1, row[x]->mSource->mPlane, loc0);
//mSrfNrm.mPrint = ((x >= 209 && x <= 211) && y == 46);		// edge
//mSrfNrm.mPrint = ((x >= 144 && x <= 145) && y == 140);
//mSrfNrm.mPrint = (x == 205 || x == 210) && y == 48;
//mSrfNrm.mPrint = (x == 190 && (y == 26 || y == 27));
//mSrfNrm.mPrint = (x == 90 && y == 166);	// on line
//mSrfNrm.mPrint = (x == 128 && y == 96);
				mSrfNrm.getNormal(*(row[x]), loc0, nrm);
				cell[pix].mNormalX = nrm.x;
				cell[pix].mNormalY = nrm.y;
				cell[pix].mNormalZ = nrm.z;
			}
			pix++;
		}
	}
}

/* TEST QUALITY
 ******************************************************************/
bool CellNormal::testQuality(Cell* cell, const QSize& size)
{
	Extrude			extrude;
	if (!extrude.on(mShape, mDepth, mSources)) return false;

	SourceListIterator		slit(mSources);
	math::Normal			nrm;
	while (slit.hasValue()) {
		Source&				s = **slit;
		++slit;

		nrm.setOn(s.mPoly, s.mNormal);
		s.mPoly.getFrame(s.mFrame);
	}

	setPixelNormals(cell, size, mSources);
	return true;
}

/* misc
 ******************************************************************/
static const Surface* findSurface(	const Point3& pt, const SurfaceList& sl,
									const Surface* prv, const float d)
{
	if (prv != NULL && prv->contains(pt)) return prv;
	const Surface*			ans = sl.find(pt);
//	if (ans == NULL && d >= 0) qDebug() << "d=" << d;
	return ans;
}
