#include "bevel_edge.h"

#include "math/geo.h"

using namespace bevel;

/* BEVEL::EDGE
 ******************************************************************/
EdgePoint::EdgePoint()
{
	clear();
}

void EdgePoint::clear()
{
	mIsSegment = false;
	mType = NO_TYPE;
	mCount = 0;
}

void EdgePoint::set(const EdgePoint& o)
{
	if (this != &o) {
		mLoc.set(o.mLoc);
		mNrm.set(o.mNrm);
		mD = o.mD;
		mIsSegment = o.mIsSegment;
		mType = o.mType;
		mCount = o.mCount;
	}
}

void EdgePoint::start(	const Point3& loc, const Point3& nrm,
						const Point3& origin, const int type,
						const bool isSegment)
{
	mLoc.set(loc);
	mNrm.set(nrm);
//	mD = origin.dist3Squared(loc);
	mD = origin.dist3(loc);
	mIsSegment = isSegment;
	mType = type;
	mCount = 1;
}

void EdgePoint::add(const EdgePoint& o)
{
	mNrm.asAdd(mNrm, o.mNrm);
	// Preserve the "edgeness"
	if (o.mIsSegment) mIsSegment = true;
	mCount++;
}

/* BEVEL::EDGE-POINT-LIST
 ******************************************************************/
class EdgePointList::Factory : public ManagedFactory<EdgePoint*>
{
public:
	Factory() : ManagedFactory<EdgePoint*>(true) { }

	virtual EdgePoint*	newObject() const {
		EdgePoint*		ans = new EdgePoint();
		if (ans == NULL) throw std::exception();
		return ans;
	}

	virtual void		deleteObject(EdgePoint* o) const { delete o; }
	virtual void		initializeObject(EdgePoint* o) const { o->clear(); }
	virtual void		setObject(EdgePoint* dst, EdgePoint* src) const { if (src != NULL) dst->set(*src); }
	virtual int			sortObject(EdgePoint* o1, EdgePoint* o2) const {
		// Need some slop, to make sure my endpoints are at the start of similar items
		if (o1->mLoc.sameAs3(o2->mLoc, 0.00001)) {
			return (o1->mType < o2->mType ? -1 : 1);
		}
		return geo::sort_doubles(o1->mD, o2->mD);
	}
};

static EdgePointList::Factory	EDGE_POINT_FACTORY;

EdgePointList::EdgePointList()
	: inherited(EDGE_POINT_FACTORY)
{
}

void EdgePointList::add(EdgePoint& pt)
{
	// All new points initialized to a count of 1
	pt.mCount = 1;
//qDebug() << "PUSH EDGEPT " << pt.mLoc << " d=" << pt.mD << " nrm=" << pt.mNrm << " type=" << pt.mType;
	pushSorted((EdgePoint*)&pt);
}

/* BEVEL::EDGE
 ******************************************************************/
Edge::Edge()
{
	clear();
}

void Edge::clear()
{
	mSource = NULL;
	mPts.clear();
}

void Edge::set(const Point3& a, const Point3& b, const Source& src)
{
	mA.set(a);
	mB.set(b);
	mSource = &src;
}

/* BEVEL::EDGE-LIST
 ******************************************************************/
class EdgeList::Factory : public ManagedFactory<Edge*>
{
public:
	virtual Edge*		newObject() const {
		Edge*			ans = new Edge();
		if (ans == NULL) throw std::exception();
		return ans;
	}

	virtual void		deleteObject(Edge* o) const { delete o; }
	virtual void		initializeObject(Edge* o) const { o->clear(); }
	virtual void		setObject(Edge*, Edge*) const { throw std::exception(); }
};
static EdgeList::Factory	EDGE_LIST_FACTORY;

EdgeList::EdgeList()
	: inherited(EDGE_LIST_FACTORY)
{
}

Edge* EdgeList::start(const Point3& a, const Point3& b, const Source& src)
{
	Edge*			e = pushFront();
	if (e != NULL) {
		e->mA.set(a);
		e->mB.set(b);
		e->mSource = &src;
		e->mPts.clear();
	}
	return e;
}

void EdgeList::add(const Point3& a, const Point3& b, const Source& src)
{
	Edge*			e = pushFront();
	if (e != NULL) e->set(a, b, src);
}
