#include "bevel_edgebuilder.h"

#include "math/geo.h"
#include "fx/bevel_source.h"
#include "fx/bevel_surface.h"

using namespace bevel;

static const double			EPS = 0.2;

static bool					PRINT = false;

/* EDGE-BUILDER::STATE
 ******************************************************************/
EdgeBuilder::State::State()
{
	mSI.mEps = EPS;
}

/* EDGE-BUILDER::COMBINE-BUILD
 * Build rules for removing duplicate edge points.  The build rules
 * mean to combine only those points marked BUILD_TYPE.
 ******************************************************************/
class EdgeBuilder::CombineBuild : public ManagedCombiner<EdgePoint*>
{
public:
	CombineBuild() : mType(EdgePoint::BUILD_TYPE) { }

	const int		mType;

	virtual bool	combine(EdgePoint* dst, EdgePoint* src) {
		if (dst == NULL || src == NULL) return false;
		if (mType == 0 || ( (dst->mType&mType) != 0 && (src->mType&mType) != 0)) {
			if (geo::abs(dst->mD - src->mD) < 1) {
				dst->add(*src);
				return true;
			}
		}
		return false;
	}

	void			finish(EdgePointList& l) {
		// Even if no one was combined, we still need to clear out the
		// build flag.
		EdgePointIterator		it(l);
		while (it.hasValue()) {
			EdgePoint&			ep = **it;
			++it;
			if (ep.mCount > 1) ep.mNrm /= (double)ep.mCount;
			ep.mCount = 0;
			if (ep.mType == EdgePoint::BUILD_TYPE) ep.mType = EdgePoint::NO_TYPE;
		}
	}
};

/* EDGE-BUILDER::COMBINE-FINISH
 * Finish rules for removing duplicate edge points.  The finish
 * combines all duplicate points, overwriting the endpoints.
 ******************************************************************/
class EdgeBuilder::CombineFinish : public ManagedCombiner<EdgePoint*>
{
public:
	CombineFinish()  { }

	virtual bool	combine(EdgePoint* dst, EdgePoint* src) {
		if (dst == NULL || src == NULL) return false;
		if (geo::abs(dst->mD - src->mD) >= 1) return false;

		// Rules: endpoints are overwritten.  All others accumulate.
		if (src->mType == EdgePoint::ENDPOINT_TYPE) return true;
		if (dst->mType == EdgePoint::ENDPOINT_TYPE) dst->set(*src);
		else dst->add(*src);
		return true;
	}

	void			finish(EdgePointList& l, const Point3& nrm) {
		// Even if no one was combined, we still need to clear out the
		// build flag.
		EdgePointIterator		it(l);
		while (it.hasValue()) {
			EdgePoint&			ep = **it;
			++it;
			ep.mNrm.asAdd(ep.mNrm, nrm);
			ep.mCount++;

			if (ep.mCount > 1) ep.mNrm /= (double)ep.mCount;
		}
	}
};

/* EDGE-BUILDER
 ******************************************************************/
EdgeBuilder::EdgeBuilder(State& state, Surface& a)
	: mState(state)
	, mA(a)
{
//qDebug() << "\nSTART BUILDING ON " << mA.mIdx << " " << mA.mPoly;
PRINT = (mA.mIdx == 0 || mA.mIdx == 1);
PRINT = false;
if (PRINT) qDebug() << "\nSTART BUILDING ON " << mA.mIdx << " " << mA.mPoly;
	mA.mEdges.clear();
	start();
}

EdgeBuilder::~EdgeBuilder()
{
	finish();

if (PRINT) qDebug() << "FINAL EDGES";

	EdgeListIterator			itA(mA.mEdges);
	while (itA.hasValue()) {
		Edge&					edge = **itA;
		++itA;

if (PRINT) qDebug() << "edge a " << edge.mA << " b " << edge.mB;

		EdgePointIterator		lit(edge.mPts);
		while (lit.hasValue()) {
			const EdgePoint&	pt = **lit;
			++lit;

if (PRINT) qDebug() << "\tpt loc=" << pt.mLoc << " d=" << pt.mD << " nrm=" << pt.mNrm;
		}
	}
}

// Yeah... Qt rect.intersects() doesn't include rects with coincident edges.
static bool intersects(const QRectF& a, const QRectF& b)
{
	if (a.right() < b.left()) return false;
	if (a.left() > b.right()) return false;
	if (a.bottom() < b.top()) return false;
	if (a.top() > b.bottom()) return false;
	return true;
}

void EdgeBuilder::add(const SurfaceList& allB)
{
//if (!(mA.mIdx >= 9 && mA.mIdx <= 10)) return;
//if (mA.mIdx != 11) return;

	SurfaceListIterator		slit(allB);
	while (slit.hasValue()) {
		const Surface*		s = *slit;
//		if (s != NULL && s->mSource != mA.mSource && mA.mFrame.intersects(s->mFrame)) {
		if (s != NULL && s->mSource != mA.mSource && intersects(mA.mFrame, s->mFrame)) {
			addEdges(*s);
		}
		++slit;
	}
}

void EdgeBuilder::addEdges(const Surface& b)
{
	const Point3&				nrmB = b.mSource->mNormal;
	Point3&						a0 = mState.mPtA0;
	Point3&						a1 = mState.mPtA1;
	EdgePoint&					ep = mState.mEdgePt;
	CombineBuild				cmb;
	EdgeListIterator			itA(mA.mEdges);
if (PRINT) qDebug() << "\nADD EDGES for " << b.mPoly;
	while (itA.hasValue()) {
		Edge&					edge = **itA;
		math::EdgeIterator		itB(b.mPoly);
		bool					added = false;
if (PRINT) qDebug() << "\npush edge " << edge.mA << ", " << edge.mB;
		while (itB.hasValue()) {
if (PRINT) qDebug() << "\tB " << *(itB.mA) << ", " << *(itB.mB);
			// Find an overlapping segment between a and b.
			if (mState.mSI.on(edge.mA, edge.mB, *(itB.mA), *(itB.mB), a0, a1) >= math::SegIntersection::INTERSECTS) {
				added = true;
				const bool		isSegment = (!a0.sameAs3(a1, EPS));
				ep.start(a0, nrmB, edge.mA, EdgePoint::BUILD_TYPE, isSegment);
				edge.mPts.add(ep);
if (PRINT) qDebug() << "\t\tadd pt " << a0;
				// small optimization, the duplicate would get scrubbed later anyway.
				if (isSegment) {
if (PRINT) qDebug() << "\t\tadd pt " << a1;
					ep.start(a1, nrmB, edge.mA, EdgePoint::BUILD_TYPE, false);
					edge.mPts.add(ep);
				}
			}

			++itB;
		}
if (PRINT) qDebug() << "\tedge point count 1=" << edge.mPts.size();
		// Due to the way I'm adding points, I might end up with a lot
		// of duplicate hits, so clear those out.
		if (added) {
			edge.mPts.removeDuplicates(cmb);
			cmb.finish(edge.mPts);
		}
		++itA;
if (PRINT) qDebug() << "\tedge point count 2=" << edge.mPts.size();
	}
}

void EdgeBuilder::start()
{
	mA.mEdges.clear();
	if (mA.mSource == NULL) return;

	EdgePoint&				ep = mState.mEdgePt;
	math::EdgeIterator		itA(mA.mPoly);
	while (itA.hasValue()) {
		const Point3		&a = *(itA.mA), &b = *(itA.mB);
		Edge*				e = mA.mEdges.start(a, b, *(mA.mSource));
		// Initialize the edge with the segment endpoints
		if (e != NULL) {
//qDebug() << "\nSTART EDGE " << a << ", " << b;
			ep.start(a, mA.mSource->mNormal, a, EdgePoint::ENDPOINT_TYPE, false);
			e->mPts.add(ep);
			ep.start(b, mA.mSource->mNormal, a, EdgePoint::ENDPOINT_TYPE, false);
			e->mPts.add(ep);
		}
		++itA;
	}
}

void EdgeBuilder::finish()
{
if (PRINT) qDebug() << "BuildEdges for " << mA.mIdx << " nrm " << mA.mSource->mNormal << " " << mA.mPoly;
	// Create the normals for each point by averaging any overlaps.
	const Point3&				nrm = mA.mSource->mNormal;
	CombineFinish				cmb;
	EdgeListIterator			itA(mA.mEdges);
	while (itA.hasValue()) {
		Edge&					edge = **itA;
		++itA;

if (PRINT) {
	qDebug() << "\tedge " << edge.mA << ", " << edge.mB;
	EdgePointIterator		itB(edge.mPts);
	while (itB.hasValue()) {
		EdgePoint&			ep = **itB;
		qDebug() << "\t\tpt=" << ep.mLoc << " nrm=" << ep.mNrm;

		++itB;
	}
}
		// Average any overlaps, but with a special rule:  Endpoints are
		// only placeholders, and will get overwritten.
		edge.mPts.removeDuplicates(cmb);
		cmb.finish(edge.mPts, nrm);

if (PRINT) {
	EdgePointIterator		itB(edge.mPts);
	while (itB.hasValue()) {
		EdgePoint&			ep = **itB;
		qDebug() << "\t\tremoved=" << ep.mLoc << " nrm=" << ep.mNrm;
		++itB;
	}
}

		EdgePointIterator		itB(edge.mPts);
		while (itB.hasValue()) {
			EdgePoint&			ep = **itB;
if (PRINT) qDebug() << "\tpt=" << ep.mLoc << " d=" << ep.mD << " nrm=" << ep.mNrm << " type=" << ep.mType;

			++itB;
		}
	}
}
