#include "bevel_extrude.h"

#include <math.h>

#include "fx/bevel_source.h"
#include "math/geo.h"

using namespace bevel;

static const double		EPS = 0.1;

static float directioned_dist(	const float ax, const float ay,
								const float bx, const float by,
								const float dist);

/* BEVEL::EXTRUDE
 ******************************************************************/
Extrude::Extrude()
{
}

bool Extrude::on(const Polygon& src, const float depth, SourceList& out)
{
	const qint32	size = src.getSize();
	const Point3*	pts = src.getPoints();
	if (size < 3) return false;
	const bool		isClosed = pts[0].sameAs2(pts[size-1], EPS);

	// Reverse the order of iteration for opposite effects
	const int		min = 0, max = size-1 + (isClosed ? -1 : 0);
	int				i = 0, stop = max + 1, inc = 1;
	if (stop <= i) return false;
	// If the points are ordered clockwise, iterate backwards.
	if (src.isClockWise(EPS)) {
		i = max;
		stop = -1;
		inc = -1;
	}
	int				pi = i - inc,
					ni = i + inc;
	Point3			p1, p3, prvP3;
	const Point3	*prv, *cur, *nxt;
	const int		start = i;

	while (i != stop) {
		// Constrain the indexes to valid values
		if (pi > max) pi = 0;
		else if (pi < min) pi = max;
		if (ni > max) ni = 0;
		else if (ni < min) ni = max;

		prv = &(pts[pi]);
		cur = &(pts[i]);
		nxt = &(pts[ni]);

		// Every 2 points are extruded into a new quad, by generating
		// a new point for each.
		extrudePoints(*cur, *nxt, depth, p1, p3);
		p1.z = depth;
		p3.z = depth;

		if (i != start) fillGaps(*prv, *cur, *nxt, prvP3, p1, depth, out);

		out.add(cur, &p1, &p3, nxt, mOrder);

		prvP3.set(p3);

		// Increment
		pi = i;
		i += inc;
		ni += inc;
	}
	return true;
}

// Given two ordered points (b, c) and a depth, find
// a new point representing the bevel at the given depth.
void Extrude::extrudePoints(	const Point3& b, const Point3& c, const float depth,
								Point3& out1, Point3& out2) const
{
	const float			m_bc = geo::perpendicularSlope(b.x, b.y, c.x, c.y);

	// Use the segment direction to decide which direction
	// to extend the new point in.
	const float			dist_bc = directioned_dist(b.x, b.y, c.x, c.y, depth);

	// Get two new line segments, perpendicular to the supplied points.
	geo::newPtOnLine(b.x, b.y, m_bc, dist_bc, out1);
	geo::newPtOnLine(c.x, c.y, m_bc, dist_bc, out2);
}

void Extrude::fillGaps(	const Point3& prv, const Point3& cur, const Point3& nxt,
						const Point3& prvP3, const Point3& p1,
						float depth,
						PolygonList& out) {
	// Using my current point as the center, find the degree change
	// between the previous and next points.  If it's too high (since
	// we evaluate points clockwise, we reverse the degree evaluation),
	// then the extruded polygons overlap.
	const float			prvDeg = geo::degree(prv.x-cur.x, prv.y-cur.y),
						nxtDeg = geo::degree(nxt.x-cur.x, nxt.y-cur.y);
	float				check = nxtDeg - prvDeg;
	if (check < 0) check += 360;
	if (check >= 180) return;

	// Get the midpoint, use its slope to determine the line on
	// which the gap point resides.
	mMid.set(prvP3);
	mMid += p1;
	mMid /= 2;
	const float			m = geo::slope(cur.x, cur.y, mMid.x, mMid.y);
	// Brute-force it:  Make two new gap points, going in opposite directions.
	// The one closest to the extruded points is the winner.
	geo::newPtOnLine(cur.x, cur.y, m, depth, mGap0);
	geo::newPtOnLine(cur.x, cur.y, m, -depth, mGap1);
	mGap0.z = depth;
	mGap1.z = depth;
	const Point3*		gap = &mGap0;
	if (prvP3.dist3Squared(mGap1) < prvP3.dist3Squared(mGap0)) gap = &mGap1;
	add(&cur, &prvP3, gap, NULL, out);
	add(&cur, gap, &p1, NULL, out);
}

void Extrude::fillGaps(	const Point3& prv, const Point3& cur, const Point3& nxt,
						const Point3& prvP3, const Point3& p1,
						float depth, SourceList& out) {
	// Using my current point as the center, find the degree change
	// between the previous and next points.  If it's too high (since
	// we evaluate points clockwise, we reverse the degree evaluation),
	// then the extruded polygons overlap.
	const float			prvDeg = geo::degree(prv.x-cur.x, prv.y-cur.y),
						nxtDeg = geo::degree(nxt.x-cur.x, nxt.y-cur.y);
	float				check = nxtDeg - prvDeg;
	if (check < 0) check += 360;
	if (check >= 180) return;

	// Get the midpoint, use its slope to determine the line on
	// which the gap point resides.
	mMid.set(prvP3);
	mMid += p1;
	mMid /= 2;
	const float			m = geo::slope(cur.x, cur.y, mMid.x, mMid.y);
	// Brute-force it:  Make two new gap points, going in opposite directions.
	// The one closest to the extruded points is the winner.
	geo::newPtOnLine(cur.x, cur.y, m, depth, mGap0);
	geo::newPtOnLine(cur.x, cur.y, m, -depth, mGap1);
	mGap0.z = depth;
	mGap1.z = depth;
	const Point3*		gap = &mGap0;
	if (prvP3.dist3Squared(mGap1) < prvP3.dist3Squared(mGap0)) gap = &mGap1;
	out.add(&cur, &prvP3, gap, NULL, mOrder);
	out.add(&cur, gap, &p1, NULL, mOrder);
}

void Extrude::add(	const Point3* a, const Point3* b, const Point3* c, const Point3* d,
					PolygonList& out)
{
	Polygon*		p = out.pushBack();
	if (p == NULL) return;
	p->clear();
	if (a != NULL) p->add(*a);
	if (b != NULL) p->add(*b);
	if (c != NULL) p->add(*c);
	if (d != NULL) p->add(*d);
	mOrder.makeCcw(p->getPoints(), p->getSize());
}

/* misc
 ******************************************************************/
static float directioned_dist(	const float ax, const float ay,
								const float bx, const float by,
								const float dist)
{
	double				deg = geo::degree(bx-ax, by-ay);
	// My getDegree func doesn't correspond exactly to the
	// angles that determine direction.
	deg -= 90;
	if (deg < 0) deg += 360;
	if (deg >= 0 && deg < 180) return -dist;
	return dist;
}
