#include "bevel_surface.h"

#include "math/primitives/range.h"
#include "math/geo.h"
#include "fx/bevel_source.h"

//#define BEVEL_SMOOTHING_OFF		(1)

using namespace bevel;

static const double			EPS = 0.2;

/* BEVEL::SURFACE
 ******************************************************************/
Surface::Surface()
{
}

void Surface::clear()
{
	mSource = NULL;
	mEdges.clear();
	mPoly.clear();
}

void Surface::set(const Source& src, const Polygon& p)
{
	mSource = &src;
	mPoly.set(p);
	mPoly.getFrame(mFrame);
}

bool Surface::contains(const Point3& pt) const
{
	if (!mFrame.contains(pt.x, pt.y)) return false;
	return mPoly.contains(pt);
}

/* BEVEL::SURFACE-LIST
 ******************************************************************/
class SurfaceList::Factory : public ManagedFactory<Surface*>
{
public:
	virtual Surface*		newObject() const {
		Surface*			ans = new Surface();
		if (ans == NULL) throw std::exception();
		return ans;
	}

	virtual void		deleteObject(Surface* o) const {
		delete o;
	}

	virtual void		initializeObject(Surface* o) const {
		o->clear();
	}

	virtual void		setObject(Surface*, Surface*) const {
		// Should never be called
		throw std::exception();
	}
};
static SurfaceList::Factory		FACTORY;

SurfaceList::SurfaceList()
	: inherited(FACTORY)
{
}

void SurfaceList::add(const Source& src, const Polygon& p)
{
	Surface*			s = pushBack();
	if (s != NULL) {
		s->set(src, p);
		s->mIdx = size()-1;
	}
}

const Surface* SurfaceList::find(const Point3& pt) const
{
	SurfaceListIterator		slit(*this);
	while (slit.hasValue()) {
		const Surface*		s = *slit;
		if (s != NULL && s->contains(pt)) return s;
		++slit;
	}
	return NULL;
}

/* BEVEL::SURFACE-NORMAL
 ******************************************************************/
SurfaceNormal::SurfaceNormal()
	: mMaxD(10)
{
	// Hopefully prevents allocations.
	mFound.get(32);
}

void SurfaceNormal::setMaxDistance(const double d)
{
//qDebug() << "Set depth=" << d;
	mMaxD = d;
//mMaxD = 20;
}

void SurfaceNormal::getNormal(const Surface& s, const Point3& loc, Point3& nrm)
{
/*
if (!(s.mIdx == 0 || s.mIdx == 1)) {
	nrm.set(0, 0, 0);
	return;
}
*/
	nrm.set(s.mSource->mNormal);
//if (s.mIdx != 5) return;
//if (s.mIdx != 6) return;

#ifdef BEVEL_SMOOTHING_OFF
	return;
#endif
	if (mPrint) qDebug() << "\ngetNormal for " << loc << " maxD=" << mMaxD << " srf=" << s.mIdx;
//	if (mPrint) qDebug() << "\tstart nrm=" << nrm;

	// Collect all edges within range
	Found*						found = NULL;
	int							size = 0;
	EdgeListIterator			itA(s.mEdges);
	while (itA.hasValue()) {
		const Edge&				e = **itA;
		++itA;

if (mPrint) qDebug() << "\t" << size << " Edge a=" << e.mA << " b=" << e.mB;

		const double			d = mPS.on(loc, e.mA, e.mB, mPS.mHit);
		if (d <= 0) {
			getEdgeNormal(e, mPS.mHit, nrm);
			return;
		} else {
			if ((found = mFound.get(size+1)) == NULL) return;
			Found&				f = found[size];
			f.mD = d;
			getEdgeNormal(e, mPS.mHit, f.mNrm);
if (mPrint) qDebug() << "\t" << size << " d=" << f.mD << " nrm=" << f.mNrm;
			size++;
		}
	}
	if (size < 1) return;
	if (size == 1) {
		nrm.set(found[0].mNrm);
		return;
	}

	double					lowD;
	assignAmounts(found, size, lowD);

	// Accumulate the averages
	nrm.set(0, 0, 0);
	for (int k=0; k<size; k++) {
		Found&			f = found[k];
if (mPrint) qDebug() << "\t" << k << " amt=" << f.mAmt << " for " << f.mNrm;
		f.mNrm *= f.mAmt;
		nrm.asAdd(nrm, f.mNrm);
	}
if (mPrint) qDebug() << "FINAL=" << nrm;

	// Blend in the surface normal based on the lowest distance to edge.
//	const double		blend = (lowD >= mMaxD ? 1 : lowD / mMaxD);
//	nrm.asBlend(nrm, s.mSource->mNormal, blend);
}

void SurfaceNormal::assignAmounts(Found* found, const int size, double& lowD) const
{
	// Find the proportions between each found item.  Do this
	// by finding the absolute value of each item (distance / sum of all distances),
	// then finding the ratio between each and normalizing those values,
	// and finally assigning an amount based on the normalized ratios.
	// So let's trace through some numbers.  Let's say we have 3 pts, that
	// are a=6, b=30 and c=180 pixels away.
	// Find the low (6) and sum (216).
	// Find the ratios based on the low.  So the low numerator is 6/216= 0.027.
	// a=0.027/(6/216)=1, b=0.027/(30/216)=0.194, c=0.032
	// Find the sum of the ratios, 0.253
	// Finally, assign the amounts as the normalized ratios

	// distance / sum of all distances
	lowD = found[0].mD;
	int					lowIdx = 0;
	double				sumD = found[0].mD;
	for (int k=1; k<size; k++) {
		Found&			f = found[k];
		sumD += f.mD;
		if (f.mD < lowD) {
			lowD = f.mD;
			lowIdx = k;
		}
	}
	// Assign the low value to 1, express each non-low as a ratio of the low
	const double			lowNumerator = lowD / sumD;
	double					sum = 0;
	for (int k=0; k<size; k++) {
		Found&				f = found[k];
		const double		div = f.mD / sumD;
		f.mRatio = lowNumerator / div;
//		f.mRatio = lowNumerator / (GEO_S_CURVE(div));
		sum += f.mRatio;
	}

//if (mPrint) qDebug() << "SUM_D=" << sumD << " sum=" << sum;

	// Normalize
	for (int k=0; k<size; k++) {
		Found&				f = found[k];
//if (mPrint) qDebug() << "\t" << k << " NEW d=" << f.mD << " ratio=" << f.mRatio << " amt=" << (f.mRatio / sum);
		f.mAmt = (f.mRatio / sum);
	}
}

void SurfaceNormal::getEdgeNormal(const Edge& e, const Point3& pt, Point3& nrm)
{
	// Default, covers error conditions
//	nrm.set(e.mSource->mNormal);
//	const double			d = e.mA.dist3Squared(pt);
	// Add in a touch of undershoot so that I don't go out of bounds on a technicality
	const double			d = e.mA.dist3(pt) - 0.00001;
	EdgePointIterator		it(e.mPts);
	const EdgePoint*		prv = NULL;
//if (mPrint) qDebug() << "getEdgeNrm " << pt;
	while (it.hasValue()) {
		const EdgePoint&	nxt = **it;
		++it;
if (mPrint) qDebug() << "\t\tpt d=" << nxt.mD << " seg=" << nxt.mIsSegment << " cmpD=" << d;

		if (d <= nxt.mD) {
			if (prv == NULL) nrm.set(nxt.mNrm);
			else {
				if (!prv->mIsSegment) {
					// Blend based on distance
					const double	fullD = d-prv->mD;
if (mPrint) qDebug() << "\t\tfullD=" << fullD;
					if (fullD <= 0) nrm.set(prv->mNrm);
					else if (fullD >= mMaxD) nrm.set(e.mSource->mNormal);
					else nrm.asBlend(prv->mNrm, e.mSource->mNormal, fullD / mMaxD);
//					nrm.set(0, 0, 0);
				} else {
					const double		dA = d - prv->mD;
					nrm.asBlend(prv->mNrm, nxt.mNrm, dA / (nxt.mD - prv->mD));
//if (mPrint) qDebug() << "\tblend dA=" << dA << " pos=" << (dA / (nxt.mD - prv->mD));
//if (mPrint) qDebug() << "\tprv=" << prv->mNrm << " nxt=" << nxt.mNrm;
				}
			}
			return;
		}
		prv = &nxt;
	}

if (mPrint) qDebug() << "\t\toob";

	// out of bounds
	if (prv != NULL) nrm.set(prv->mNrm);
	// error
	else nrm.set(e.mSource->mNormal);
}
