#ifndef BEVEL_SURFACE_H
#define BEVEL_SURFACE_H

#include "collections/recyclearray.h"
#include "math/primitives/polygon.h"
#include "math/primitives/segment3.h"
#include "fx/bevel_edge.h"

namespace bevel {
class Source;

/* BEVEL::SURFACE
 * I contain all the necessary info to shade a surface.  This
 * means a source polygon, broken into a series of line segments,
 * which include a sorted list of points for all shading steps
 * along that line.  Additionally, I contain the center line
 * for my normal value.
 ******************************************************************/
class Surface
{
public:
	int					mIdx;		// debugging

public:
	Surface();

	// Each source will become one or more surfaces.  We want
	// to know the source because it has info like the normal,
	// and also lets as avoid redundant processing.
	const Source*		mSource;
	Polygon				mPoly;
	QRectF				mFrame;
	// A list of all edges that coincide with mine.
	EdgeList			mEdges;

	void				clear();
	void				set(const Source&, const Polygon&);
	bool				contains(const Point3&) const;
};

/* BEVEL::SURFACE-LIST
 ******************************************************************/
class SurfaceList : public ManagedList<Surface*>
{
public:
	SurfaceList();

	void				add(const Source&, const Polygon&);

	const Surface*		find(const Point3&) const;

private:
	typedef ManagedList<Surface*>	inherited;

// internal
public:
	class Factory;
};

/* BEVEL::SURFACE-NORMAL
 * A utility for finding the normal of a point in a surface.
 ******************************************************************/
class SurfaceNormal
{
public:
	bool				mPrint;		// tmp debug

public:
	SurfaceNormal();

	// Set the maximum distance to shade
	void				setMaxDistance(const double);
	void				getNormal(const Surface&, const Point3& loc, Point3& nrm);

private:
	/* FOUND
	 **************************************************************/
	struct Found {
		double			mD;			// distance from edge
		double			mAmt;		// amount I will contribute to final output
		double			mRatio;
		Point3			mNrm;
	};
	typedef RecycleArray<Found>	FoundBlock;

private:
	math::PtSeg			mPS;
	double				mMaxD;
	FoundBlock			mFound;

	void				assignAmounts(Found*, const int size, double& lowD) const;
	void				getEdgeNormal(const Edge&, const Point3& pt, Point3& nrm);
};

typedef ManagedListIterator<Surface*>	SurfaceListIterator;

} // namespace bevel

#endif // BEVEL_SURFACE_H
