#include "chain.h"

#include <QDebug>

using namespace fx;

static const QString		FX_NAME("chain");
static const QString		ALL_STR("all");

/* CHAIN
 ******************************************************************/
Chain::Chain()
	: Fx(-1)
{
}

Chain::Chain(const Chain& o)
	: Fx(o)
	, fx::Reader::Target(o)
{
	for (int k=0; k<o.mFx.size(); k++) {
		const Fx*		src = o.mFx[k];
		if (src == NULL) continue;
		Fx*				dst = src->copy();
		if (dst != NULL) mFx.append(dst);
	}
}

Chain::~Chain()
{
	for (int k=0; k<mFx.size(); k++) delete mFx[k];
	mFx.clear();
}

Chain* Chain::copy() const
{
	return new Chain(*this);
}

void Chain::init(fx::Setup& setup)
{
	Fx::init(setup);
	for (int k=0; k<mFx.size(); k++) {
		if (mFx[k] == NULL) continue;

		mFx[k]->init(setup);
	}
}

QString Chain::getLabel() const
{
	return FX_NAME;
}

void Chain::getParams(param::List& list, param::GetArgs& get) const
{
	for (int k=0; k<mFx.size(); k++) {
		if (mFx[k] == NULL) continue;

		const Fx&		fx = *(mFx[k]);
		param::List		l;
		fx.getParams(l, get);
		l.prepend(fx.getId());
		list.takeAll(l);
	}
}

param::Parent* Chain::findParent(param::UrlIterator& it)
{
	for (int k=0; k<mFx.size(); k++) {
		Fx*			fx = mFx[k];
		if (fx == NULL) continue;

		if (it.matches(fx->getId())) {
			++it;
			return fx->findParent(it);
		}
	}
	return NULL;
}

void Chain::setParam(const param::UrlIterator&, param::SetArgs&)
{
}

int Chain::forEach(Visitor& op, VisitParams& vp)
{
	for (int i=0; i<mFx.size(); i++) {
		vp.mDepth++;
		const int		ans = (mFx[i] != NULL ? mFx[i]->forEach(op, vp) : fx::CONTINUE);
		vp.mDepth--;
		if (ans < fx::CONTINUE) return ans;
	}
	return fx::CONTINUE;
}

int Chain::edit(fx::EditOp& op)
{
	const int			mode = op.mode();
	// If the op refers to this FX and it's add mode, append the new fx at the end
	if (mode == EditOp::ADD_MODE) {
		if (op.fxId() == mId) {
			Fx*			fx = op.popFx();
			if (fx != NULL) {
				append(fx);
				op.setToSuccess();
			}
			return fx::STOP;
		}
	// If the op is a remove and I contain the selected ID, remove it.
	} else if (mode == EditOp::REMOVE_MODE) {
		const int		removeId = op.fxId();
		for (int i=0; i<mFx.size(); i++) {
			if (mFx[i] != NULL && mFx[i]->getId() == removeId) {
				remove(i);
				op.setToSuccess();
				return fx::STOP;
			}
		}
	}
	// Otherwise pass to my children.
	for (int i=0; i<mFx.size(); i++) {
		const int		ans = (mFx[i] != NULL ? mFx[i]->edit(op) : fx::CONTINUE);
		if (ans < fx::CONTINUE) return ans;
	}
	return fx::CONTINUE;
}

void Chain::append(Fx* fx)
{
	if (fx == NULL) return;

	mFx.append(fx);
	fx->setParent(this);
}

void Chain::remove(const int idx)
{
	if (idx < 0 || idx >= mFx.size()) return;

	delete mFx.takeAt(idx);
}

void Chain::work(fx::WorkArgs& a, fx::WorkOutput& o)
{
	for (int i=0; i<mFx.size(); i++) {
		if (mFx[i] != NULL) mFx[i]->work(a, o);
	}
}

/* IO
 ******************************************************************/
void Chain::writeJson(json::Writer& w) const
{
	writeJsonHeader(w, FX_NAME);

	// body
	w.key(ALL_STR);
	w.beginArray();
	for (int i=0; i<mFx.size(); i++) {
		if (mFx[i] != NULL) mFx[i]->writeJson(w);
	}
	w.endArray();

	writeJsonFooter(w);
}

void Chain::startArray(json::Stack& stack, const QString& key)
{
	if (key == ALL_STR) stack.push(new fx::Reader(*this));
}

void Chain::parseFx(Fx* fx)
{
	append(fx);
}

/* FX::CHAIN-FACTORY
 ******************************************************************/
ChainFactory::ChainFactory()
{
}

QString ChainFactory::label() const
{
	return FX_NAME;
}

QString ChainFactory::key() const
{
	return FX_NAME;
}

float ChainFactory::match(const QString& key) const
{
	if (key == FX_NAME) return 1;
	return 0;
}

Fx* ChainFactory::create() const
{
	return new Chain();
}
