#include "glow.h"

#include "math/geo.h"
#include "engine/arc.h"
#include "engine/celldistance.h"
#include "engine/colorf.h"

using namespace fx;

static const QString		FX_NAME("glow");
static const QString		IN_NAME("in");
static const int			IN_ID = 0;
static const QString		OUT_NAME("out");
static const int			OUT_ID = 1;
static const int			CLR_ID = 100;
static const int			ARC_ID = 101;

// FLAGS
static const int			CLR_EXP = (1<<0);
static const int			ARC_EXP = (1<<1);

// Hardcoded limit on depth amount right now
static const float			DEPTH_MAX = 256;

/* FX::GLOW
 ******************************************************************/
Glow::Glow()
	: Fx(-1)
	, mIn(40)
//	, mIn(0)
	, mOut(0)
	, mWorkMasks(2)
{
	mClr.setId(CLR_ID);
	mClr.scaleAlpha(0.65f);

	mArc.setId(ARC_ID);
}

Glow::Glow(const Glow& o)
	: Fx(o)
	, mIn(o.mIn)
	, mOut(o.mOut)
	, mClr(o.mClr)
	, mArc(o.mArc)
	, mWorkMasks(2)
{
	mClr.setId(CLR_ID);
	mArc.setId(ARC_ID);
}

Glow* Glow::copy() const
{
	return new Glow(*this);
}

void Glow::init(fx::Setup& setup)
{
	Fx::init(setup);
	if (mClr.isEmpty()) setup.getColorSrc(mClr, COLOR_SRC_SOLO);
	if (mArc.isEmpty()) setup.getArc(mArc, 2);
}

QString Glow::getLabel() const
{
	return FX_NAME;
}

void Glow::getParams(param::List& list, param::GetArgs& get) const
{
	list.addParam(&FX_NAME, IN_NAME, IN_ID, mIn / DEPTH_MAX, get);
	list.addParam(&FX_NAME, OUT_NAME, OUT_ID, mOut / DEPTH_MAX, get);

	if (get.supports(param::COLOR_FEATURE)) {
		param::List::Grouper	g(list, COLOR_SZ);
		mClr.addTypeParam(list, get, CLR_EXP, mFlags);
	}
	if (get.supports(param::ARC_FEATURE)) {
		param::List::Grouper	g(list, ARC_SZ);
		mArc.addTypeParam(list, get, ARC_EXP, mFlags);
	}
}

param::Parent* Glow::findParent(param::UrlIterator& it)
{
	// If I hit a composite object, there are two possibilities:
	// If the URL is a leaf, then someone is trying to replace
	// the entire composite.  Otherwise they're looking to set a param.
	if (it.matches(CLR_ID)) {
		if (it.isLeaf()) return this;
		++it;
		return mClr.findParent(it);
	} else if (it.matches(ARC_ID)) {
		if (it.isLeaf()) return this;
		++it;
		return mArc.findParent(it);
	}
	return this;
}

static bool set_depth(const float src, float& dst)
{
	float		trans = src * DEPTH_MAX;
	if (trans < 0) trans = 0;
	else if (trans > DEPTH_MAX) trans = DEPTH_MAX;
	if (dst == trans) return false;		// no change
	dst = trans;
	return true;
}

void Glow::setParam(const param::UrlIterator& it, param::SetArgs& sa)
{
	if (it.matches(IN_ID)) {
		if (!set_depth(sa.mV, mIn)) return;
	} else if (it.matches(OUT_ID)) {
		if (!set_depth(sa.mV, mOut)) return;
	// If I reached here, then I am a direct set.  Any parameters
	// inside the composites would go to their parent.
	} else if (it.matches(CLR_ID)) {
		mClr.swap(sa.mClr);
	} else if (it.matches(ARC_ID)) {
		mArc.setContentTo(sa.mArc);
	} else {
		return;
	}
	sa.mChanged = true;
}

void Glow::work(fx::WorkArgs& arg, fx::WorkOutput& out)
{
	CellDistance			cd;
	float					minD, maxD;
	arg.cache(cd);
	arg.getDistanceRange(&minD, &maxD);

	// Operate on buffers equal to the width of the image
	const int				w = out.mImg.width(), h = out.mImg.height();
	if (!mArcP.begin(&mArc, w, 0)) return;
	Cell*					cell = arg.getCell();
	float*					img = arg.getImg();
	const ColorSrc*			csrc = mClr.getSrc();
	if (cell == NULL || img == NULL || csrc == NULL) return;
	if (!mWorkMasks.start(arg, w)) return;

	color_f					clr;
	Arc*					arc = mArc.getArc();
	// The mask determines the visibility, including transparency
	float*					mask = mWorkMasks.d[0];
	float*					rowV = mWorkMasks.d[1];

	// Find pixels within the range of my in and out depth amounts.
	// If my amounts are at their limits or beyond the actual values,
	// then clip to thos values.  You might think we want to let users
	// overdrive it, but that's what the arc is for.  And for the arc
	// to work, we need an exact scaling to 0 - 1.
	Range					r(-mOut, mIn);
	if (mOut >= DEPTH_MAX || r.mMin < minD) r.mMin = minD;
	if (mIn >= DEPTH_MAX || r.mMax > maxD) r.mMax = maxD;

	for (int y=0; y<h; y++) {
		// Fill values of 0 - 1 to be sent to the arc.
		int				pix = IDX_FOR_PIXEL(0, y, w);
		set_to_0(mask, w);
		set_to_0(rowV, w);
		for (int x=0; x<w; x++) {
			float			v = cell[pix].mD;
			if (v >= r.mMin && v <= r.mMax) {
				rowV[x] = 1-r.translateTo(v, RANGE_UNI);
				// aa
				if (v < r.mMin+1) mask[x] = v-r.mMin;
				else if (v > r.mMax-1) mask[x] = r.mMax-v;
				else mask[x] = 1;
			}
			pix++;
		}

		// Process the row
		if (arc != NULL) arc->work(mArcP, rowV);

		// Fill values of 0 - 1 to be sent to the arc.
		pix = IDX_FOR_PIXEL(0, y, w);
		for (int x=0; x<w; x++) {
			if (mask[x] > 0) {
				csrc->get(rowV[x], clr);
				if (clr.a > 0) {
					if (mask[x] < 1) clr.setA_Pm(clr.a*mask[x]);
					srcOver(clr, img, pix);
				}
			}
			pix++;
		}
	}
/*
	color_f		clr2;
	clr2.setRgba(255, 0, 0, 255);
	srcOver(clr2, img, (140*w)+143);
*/
	mWorkMasks.stop(arg);
}

/* IO
 ******************************************************************/
void Glow::writeJson(json::Writer& w) const
{
	writeJsonHeader(w, FX_NAME);

	w.key(IN_NAME);
	w.floatValue(mIn);
	w.key(OUT_NAME);
	w.floatValue(mOut);
	mClr.writeJsonObject(w, COLOR_SZ);
	mArc.writeJsonObject(w, ARC_SZ);

	writeJsonFooter(w);
}

void Glow::startObject(json::Stack& s, const QString& k)
{
	if (k == COLOR_SZ) s.push(&mClr);
	else if (k == ARC_SZ) s.push(&mArc);
}

void Glow::jsonFloat(json::Stack& s, const QString& k, const float v)
{
	if (k == IN_NAME) mIn = v;
	else if (k == OUT_NAME) mOut = v;
	else Fx::jsonFloat(s, k, v);
}

/* FX::GLOW-FACTORY
 ******************************************************************/
GlowFactory::GlowFactory()
{
}

QString GlowFactory::label() const
{
	return FX_NAME;
}

QString GlowFactory::key() const
{
	return FX_NAME;
}

float GlowFactory::match(const QString& key) const
{
	if (key == FX_NAME) return 1;
	return 0;
}

Fx* GlowFactory::create() const
{
	return new Glow();
}
