#include "light.h"

#include <stdlib.h>
#include <math.h>
#include <QtDebug>
#include "engine/arc.h"
#include "engine/celldistance.h"
#include "engine/colorf.h"
#include "math/geo.h"
#include "math/primitives/point3.h"

#include "fx/bevel_cellnormal.h"

#include "experimental/exp_defs.h"

using namespace fx;

static const QString		FX_NAME("light");

static const int			CLR_ID = 100;		// XXX Shared with bevel.cpp
static const int			ARC_ID = 101;		// XXX Shared with bevel.cpp

/* FX::LIGHT
 ******************************************************************/
Light::Light()
	: Fx(-1)
{
}

Light::Light(Context& ctx)
	: Fx(ctx.nextFxUniqueId())
{
}

Light::Light(const Light& o)
	: Fx(o)
	, mClr(o.mClr)
	, mArc(o.mArc)
{
}

Light::~Light()
{
}

Light* Light::copy() const
{
	return new Light(*this);
}

void Light::init(fx::Setup& setup)
{
	Fx::init(setup);
	if (mClr.isEmpty()) setup.getColorSrc(mClr, COLOR_SRC_DUO);
	if (mArc.isEmpty()) setup.getArc(mArc, 1);
//	if (mArc.isEmpty()) setup.getArc(mArc, 3);

	if (mArc.isEmpty()) setup.getArc(mArc, 1);
}

void Light::setLight(const float bevelDepth, const Point3& loc, const float min, const float max)
{
	mBevelDepth = bevelDepth,
	mLightLoc.set(loc);
	mLightMin = min;
	mLightMax = max;
}

QString Light::getLabel() const
{
	return FX_NAME;
}

void Light::getParams(param::List&, param::GetArgs&) const
{
/*
	bool			ans = true;
	if (f != NULL && f->supports(Param::Filter::ARC_FEATURE)) {
		add_arc_browser_param(list, ARC_NAME, getId(), ARC_ID, f, &ans);
	}
	return ans;
*/
}

/*
bool Light::setParam(fx::ParamArgs& p)
{
	if (p.mKey.mParam == ARC_ID) {
		mArc.setTo(p.mArc);
		return true;
	}
	return false;
}
*/
param::Parent* Light::findParent(param::UrlIterator& it)
{
	if (!it.hasValue()) return NULL;
	return NULL;
}

void Light::setParam(const param::UrlIterator&, param::SetArgs&)
{
}

void Light::work(fx::WorkArgs& arg, fx::WorkOutput& out)
{
	CellDistance		cd;
	float				maxD;
	arg.cache(cd);
	arg.getDistanceRange(NULL, &maxD);
	if (maxD <= 0) maxD = 1;

	const QSize			imgSize = out.mImg.size();
	const Cell*			cell = arg.getCell();
	float*				img = arg.getImg();
	const ColorSrc*		csrc = mClr.getSrc();
	if (cell == NULL || img == NULL || csrc == NULL) return;

	const int			w = imgSize.width(), h = imgSize.height();
	Point3				pt;
	const float			globalLength = mLightLoc.length();
	int					pix = 0;

	if (!mArcP.begin(&mArc, w, arc::HAS_DISTANCE_SRC | arc::HAS_NORMAL_SRC)) return;

	// The final values are written to a float row, in case I have
	// an arc that needs to do processing.
	float*				rowV = mRowV.get(w);
	float*				rowMask = mRowMask.get(w);
	if (rowV == NULL || rowMask == NULL) return;

	float				fadeSize = mBevelDepth * 0.45;
	if (fadeSize < 4) fadeSize = 4;
	float				fadeStart = mBevelDepth - fadeSize;
	if (fadeStart < 0) fadeStart = 0;
	float				fadeEnd = fadeStart + fadeSize;
	if (fadeEnd > mBevelDepth) fadeEnd = mBevelDepth;

	const float			expMin = -(1/mLightMin), expMax = 1/mLightMax;
	Arc*				arc = mArc.getArc();
	color_f				clr;
	Point3				nrmPt;
	for (int y=0; y<h; y++) {
		pt.y = y;

		set_to_neg_1(rowMask, w);

		// Find the light and distance values for each row
		for (int x=0; x<w; x++) {
			pt.x = x;

			const int		cp = (y*w)+x;
			if (cell[cp].mD > 0 && cell[cp].mD < mBevelDepth) {
				rowMask[x] = 1;
				nrmPt.set(cell[cp].mNormalX, cell[cp].mNormalY, cell[cp].mNormalZ);
				// Find the light value for the normal and normalize it
				const float			light = bevel::lightFrom(nrmPt, mLightLoc, globalLength);
				if (light <= mLightMin) mArcP.mNormalRow[x] = -1;
				else if (light >= mLightMax) mArcP.mNormalRow[x] = 1;
				else if (light < 0) mArcP.mNormalRow[x] = light * expMin;
				else if (light > 0) mArcP.mNormalRow[x] = light * expMax;
// Without normalizing...
//				mArcP.mNormalRow[x] = bevel::lightFrom(nrmPt, mLightLoc, globalLength);
				mArcP.mDistanceRow[x] = (cell[cp].mD / maxD);
			}
		}

		// Process the row
		set_to_1(rowV, w);
		if (arc != NULL) arc->work(mArcP, rowV);

		// Apply the row
		for (int x=0; x<w; x++) {
			const float			d = cell[pix].mD;
			if (rowMask[x] > 0 && rowV[x] != 0 && d > 0 && d < mBevelDepth) {
				// XXX Currently it's -1 to 1 but needs to be 0 to 1
				if (rowV[x] <= -1) rowV[x] = 0;
				else if (rowV[x] >= 1) rowV[x] = 1;
				else rowV[x] = (rowV[x]+1) * 0.5f;

				csrc->get(rowV[x], clr);
				// aa
				float			aa = 1;
				// Alpha blend the inner edge
				if (d > fadeStart) {
					if (d >= fadeEnd) aa = 0;
					else {
						const float		interp = 1-((d-fadeStart)/(fadeEnd-fadeStart));
						aa *= GEO_S_CURVE(interp);
					}
				}
				// Apply an antialias around the outer edge
				if (d < 1) aa *= d;
				if (aa < 1) clr.setA_Pm(clr.a * aa);
				if (clr.a > 0) srcOver(clr, img, pix);
			}
			pix++;
		}
	}
}

/* FX::LIGHT IO
 ******************************************************************/
void Light::writeJson(json::Writer& w) const
{
	w.beginObject();
	w.key(FX_SZ);
	w.stringValue(FX_NAME);
	w.endObject();
}
