#include "shape.h"

#include <QDebug>

#include "app/defs.h"
#include "engine/celldistance.h"
#include "io/svgreader.h"
#include "fxui/shapebrowser.h"

// For testing, force the shape to always be a square
//#define SQUARE_SHAPE		(1)

using namespace fx;

static const QString		FX_NAME("shape");
static const QString		FN_NAME("filename");
static const int			FN_ID = 0;
static const int			CLR_ID = 100;

static const int			LEFT_ID = 30;
static const int			TOP_ID = 31;
static const int			RIGHT_ID = 32;
static const int			BOTTOM_ID = 33;

// FLAGS
static const int			CLR_EXP = (1<<0);

static bool load_poly_from_svg(const QString& filename, const QRectF& dstF,
								QPolygonF& dst);

/* FX::SHAPE-SRC
 ******************************************************************/
Shape::Shape(const QString* filename)
	: Fx(-1)
	, mFrame(0, 0, -1, -1)
	, mNeedsCache(true)
{
	if (filename != NULL) mFilename = *filename;
	mClr.setId(CLR_ID);
}

Shape::Shape(const Shape& o)
	: Fx(o)
	, mFilename(o.mFilename)
	, mFrame(o.mFrame)
	, mClr(o.mClr)
	, mNeedsCache(o.mNeedsCache)
	, mShapeDir(o.mShapeDir)
{
}

Shape* Shape::copy() const
{
	return new Shape(*this);
}

void Shape::init(fx::Setup& setup)
{
	Fx::init(setup);
	if (mClr.isEmpty()) setup.getColorSrc(mClr, COLOR_SRC_SOLO);
	if (mFrame.right() <= mFrame.left() || mFrame.bottom() <= mFrame.top()) {
		const QSize		size = setup.getCanvasSize();
		mFrame.setLeft(0);
		mFrame.setTop(0);
		mFrame.setRight(size.width());
		mFrame.setBottom(size.height());
	}
	mShapeDir = setup.getInstallPath(SHAPES_INSTALL_DIR);
}

QString Shape::getLabel() const
{
	return FX_NAME;
}

static void	addRawFloat(	param::List& list, const QString& group, const QString& name,
							const int paramId, param::GetArgs& get, const float floatV)
{
	list.addParam(param::RAW_FLOAT_TYPE, &group, name, paramId, get, &floatV);
}

void Shape::getParams(param::List& list, param::GetArgs& get) const
{
	{
		param::List::Grouper	g(list, COLOR_SZ);
		mClr.addTypeParam(list, get, CLR_EXP, mFlags);
	}
	{
		param::List::Grouper	g(list, SHAPE_SZ);
		list.addTypeParam(FX_NAME, FN_ID, param::SHAPE_BROWSER_TYPE, get, 0, 0, &mFilename);
	}
	{
		param::List::Grouper	g(list, param::FRAME_SZ);
		addRawFloat(list, FX_NAME, param::LEFT_SZ, LEFT_ID, get, mFrame.left());
		addRawFloat(list, FX_NAME, param::TOP_SZ, TOP_ID, get, mFrame.top());
		addRawFloat(list, FX_NAME, param::RIGHT_SZ, RIGHT_ID, get, mFrame.right());
		addRawFloat(list, FX_NAME, param::BOTTOM_SZ, BOTTOM_ID, get, mFrame.bottom());
	}
}

param::Parent* Shape::findParent(param::UrlIterator& it)
{
	// If I hit a composite object, there are two possibilities:
	// If the URL is a leaf, then someone is trying to replace
	// the entire composite.  Otherwise they're looking to set a param.
	if (it.matches(CLR_ID)) {
		if (it.isLeaf()) return this;
		++it;
		return mClr.findParent(it);
	}
	return this;
}

void Shape::setParam(const param::UrlIterator& it, param::SetArgs& sa)
{
	if (it.matches(FN_ID)) {
		if (mFilename == sa.mStr) return;		// no change
		mFilename = sa.mStr;
	} else if (it.matches(CLR_ID)) {
		mClr.swap(sa.mClr);
	} else if (it.matches(LEFT_ID)) {
		if (mFrame.left() == sa.mV) return;
		mFrame.setLeft(sa.mV);
	} else if (it.matches(TOP_ID)) {
		if (mFrame.top() == sa.mV) return;
		mFrame.setTop(sa.mV);
	} else if (it.matches(RIGHT_ID)) {
		if (mFrame.right() == sa.mV) return;
		mFrame.setRight(sa.mV);
	} else if (it.matches(BOTTOM_ID)) {
		if (mFrame.bottom() == sa.mV) return;
		mFrame.setBottom(sa.mV);
	} else return;
	mNeedsCache = true;
	sa.mChanged = true;
}

void Shape::work(fx::WorkArgs& arg, fx::WorkOutput& o)
{
	if (mNeedsCache) {
		mNeedsCache = false;

#ifdef SQUARE_SHAPE
		mFrame.setCoords(4, 4, 252, 252);
#endif

		// Load the polygon
		QPolygonF&		shape = arg.editShape();
		shape.clear();
		// Convert the filename if necessary.
		QString			fn = mFilename;
		if (fn.startsWith(SHAPES_DIR)) fn.replace(0, SHAPES_DIR.length(), mShapeDir);
		load_poly_from_svg(fn, mFrame, shape);
	}

	// Draw it, if I've got a color.  Do this by using the
	// precomputed distance maps (all positive values are interiors).
	CellDistance		cd;
	arg.cache(cd);

	const QSize				imgSize = o.mImg.size();
	const int				w = imgSize.width(), h = imgSize.height();
	Cell*					cell = arg.getCell();
	float*					img = arg.getImg();
	const ColorSrc*			csrc = mClr.getSrc();
	if (cell == NULL || img == NULL || csrc == NULL) return;
	color_f					clr;
	csrc->get(1, clr);
	if (clr.a <= 0) return;	 // bail if completely transparent;

	int						pix = 0;
	color_f					clrAa;
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			// bypass the extra work for antialiasing if I can
			if (cell[pix].mD >= 1) {
				srcOver(clr, img, pix);
			} else if (cell[pix].mD > 0) {
				clrAa = clr;
				clrAa.setA_Pm(clr.a * cell[pix].mD);
				srcOver(clrAa, img, pix);
			}
			pix++;
		}
	}
}

/* IO
 ******************************************************************/
void Shape::writeJson(json::Writer& w) const
{
	writeJsonHeader(w, FX_NAME);

	w.key(FN_NAME);
	w.stringValue(mFilename);

	// No doubt we will do this a different way eventually
	w.key(param::LEFT_SZ);
	w.floatValue(mFrame.left());
	w.key(param::TOP_SZ);
	w.floatValue(mFrame.top());
	w.key(param::RIGHT_SZ);
	w.floatValue(mFrame.right());
	w.key(param::BOTTOM_SZ);
	w.floatValue(mFrame.bottom());

	mClr.writeJsonObject(w, COLOR_SZ);

	writeJsonFooter(w);
}

void Shape::startObject(json::Stack& s, const QString& k)
{
	if (k == COLOR_SZ) s.push(&mClr);
}

void Shape::jsonString(json::Stack&, const QString& k, const QString& v)
{
	if (k == FN_NAME) {
		mFilename = v;
		mNeedsCache = true;
	}
}

void Shape::jsonFloat(json::Stack& s, const QString& k, const float v)
{
	if (k == param::LEFT_SZ) mFrame.setLeft(v);
	else if (k == param::TOP_SZ) mFrame.setTop(v);
	else if (k == param::RIGHT_SZ) mFrame.setRight(v);
	else if (k == param::BOTTOM_SZ) mFrame.setBottom(v);
	else Fx::jsonFloat(s, k, v);
}

/* FX::SHAPE-FACTORY
 ******************************************************************/
ShapeFactory::ShapeFactory()
{
}

QString ShapeFactory::label() const
{
	return FX_NAME;
}

QString ShapeFactory::key() const
{
	return FX_NAME;
}

float ShapeFactory::match(const QString& key) const
{
	if (key == FX_NAME) return 1;
	return 0;
}

Fx* ShapeFactory::create() const
{
	return new Shape();
}

/* Misc
 ********************************************/
static bool load_poly_from_svg(const QString& filename, const QRectF& dstF,
								QPolygonF& dst)
{
#ifdef SQUARE_SHAPE
	QPolygonF			square;
	square.append(QPointF(0, 0));
	square.append(QPointF(200, 0));
	square.append(QPointF(200, 200));
	square.append(QPointF(0, 200));
	QList<QPolygonF>	pl;
	pl.append(square);
#else
	// Load the path from the file
	SvgReader			r;
	QPainterPath		p;
	if (!r.load(filename, &p)) return false;

	// Convert the path to a polygon
	QList<QPolygonF>	pl = p.toSubpathPolygons();
#endif
	if (pl.size() < 1) return false;

	const QPolygonF&	poly = pl.at(0);

	// Translate the polygon to local coords
	const QRectF		srcF(poly.boundingRect());
	QTransform			trans;
	{
		QPolygonF		a, b;
		a << srcF.topLeft() << srcF.topRight() << srcF.bottomRight() << srcF.bottomLeft();
		b << dstF.topLeft() << dstF.topRight() << dstF.bottomRight() << dstF.bottomLeft();
		if (!QTransform::quadToQuad(a, b, trans)) return false;
	}
	for (int k=0; k<poly.size(); k++) {
		dst.append(poly.at(k) * trans);
	}
	return true;
}
