#include "fxui/parameditor.h"

#include <assert.h>

#include <QDebug>
#include <QPainter>

#include "engine/engineparam.h"
#include "model/context.h"
#include "ui/look.h"
#include "fxui/arcbrowser.h"
#include "fxui/colorbrowser.h"
#include "fxui/shapebrowser.h"

using namespace fxui;
/* FX-E.  One param entry.
 **************************************************************/
class ParamEditor::ExpandE : public ParamEditor::FxE {
public:
	ExpandE(ParamEditor& pe, const Param& p, const QSize& s,
			const int f, const bool expanded)
		: FxE(pe, p, s, f)
		, mBit(p.mIV)
		, mExpanded(expanded)
	{
	}

	virtual void		onPaint(paint_args& pa)
	{
		QPolygonF			poly;
		const QRect&		frame = pa.mFrame;
		const float			borderX = (pa.mFrame.width()*0.25f),
							borderY = (pa.mFrame.height()*0.15f);
		const float			l = pa.mFrame.left() + borderX,
							t = pa.mFrame.top() + borderY,
							r = pa.mFrame.right() - borderX,
							b = pa.mFrame.bottom() - borderY;
		const QPointF		cen(frame.center());
		if (mExpanded) {
			poly.append(QPointF(l, cen.y()));
			poly.append(QPointF(r, t));
			poly.append(QPointF(r, b));
		} else {
			poly.append(QPointF(l, t));
			poly.append(QPointF(r, cen.y()));
			poly.append(QPointF(l, b));
		}

		QPainter&			painter = *pa.mPainter;
		const Look&			look = *pa.mLook;
		painter.setRenderHint(QPainter::Antialiasing);
		painter.setPen(Qt::NoPen);
		painter.setBrush(look.mParamFgB);
		painter.drawPolygon(poly);
	}

	virtual void		mousePressEvent(const mouse_args& down) { }
	virtual void		mouseMoveEvent(const mouse_args& down) { }

	virtual void		mouseReleaseEvent(const mouse_args& down)
	{
		if (down.mEntry != this) return;
		param::Url		url(mUrl);
		url.popBack();	// go to the FX level
		mP.mPrj.setFlag(url, mBit, !mExpanded);
	}

private:
	const int			mBit;
	const bool			mExpanded;
};

/* PARAM-EDITOR::PARAM-FILTER
 ******************************************************************/
class ParamEditor::ParamFilter : public param::Filter
{
public:
	virtual bool accept(const QString*, const QString*, const int type) const
	{
		return type == Param::FLOAT_TYPE
				|| type == param::ARC_BROWSER_TYPE
				|| type == param::COLOR_BROWSER_TYPE
				|| type == param::SHAPE_BROWSER_TYPE;
	}

	virtual bool supports(const int feature) const
	{
		return (feature&(param::ARC_FEATURE | param::COLOR_FEATURE)) != 0;
	}

};
static const ParamEditor::ParamFilter		gFilter;

/* PARAM-EDITOR
 ******************************************************************/
ParamEditor::ParamEditor(Create& c, const param::Url& url)
	: mPrj(c.mPrj)
	, mEnv(c.mEnv)
	, mArcs(c.mArcs)
	, mClrs(c.mClrs)
	, mUrl(url)
	, mNeedsCache(true)
{
}

ParamEditor::~ParamEditor()
{
	uncache();
}

void ParamEditor::setParent(Parent* p, const Look& look)
{
	assert(p != NULL);
	inherited::setParent(p, look);

	// Should check to see if the folder date stamp changed
	if (mNeedsCache) {
		cache(p->entryH(), look);
		mNeedsCache = false;
	}

	for (int i=0; i<mS.size(); i++) {
		if (mS[i] == NULL) continue;

		if (mS[i]->mBrowser != NULL) {
			p->addBrowser(mS[i]->mBrowser, mS[i]->mFlags, mS[i]->isExpanded());
		} else p->add(mS[i]);
	}
}

int ParamEditor::onProjectChanged()
{
	if (!mPrj.contains(mUrl)) return FxUi::UI_ERROR;
	uncache();
	mNeedsCache = true;
	return FxUi::UI_CHANGED;
}

void ParamEditor::cache(const int entryH, const Look& look)
{
	param::List				list;
	param::GetArgs			ga;
	ga.mUrl = mUrl;
	ga.mFilter = &gFilter;
	mPrj.getParams(list, ga);

	cacheList(list, entryH, look);
}

void ParamEditor::cacheList(const param::List& list, const int entryH, const Look& look)
{
	param::ListIterator		it(list);
	const Param*			p;
	QString					groupName;
	while ((p = *it) != NULL) {
		FxE*				lastE = (!mS.isEmpty() ? mS.back() : NULL);
		int					flags = 0;
		if (groupName == p->mGroupName) {
			if (lastE != NULL) lastE->mFlags |= Entry::CONNECT_RIGHT;
			flags = Entry::CONNECT_LEFT;
		}
		// If the previous connects to me, I must connect to it.
		if (lastE != NULL && (lastE->mFlags&Entry::CONNECT_RIGHT) != 0) flags |= Entry::CONNECT_LEFT;
		groupName = p->mGroupName;
		if (p->mType == param::ARC_BROWSER_TYPE) {
			cacheBrowser(	*p, new ArcBrowser(mPrj, mArcs, *p, QSize(entryH, entryH)),
							entryH, look);
		} else if (p->mType == param::COLOR_BROWSER_TYPE) {
			cacheBrowser(	*p, new ColorBrowser(mPrj, mClrs, *p, QSize(entryH, entryH)),
							entryH, look);
		} else if (p->mType == param::SHAPE_BROWSER_TYPE) {
			cacheBrowser(	*p, new ShapeBrowser(mPrj, *p, QSize(entryH, entryH), mEnv, look),
							entryH, look);
		} else {
			FxE*				e = new FxE(*this, *(p), QSize(60, entryH), flags);
			if (e != NULL) mS.append(e);
		}
		++it;
	}
}


void ParamEditor::cacheBrowser(const Param& p, fxui::BrowserModel* bm,
							   const int entryH, const Look& look)
{
	if (bm == NULL) return;
	const QSize			size(60, entryH);
	const bool			isEmpty = p.mList.isEmpty();
	// If the browser has no subparams, then it's drawn as its own element,
	// otherwise it needs to connect to the next.
	int					flags = (isEmpty ? 0 : Entry::CONNECT_RIGHT);
	FxE*				e = new FxE(*this, p, size, flags);
	if (e == NULL) {
		delete bm;
		return;
	}
	e->mBrowser = bm;
	mS.append(e);
	if (!isEmpty) {
		// Add the expand/contract button
		FxE*		ee = NULL;
		if (param::isExpanded(p.mFV)) {
			cacheList(p.mList, entryH, look);
			FxE*	last = (mS.isEmpty() ? NULL : mS.back());
			if (last != NULL) last->mFlags |= Entry::CONNECT_RIGHT;
			ee = new ExpandE(*this, p, size, Entry::CONNECT_LEFT, true);
		} else {
			ee = new ExpandE(*this, p, size, Entry::CONNECT_LEFT, false);
		}
		if (ee != NULL) mS.append(ee);
	}
}

void ParamEditor::uncache()
{
	for (int i=0; i<mS.size(); i++) delete mS[i];
	mS.clear();
}

/* FX-E
 *******************************/
ParamEditor::FxE::FxE(ParamEditor& pe, const Param& p,
					  const QSize& size, const int flags)
	: mBrowser(NULL)
	, mFlags(flags)
	, mP(pe)
	, mSize(size)
	, mParamType(p.mType)
{
	mUrl = p.mUrl;
	mLabel = p.mParamName;
	mV = p.mFV;
}

ParamEditor::FxE::~FxE()
{
	delete mBrowser;
}

QSize ParamEditor::FxE::size() const
{
	return mSize;
}

bool ParamEditor::FxE::getParam(Param& p) const
{
	p.mUrl = mUrl;
	p.mParamName = mLabel;
	p.mType = mParamType;
	return true;
}

void ParamEditor::FxE::onPaint(paint_args& pa)
{
	QPainter&			painter = *pa.mPainter;
	const Look&			look = *pa.mLook;
	const QRect&		frame = pa.mFrame;

	painter.setRenderHint(QPainter::Antialiasing);

	painter.setPen(Qt::NoPen);

	// Label
	if (mLabel.size() > 0) {
		painter.setPen(look.mParamFgC);
		painter.drawText(frame.left() + look.mPadX, frame.top() + look.mPadY + look.mFa, mLabel);
		painter.setPen(Qt::NoPen);
	}

	// Param value background
	const int		l = frame.left() + look.mPadX;
	const int		t = frame.top() + look.mPadY + look.mFh + look.mPadY;
	const int		valueW = frame.width() - look.mPadX * 2;
	const int		valueH = (frame.bottom() - look.mPadY) - t;
	painter.setBrush(look.mParamBgToneB);
	painter.drawRoundedRect(l, t, valueW, valueH, look.mRound, look.mRound);

	// Param value
	const int		curW = (int)(float(valueW) * mV);
	painter.setBrush(look.mParamFgB);
	painter.drawRoundedRect(l, t, curW, valueH, look.mRound, look.mRound);
}

void ParamEditor::FxE::mousePressEvent(const mouse_args& down)
{
	mP.mDownPt.setX(down.mEvent->x());
	mP.mDownPt.setY(down.mEvent->y());
	mP.mDownV = mV;
}

static float new_value(const QPoint& startPt, const int curX, const QRect& f, const float origV)
{
	const int		offset = curX - startPt.x();
	const float		newV = origV + (float(offset) / (float(f.width()*2)));
	if (newV <= 0) return 0;
	if (newV >= 1) return 1;
	return newV;
}

void ParamEditor::FxE::mouseMoveEvent(const mouse_args& ma)
{
	const float		newV = new_value(mP.mDownPt, ma.mEvent->x(), ma.mFrame, mP.mDownV);
	mP.mPrj.setParamF(mUrl, newV);
}

void ParamEditor::FxE::mouseReleaseEvent(const mouse_args&)
{
}

int ParamEditor::FxE::onParamChanged(const Param& p)
{
	if (mUrl == p.mUrl) {
		// Indicate that not only is there nothing to update,
		// but I can stop looking for something.
		if (mV == p.mFV) return NO_PARAMS;
		mV = p.mFV;
		return PARAM;
	}
	return WRONG_PARAM;
}

bool ParamEditor::FxE::isExpanded() const
{
	return param::isExpanded(mV);
}
