#include "fxui/shapebrowser.h"

#include <QDir>
#include <QPixmap>
#include <QStringList>

#include "model/context.h"
#include "engine/fxdefs.h"
#include "ui/look.h"
#include "io/svgreader.h"
#include "app/defs.h"
#include "app/environment.h"

using namespace fxui;

static bool transform_frame(const QRectF& srcF, const int dstS, QTransform&, const Look&);

/* SHAPE-BROWSER
 ******************************************************************/
ShapeBrowser::ShapeBrowser(Context& prj,
							const Param& p, const QSize& iconSize,
							const Environment& env, const Look& look)
	: mPrj(prj)
	, mIconSize(iconSize)
	, mIdx(0)
{
	mUrl = p.mUrl;
	cacheAll(env, look);
	for (int k=0; k<mE.size(); k++) {
		const entry*		e = mE[k];
		if (e != NULL && e->mFn == p.mSV) {
			mIdx = k;
			break;
		}
	}
}

ShapeBrowser::~ShapeBrowser()
{
	clear();
}

int ShapeBrowser::size() const
{
	return mE.size();
}

QSize ShapeBrowser::iconSize() const
{
	return mIconSize;
}

const QPixmap* ShapeBrowser::at(const int idx)
{
	const entry*		e = getAt(idx);
	if (e != NULL) return &(e->mBm);
	return NULL;
}

int ShapeBrowser::getIndex() const
{
	return mIdx;
}

void ShapeBrowser::setIndex(const int idx)
{
	const entry*		e = getAt(idx);
	if (e != NULL) {
		mPrj.setParamStr(mUrl, e->mFn);
		mIdx = idx;
	}
}

void ShapeBrowser::clear()
{
	while (!mE.isEmpty()) delete mE.takeFirst();
}

const ShapeBrowser::entry* ShapeBrowser::getAt(const int idx) const
{
	if (idx < 0 || idx >= mE.size()) return NULL;
	return mE[idx];
}

void ShapeBrowser::cacheAll(const Environment& env, const Look& look)
{
	// Create all the entries here.
	// Of course, this should be in a separate thread
	// and there should be a signal to indicate when a
	// new entry is ready to display.  That separate thread
	// also needs an object that copies the relevant fields
	// from "Look"

	const QDir			dir(env.getInstallPath(SHAPES_INSTALL_DIR));
	const QString		pdir(dir.absolutePath());
	const QStringList	files = dir.entryList(QDir::Files);
	for (int k=0; k<files.size(); k++) {
		// Create an entry with the full file path (for loading it in and generating a preview)
		// and the abstracted file path (for compatibility with the shape FX).
		entry*			e = newEntry(	pdir + QDir::separator() + files.at(k),
										SHAPES_DIR + QDir::separator() + files.at(k), look);
		if (e != NULL) mE.append(e);
	}
}

ShapeBrowser::entry* ShapeBrowser::newEntry(const QString& fullFn, const QString& fn, const Look& look) const
{
	// Load a path for the underlying SVG file.
	SvgReader		reader;
	QPainterPath	path;
	if (!reader.load(fullFn, &path)) {
		qDebug() << "ERROR LOADING " << fullFn;
		return NULL;
	}

	entry*			e = new entry();
	if (e == NULL) return NULL;

	const int		size = mIconSize.width();
	QImage*			img = new QImage(size, size, QImage::Format_RGB888);
	if (img == NULL) {
		delete e;
		return NULL;
	}

	{
		QPainter		painter(img);
		// Clear the background
		painter.setPen(Qt::NoPen);
		painter.setBrush(look.mDashBgB); //.mAlphaB);
		painter.drawRect(0, 0, size, size);
		// Draw a param panel
		painter.setBrush(look.mParamBgB);
		painter.setRenderHint(QPainter::Antialiasing);
		painter.drawRoundedRect(0, 0, size, size, look.mRound, look.mRound);
		// Translate the path to local coords
		const QRectF	pathF(path.boundingRect());
		QTransform		trans;
		transform_frame(pathF, size, trans, look);
		painter.setTransform(trans);
		// Draw the path
		painter.setBrush(look.mParamFgB);
		painter.drawPath(path);
	}

	e->mFn = fn;
	// Kinda weird -- where's the error check?
	e->mBm = QPixmap::fromImage(*img);
	delete img;
	return e;
}

/* MISC
 *******************************/
static bool transform_frame(const QRectF& srcF, const int dstS, QTransform& trans, const Look& look)
{
	QPolygonF		a, b;
	const int		w = dstS - look.mPadX*2, h = dstS - look.mPadY*2;
	QRectF			dstF(look.mPadX, look.mPadY, w-1, h-1);
	a << srcF.topLeft() << srcF.topRight() << srcF.bottomRight() << srcF.bottomLeft();
	b << dstF.topLeft() << dstF.topRight() << dstF.bottomRight() << dstF.bottomLeft();

	return QTransform::quadToQuad(a, b, trans);
}
