#include "json_parser.h"

#include <assert.h>
#include <QIODevice>
#include <QTextStream>
#include <QDebug>

using namespace json;

static const int			BUF_SIZE = 512;
static const QString		EMPTY_KEY;

static const int			TOKEN_OFF = 0;
static const int			TOKEN_OBJECT = (1<<0);
static const int			TOKEN_ARRAY = (1<<1);

static const QChar*			parse_string(const QChar*);

// Set the string, NOT COPYING ANY DATA.
static void getReference(const QChar* c, const int size, QString& out);
// Set the string, copying the data.
static void getCopy(const QChar* c, const int size, QString& out);

/* JSON::STATE
 ******************************************************************/
class json::State
{
public:
	State() : mR(NULL), mToken(TOKEN_OFF)	{ }

	Reader*			mR;
	int				mToken;		// Off, object or array

	void			clear() {
		mR = NULL;
		mToken = TOKEN_OFF;
	}

	void			set(const State& o) {
		mR = o.mR;
		mToken = o.mToken;
	}

private:
	State(const State&);
	State&			operator=(const State& o);
};

/* JSON::STATE-FACTORY
 * Memory management for State class.
 ******************************************************************/
class json::StateFactory : public ManagedFactory<json::State*>
{
public:
	virtual State*		newObject() const {
		State*			ans = new State();
		if (ans == NULL) throw std::exception();
		return ans;
	}

	virtual void		deleteObject(State* t) const {
		delete t;
	}

	virtual void		initializeObject(State*) const {
	}

	virtual void		setObject(State* dst, State* src) const {
		if (src != NULL) dst->set(*src);
	}
};
static StateFactory		STATE_FACTORY;

/* JSON::PARSER
 ******************************************************************/
Parser::Parser()
	: mState(STATE_FACTORY)
	, mCur(NULL)
	, mValueStart(NULL)
{
}

Parser::~Parser()
{
	clear();
}

bool Parser::parse(QIODevice* io, Reader* r)
{
	if (io == NULL || r == NULL) return false;

	clear();
	// Intialize to empty state.
	if (!push(r)) return false;

	mHasSeparator = false;
	mValueStart = NULL;

	// OK...  not sure how we're going to handle things once we
	// have image data embedded in the file, but for now we'll go
	// the easy route.  Probably we'll make a rule that any single
	// piece of data must be on one line, and then use readLine().
	QTextStream			ts(io);
	QString				ans = ts.readAll();
	if (!handle(ans)) return false;
	return true;
}

bool Parser::push(Reader* r)
{
	if (r == NULL) return false;
	try {
		State*			n = mState.pushFront();
		n->clear();
		n->mR = r;
		mCur = r;
		return true;
	} catch (std::exception&) {
	}
	return false;
}

bool Parser::handle(const QString& src)
{
	const QChar*				c = src.data();
	while (!c->isNull()) {
		if (mCur == NULL) return false;
		if (*c == '{') {
//			qDebug() << "startObject";
			if (!pushToken(TOKEN_OBJECT)) return false;
			mCur->startObject(*this, mKey);
		} else if (*c == '}') {
//			qDebug() << "endObject";
			flushValue(c);
			if (!popTo(TOKEN_OBJECT)) return false;
		} else if (*c == '[') {
//			qDebug() << "startArray";
			if (!pushToken(TOKEN_ARRAY)) return false;
			mCur->startArray(*this, mKey);
		} else if (*c == ']') {
//			qDebug() << "endArray";
			flushValue(c);
			if (!popTo(TOKEN_ARRAY)) return false;
		// key-value separator
		} else if (*c == ':') {
			mHasSeparator = true;
		// array element separator
		} else if (*c == ',') {
			flushValue(c);
			mHasSeparator = false;
		} else if (*c == '"') {
			// Skip the starting quote
			++c;
			const QChar*	end = parse_string(c);
			if (end->isNull()) return false;
			const int		size = (end - c);
			if (mHasSeparator) {
				if (mCur == NULL) return false;
				mHasSeparator = false;
				getCopy(c, size, mValue);
				mCur->jsonString(*this, mKey, mValue);
				mKey.clear();
			} else {
				getReference(c, size, mKey);
			}
			c = end;
		} else {
			if (c->isSpace()) flushValue(c);
			else if (mValueStart == NULL) mValueStart = c;
//			qDebug() << "unhandled=" << *c;
		}

		++c;
	}
	return true;
}

void Parser::flushValue(const QChar* end)
{
	const QChar*			valueStart = mValueStart;
	mValueStart = NULL;
	// Number values must always appear after keys, right?
	if (valueStart == NULL || valueStart->isNull() || end == NULL || !mHasSeparator) return;

	const int				size = (end - valueStart);
	if (size < 1) return;

	getReference(valueStart, size, mValue);
	bool					ok;
	const float				v = mValue.toFloat(&ok);
	if (ok && mCur != NULL) {
		mCur->jsonFloat(*this, mKey, v);
	}
}

void Parser::clear()
{
	mState.clear();
	mCur = NULL;
}

bool Parser::pushToken(const int token)
{
	mHasSeparator = false;
	try {
		State*			n = mState.pushFront();
		n->clear();
		n->mToken = token;
		return true;
	} catch (std::exception&) {
	}
	return false;
}

bool Parser::popTo(const int token)
{
	mCur = NULL;
	mHasSeparator = false;
	if (!removeTo(token)) return false;
	if (mState.isEmpty()) return false;
	return getCurrentReader();
}

bool Parser::removeTo(const int token)
{
	while (!mState.isEmpty()) {
		State*			n = mState.front();
		int				tok = 0;
		if (n != NULL) {
			tok = n->mToken;
			if (n->mR != NULL) {
				if ((token&TOKEN_OBJECT) != 0) n->mR->endObject(*this);
				if ((token&TOKEN_ARRAY) != 0) n->mR->endArray(*this);
				n->mR->jsonPop();
			}
			mState.popFront();
		}
		if ((tok&token) != 0) return true;
	}
	return false;
}

bool Parser::getCurrentReader()
{
	mCur = NULL;
	ManagedListIterator<State*>		it(mState);
	while (it.hasValue()) {
		State*						n = *it;
		++it;

		if (n != NULL && n->mR != NULL) {
			mCur = n->mR;
			return true;
	   }
	}
	return false;
}

/* Misc
 ******************************************************************/
static const QChar* parse_string(const QChar* c)
{
	while (!c->isNull()) {
		if (*c == '"') return c;
		++c;
	}
	return c;
}

static void getReference(const QChar* c, const int size, QString& out)
{
	if (size > 0) out.setRawData(c, size);
	else out.clear();
}

static void getCopy(const QChar* c, const int size, QString& out)
{
	if (size > 0 && c != NULL) out.setUnicode(c, size);
	else out.clear();
}
