#ifndef JSON_READER_H
#define JSON_READER_H

#include <QString>
#include "io/json_stack.h"

namespace json {

/* JSON::READER
 * Read data from a JSON-formatted stream.
 * For an explanation of JSON, see: http://www.json.org/
 *
 * Clients use by implementing a reader and then passing it
 * to the parser.  A client can, at any time, make a new reader
 * active by pushing it on the Stack.  When the current context
 * (object or array) ends, the end...() call will be made on
 * every reader that's been pushed while that context has
 * been active.
 *
 * NOTE: All QStrings that come to this class are only safe to
 * read while in the given function (for performance they
 * directly reference the source stream).  The one exception
 * is the value in jsonString(): Since it's assumed most clients
 * will simply assign it locally, this is pre-copied.
 ******************************************************************/
class Reader
{
public:
	~Reader();

	// Optionally answer a new reader for the object.  Key may or
	// may not be empty.
	virtual void		startObject(Stack&, const QString&)			{ }
	virtual void		endObject(Stack&)							{ }

	// Optionally answer a new reader for the array.  Key may or
	// may not be empty.
	virtual void		startArray(Stack&, const QString&)			{ }
	virtual void		endArray(Stack&)							{ }

	// The key will only be valid if being read inside an object.
	virtual void		jsonString(Stack&, const QString&, const QString&)	{ }
	virtual void		jsonFloat(Stack&, const QString&, const float)		{ }
	// No intValue() because it can be confusing for clients -- for example,
	// the client could write a float value of "0.0" which could get read
	// back as an int of "0."  We could say that the file format requires
	// all floats have a trailing ".0" or something, but it's a human-editable
	// format, so there are no guarantees, so we'll just avoid any possible issues.

	// The client is responsible for managing any allocated Readers.
	// This indicates the framework is done with the reader, and can be deleted.
	virtual void		jsonPop()									{ }

protected:
	Reader();
	Reader(const Reader&);
};

} // namespace json

#endif // JSON_READER_H
