#include "json_writer.h"

#include <QDebug>
#include <QIODevice>

using namespace json;

static const char*			OBJECT_START = "{";
static const qint64			OBJECT_START_LENGTH = strlen(OBJECT_START);
static const char*			OBJECT_END = "}";
static const qint64			OBJECT_END_LENGTH = strlen(OBJECT_END);

static const char*			ARRAY_START = "[";
static const qint64			ARRAY_START_LENGTH = strlen(ARRAY_START);
static const char*			ARRAY_END = "]";
static const qint64			ARRAY_END_LENGTH = strlen(ARRAY_END);

static const char*			VALUE_SEP = ",";
static const qint64			VALUE_SEP_LENGTH = strlen(VALUE_SEP);

static const char*			KEY_SEP = ":";
static const qint64			KEY_SEP_LENGTH = strlen(KEY_SEP);

static const char*			NEWLINE = "\n";
static const qint64			NEWLINE_LENGTH = strlen(NEWLINE);
static const char*			TAB = "\t";
static const qint64			TAB_LENGTH = strlen(TAB);
static const char*			SPACE = " ";
static const qint64			SPACE_LENGTH = strlen(SPACE);

static const int			OBJECT_SEP_START_IDX = 0;
static const int			OBJECT_SEP_IDX = 1;
static const int			ARRAY_SEP_START_IDX = 2;
static const int			ARRAY_SEP_IDX = 3;
static const int			KEY_SEP_IDX = 4;

static const qint8			GET_ERR = -1;

/* JSON::WRITER
 ******************************************************************/
// Single argument constructor is below the NoFormat formatter

Writer::Writer(QIODevice* output, Formatter& fmt)
	: mStream(output)
	, mFormatter(fmt)
{
	init();
}

int Writer::getError() const
{
	return mError;
}

void Writer::beginObject()
{
	handleSep();
	mFormatter.beginObject(*this);
	push(OBJECT_SEP_START_IDX);
}

void Writer::endObject()
{
	mFormatter.endObject(*this);
	mDepth--;
}

void Writer::beginArray()
{
	handleSep();
	mFormatter.beginArray(*this);
	push(ARRAY_SEP_START_IDX);
}

void Writer::endArray()
{
	mFormatter.endArray(*this);
	mDepth--;
}

void Writer::key(const QString& str)
{
	// Keys are only valid when we're in an object.
	if (!isInObject()) return;

	handleSep();
	write(str);
	push(KEY_SEP_IDX);
}

void Writer::stringValue(const QString& str)
{
	handleSep();
	write(str);
}

void Writer::intValue(const int i)
{
	handleSep();
	mStream << i;
	if (mStream.status() != QTextStream::Ok) mError = ERROR;
}

void Writer::floatValue(const float f)
{
	handleSep();
	mStream << f;
	if (mStream.status() != QTextStream::Ok) mError = ERROR;
}

void Writer::push(const qint8 sep)
{
	if (!mSep.setAt(++mDepth, sep)) mError = ERROR;
}

void Writer::pop()
{
	--mDepth;
}

bool Writer::isInObject() const
{
	const qint8		v = mSep.getAt(mDepth, GET_ERR);
	return (v == OBJECT_SEP_START_IDX || v == OBJECT_SEP_IDX);
}

void Writer::handleSep()
{
	const qint8		sep = mSep.getAt(mDepth, GET_ERR);
	if (sep <= GET_ERR) return;

	if (sep == OBJECT_SEP_START_IDX) mSep.setAt(mDepth, OBJECT_SEP_IDX);
	else if (sep == OBJECT_SEP_IDX) mFormatter.valueSeparator(*this);
	else if (sep == ARRAY_SEP_START_IDX) mSep.setAt(mDepth, ARRAY_SEP_IDX);
	else if (sep == ARRAY_SEP_IDX) mFormatter.valueSeparator(*this);
	else if (sep == KEY_SEP_IDX) {
		mFormatter.keySeparator(*this);
		// KeyValue separator only gets applied once, so as soon as it's
		// used, pop it.
		pop();
	}
}

void Writer::write(const QString& str)
{
	mStream << "\"";
	mStream << str;
	mStream << "\"";
	if (mStream.status() != QTextStream::Ok) mError = ERROR;
}

void Writer::write(const char* data, const qint64)
{
	mStream << data;
	if (mStream.status() != QTextStream::Ok) mError = ERROR;
}

void Writer::init()
{
	mError = OK;
	mDepth = 0;
	mStream.setStatus(QTextStream::Ok);
}

/* JSON::FORMATTER
 * Control how the output is printed.  A basic formatter is
 * required to at least print the object/array tokens.
 ******************************************************************/
Formatter::Formatter()
{
}

Formatter::~Formatter()
{
}

/* JSON::PRETTY-FORMAT
 * Format with new lines and tabs between objects/arrays.
 ******************************************************************/
PrettyFormat::PrettyFormat()
	: mDepth(0)
{
}

void PrettyFormat::beginObject(Writer& w)
{
	push(w, OBJECT_START, OBJECT_START_LENGTH);
}

void PrettyFormat::endObject(Writer& w)
{
	pop(w, OBJECT_END, OBJECT_END_LENGTH);
}

void PrettyFormat::beginArray(Writer& w)
{
	push(w, ARRAY_START, ARRAY_START_LENGTH);
}

void PrettyFormat::endArray(Writer& w)
{
	pop(w, ARRAY_END, ARRAY_END_LENGTH);
}

void PrettyFormat::keySeparator(Writer& w)
{
	w.write(KEY_SEP, KEY_SEP_LENGTH);
	w.write(SPACE, SPACE_LENGTH);
}

void PrettyFormat::valueSeparator(Writer& w)
{
	w.write(VALUE_SEP, VALUE_SEP_LENGTH);
	w.write(SPACE, SPACE_LENGTH);
}

void PrettyFormat::push(Writer& w, const char* data, const qint64 len)
{
	w.write(data, len);
	w.write(NEWLINE, NEWLINE_LENGTH);
	mDepth++;
	for (int k=0; k<mDepth; k++) w.write(TAB, TAB_LENGTH);
}

void PrettyFormat::pop(Writer& w, const char* data, const qint64 len)
{
	w.write(NEWLINE, NEWLINE_LENGTH);
	mDepth--;
	for (int k=0; k<mDepth; k++) w.write(TAB, TAB_LENGTH);
	w.write(data, len);
}

/* JSON::NO-FORMAT
 * Class can't have any state, since there's a single global
 * instance that can be used by anyone.
 ******************************************************************/
class NoFormat : public Formatter
{
public:
	NoFormat() { }

	virtual void	beginObject(Writer& w	)	{ w.write(OBJECT_START, OBJECT_START_LENGTH); }
	virtual void	endObject(Writer& w)		{ w.write(OBJECT_END, OBJECT_END_LENGTH); }
	virtual void	beginArray(Writer& w)		{ w.write(ARRAY_START, ARRAY_START_LENGTH); }
	virtual void	endArray(Writer& w)			{ w.write(ARRAY_END, ARRAY_END_LENGTH); }
	virtual void	keySeparator(Writer& w)		{ w.write(KEY_SEP, KEY_SEP_LENGTH); }
	virtual void	valueSeparator(Writer& w)	{ w.write(VALUE_SEP, VALUE_SEP_LENGTH); }
};

static NoFormat		NO_FORMAT;

/* JSON::WRITER
 ******************************************************************/
Writer::Writer(QIODevice* output)
	: mStream(output)
	, mFormatter(NO_FORMAT)
{
	init();
}
