#ifndef JSON_WRITER_H
#define JSON_WRITER_H

#include <QString>
#include <QTextStream>
#include "collections/byteset.h"

class QIODevice;

namespace json {

class Formatter;

/* JSON::WRITER
 * Write data to a JSON-formatted stream.
 * For an explanation of JSON, see: http://www.json.org/
 ******************************************************************/
class Writer
{
public:
	static const int	ERROR = -1;
	static const int	OK = 0;

public:
	class Client {
	public:
		Client() { }
		virtual ~Client() { }

		virtual void	writeJson(json::Writer&) const = 0;
	};

public:
	explicit Writer(QIODevice* output);
	// Optionally supply a formatting object to determine how
	// white space is printed (without it, there will be none).
	Writer(QIODevice* output, Formatter&);

	int					getError() const;

	void				beginObject();
	void				endObject();
	void				beginArray();
	void				endArray();

	void				key(const QString& str);
	void				stringValue(const QString& str);
	void				intValue(const int i);
	void				floatValue(const float f);

public:					// For the formatter, other clients should go through the standard interface
	void				write(const QString& str);
	void				write(const char* data, const qint64 len);

private:
	QTextStream			mStream;
	Formatter&			mFormatter;
	int					mError;
	int					mDepth;
	// Keep a stack of the current separator state.  That is,
	// what character (if any) to write before the current command.
	ByteSet				mSep;

	void				push(const qint8 sep);
	void				pop();

	bool				isInObject() const;

	void				handleSep();

	void				init();
};

/* JSON::FORMATTER
 * Control how the output is printed.  A basic formatter is
 * required to at least print the object/array tokens.
 ******************************************************************/
class Formatter {
public:
	virtual ~Formatter();

	virtual void	beginObject(Writer&) = 0;
	virtual void	endObject(Writer&) = 0;
	virtual void	beginArray(Writer&) = 0;
	virtual void	endArray(Writer&) = 0;
	virtual void	keySeparator(Writer&) = 0;
	virtual void	valueSeparator(Writer&) = 0;

protected:
	Formatter();
};

/* JSON::PRETTY-FORMAT
 * Format with new lines and tabs between objects/arrays.
 ******************************************************************/
class PrettyFormat : public Formatter {
public:
	PrettyFormat();

	virtual void	beginObject(Writer&);
	virtual void	endObject(Writer&);
	virtual void	beginArray(Writer&);
	virtual void	endArray(Writer&);
	virtual void	keySeparator(Writer&);
	virtual void	valueSeparator(Writer&);

private:
	int				mDepth;

	void			push(Writer&, const char* data, const qint64 len);
	void			pop(Writer&, const char* data, const qint64 len);
};

} // namespace json

#endif // JSON_WRITER_H
