#include "cppolygon.h"

#include <assert.h>
#include <math.h>

static void get_dimen(const QPolygonF& p, int& l, int& t, int& r, int& b);

static int gCellCount = 0;
static int gLeafCount = 0;

/* CP-POLYGON
 ******************************************************************/
CpPolygon::CpPolygon()
{
}

void CpPolygon::setTo(const QPolygonF& p)
{
	// Prepare
	mSegs.clear();
	mRoot.clear();
	if (p.size() < 1) return;

	// Convert every segment into a leaf with a frame.
	Segment3					seg;
	for (int k=1; k<p.size(); k++) {
		const QPointF*		prev = &(p[k-1]);
		seg.set(*prev, p[k]);
		mSegs.pushFront(seg);
	}
	// Recursively generate the search tree.
	int						l, t, r, b;
	get_dimen(p, l, t, r, b);
	mRoot.generate(l, t, r, b, mSegs);
}

float CpPolygon::distanceTo(const int x, const int y) const
{
	// Remember when we hit a stop condition to use
	// the parent, to make sure I account for any neighbors.
	const Cell*			prev = &mRoot;
	const Cell*			next = mRoot.findNext(x, y);
	while (next != NULL) {
		if (next->mChildren.isEmpty()) return prev->distanceTo(x, y);
		const Cell*		n = next->findNext(x, y);
		// Remember to use the parent.
		if (n == NULL) break;
		prev = next;
		next = n;
	}
	return prev->distanceTo(x, y);
}

/* CP-POLYGON::SEG
 ******************************************************************/
CpPolygon::Seg::Seg(const Segment3& o)
	: mNext(NULL)
	, mSegment(o)
{
	QRectF		f;
	o.getFrame(f);
	mF.setCoords((int)floorf(f.left()), (int)floorf(f.top()), (int)ceilf(f.right()), (int)ceilf(f.bottom()));
}

/* CP-POLYGON::SEG-LIST
 ******************************************************************/
CpPolygon::SegList::SegList()
	: mHead(NULL)
{
}

CpPolygon::SegList::~SegList()
{
	clear();
}

void CpPolygon::SegList::pushFront(const Segment3& s)
{
	Seg*		n = new Seg(s);
	if (n != NULL) {
		n->mNext = mHead;
		mHead = n;
	}
}

void CpPolygon::SegList::clear()
{
	Seg*			n = mHead;
	while (n != NULL) {
		Seg*		next = n->mNext;
		n->mNext = NULL;
		delete n;
		n = next;
	}
	mHead = NULL;
}

/* CP-POLYGON::LEAF
 ******************************************************************/
CpPolygon::Leaf::Leaf(const Seg* s)
	: mNext(NULL)
	, mSeg(s)
{
	gLeafCount++;
}

CpPolygon::Leaf* CpPolygon::Leaf::tail()
{
	Leaf*			l = this;
	while (l->mNext != NULL) l = l->mNext;
	return l;
}

/* CP-POLYGON::LEAF-LIST
 ******************************************************************/
CpPolygon::LeafList::LeafList()
	: mHead(NULL)
{
}

CpPolygon::LeafList::~LeafList()
{
	clear();
}

void CpPolygon::LeafList::pushFront(const Seg* s)
{
	if (s == NULL) return;
	Leaf*			n = new Leaf(s);
	if (n != NULL) {
		n->mNext = mHead;
		mHead = n;
	}
}

void CpPolygon::LeafList::addAll(const SegList& segs)
{
	const Seg*		s = segs.mHead;
	while (s != NULL) {
		pushFront(s);
		s = s->mNext;
	}
}

void CpPolygon::LeafList::takeAll(LeafList& o)
{
	if (o.mHead == NULL) return;
	if (mHead == NULL) mHead = o.mHead;
	else mHead->tail()->mNext = o.mHead;
	o.mHead = NULL;
}

void CpPolygon::LeafList::clear()
{
	Leaf*				n = mHead;
	while (n != NULL) {
		Leaf*			next = n->mNext;
		n->mNext = NULL;
		delete n;
		n = next;
	}
	mHead = NULL;
}

void CpPolygon::LeafList::print() const
{
	const Leaf*		l = mHead;
	while (l != NULL) {
		qDebug() << "\tleaf=" << l->mSeg->mSegment;
		l = l->mNext;
	}
}

/* CP-POLYGON::CELL-LIST
 ******************************************************************/
CpPolygon::CellList::CellList()
	: mHead(NULL)
{
}

CpPolygon::CellList::~CellList()
{
	clear();
}

void CpPolygon::CellList::pushFront(Cell* n)
{
	if (n != NULL) {
		n->mNext = mHead;
		mHead = n;
	}
}

void CpPolygon::CellList::clear()
{
	Cell*			n = mHead;
	while (n != NULL) {
		Cell*		next = n->mNext;
		n->mNext = NULL;
		delete n;
		n = next;
	}
	mHead = NULL;
}

bool CpPolygon::CellList::generate(	const int l, const int t, const int r, const int b,
									const SegList& segs)
{
	QRect				f;
	f.setCoords(l, t, r, b);
	const Seg*			seg = segs.mHead;
	LeafList			leaves;
	while (seg != NULL) {
		if (f.intersects(seg->mF)) leaves.pushFront(seg);
		seg = seg->mNext;
	}
	if (leaves.isEmpty()) return false;
	Cell*				c = new Cell();
	if (c == NULL) return false;
	c->generate(l, t, r, b, segs);
	pushFront(c);
	return true;
}

/* CP-POLYGON::CELL
 ******************************************************************/
CpPolygon::Cell::Cell()
	: mNext(NULL)
{
	gCellCount++;
}

void CpPolygon::Cell::clear()
{
	mChildren.clear();
	mLeaves.clear();
}

const CpPolygon::Cell* CpPolygon::Cell::findNext(const int x, const int y) const
{
	if (!mF.contains(x, y)) return NULL;
	const Cell*		c = mChildren.mHead;
	while (c != NULL) {
		if (c->mF.contains(x, y)) return c;
		c = c->mNext;
	}
	return c;
}

float CpPolygon::Cell::distanceTo(const int x, const int y) const
{
	if (mLeaves.isEmpty()) return 0;
	const Leaf*			l = mLeaves.mHead;
	Point3				pt(x, y, 0);
	float				min = l->mSeg->mSegment.distanceToSquared(pt);
//qDebug() << "min=" << min;
	while (l != NULL) {
		const float		d = l->mSeg->mSegment.distanceToSquared(pt);
//		qDebug() << "\td=" << d;
		if (d < min) min = d;
		l = l->mNext;
	}
//qDebug() << "ans=" << sqrt(min);
	return sqrt(min);
}

void CpPolygon::Cell::generate(		const int l, const int t, const int r, const int b,
									const SegList& segs)
{
	mF.setCoords(l, t, r, b);
//qDebug() << "generateFor " << mF;
	const int			MIN_SIZE = 20;
	if ((r-l) < MIN_SIZE || (b-t) < MIN_SIZE) {
		mLeaves.addAll(segs);
		return;
	}
	const int			cenX = (l + r) / 2, cenY = (t + b) / 2;
	bool				needsLeaves = false;
	if (!mChildren.generate(l, t, cenX, cenY, segs)) needsLeaves = true;
	if (!mChildren.generate(cenX, t, r, cenY, segs)) needsLeaves = true;
	if (!mChildren.generate(cenX, cenY, r, b, segs)) needsLeaves = true;
	if (!mChildren.generate(l, cenY, cenX, b, segs)) needsLeaves = true;

//if (needsLeaves) qDebug() << "\t need leaves for " << mF;

	// Store every seg that intersected me.  Whenever a client reaches
	// here, they should not have to gather up segs spread around the tree.
//	if (needsLeaves) mLeaves.addAll(segs);
	mLeaves.addAll(segs);
}

/* misc
 ******************************************************************/
static void get_dimen(const QPolygonF& p, int& l, int& t, int& r, int& b)
{
	l = r = (int)(p[0].x());
	t = b = (int)(p[0].y());
	for (int k=1; k<p.size(); k++) {
		const int		x = (int)(p[k].x()),
						y = (int)(p[k].y());
		if (x < l) l = x;
		else if (x > r) r = x;
		if (y < t) t = y;
		else if (y > b) b = y;
	}
	// expand slightly to account for any rounding errors.
	l--;
	t--;
	r++;
	b++;
}
