#ifndef CPPOLYGON_H
#define CPPOLYGON_H

#include <QPolygonF>
#include <QRect>
#include "math/primitives/segment3.h"

/* CP-POLYGON
 * A cell partitioned polygon, used for quickly finding the
 * distance from any given pixel to the edge of the polygon.
 * It works like this:  The area of the polygon is divided into
 * arbitrarily-sized cells, which are then clustered together at
 * successively lower resolutions until one cell coveres the
 * entire area.  From that point, a single pixel input can
 * traverse down to finer grained cells until it finds all segment
 * intersections for a given cell.
 ******************************************************************/
class CpPolygon
{
public:
    CpPolygon();

	void			setTo(const QPolygonF&);

	// Measure how far the supplied point is to the edge of the polygon.
	// This does not tell you inside/outside.
	float			distanceTo(const int x, const int y) const;

/* Yeah, the class looks complicated.  It's nearly all custom container
 * classes.  Since this is such a key performance item, there's as little
 * overhead as possible.
 */
private:
	// forward declarations
	class Cell;

	// SEG.  Store a single segment.
	class Seg {
	public:
		Seg(const Segment3&);

		Seg*				mNext;		// Only used by the seg list.  Weird.
		const Segment3		mSegment;
		QRect				mF;
	};

	// SEG-LIST.  Seg container.
	class SegList {
	public:
		SegList();
		~SegList();

		Seg*				mHead;

		void				pushFront(const Segment3&);
		void				clear();
	};

	// LEAF.  Store a single seg (which I don't own), chainable.
	class Leaf {
	public:
		Leaf(const Seg*);

		Leaf*				mNext;		// Only used by the list.  Kinda weird.
		const Seg*			mSeg;

		Leaf*				tail();
	};

	// LEAF-LIST.  Store a list of leaves.
	class LeafList {
	public:
		LeafList();
		~LeafList();

		Leaf*				mHead;

		bool				isEmpty() const;

		void				pushFront(const Seg*);

		void				addAll(const SegList&);
		// The list arg is emptied, all leaves added to me.
		void				takeAll(LeafList&);

		void				clear();

		void				print() const;
	};

	// CELL-LIST.  Container for cells.
	class CellList {
	public:
		CellList();
		~CellList();

		Cell*			mHead;

		bool			isEmpty() const;
		void			clear();
		void			pushFront(Cell*);

		bool			generate(	const int l, const int t, const int r, const int b,
									const SegList&);
	};

	// CELL.
	class Cell {
	public:
		Cell();

		Cell*			mNext;		// only used by the list.  Kinda weird.
		// A cell can either have children or leaves.
		CellList		mChildren;
		LeafList		mLeaves;
		QRect			mF;			// frame

		void			clear();

		// Find one step of the coordinate, either myself or an immediate parent (or NULL).
		const Cell*		findNext(const int x, const int y) const;

		float			distanceTo(const int x, const int y) const;

		void			generate(	const int l, const int t, const int r, const int b,
									const SegList&);
	};

	// The leaves is the only owner of the segments.  Everything
	// else just references them.
	SegList					mSegs;
	Cell					mRoot;
//	intersection::SegQuad	mSq;

//	void					getIntersections(LeafHead& dst, const Quad& q);
	// Given the source cell list, create root cells with a copy + expansion
//	void					expandLeaves(const Cell*, const int w, const int h);
};

/* inline
 ******************************************************************/
inline bool CpPolygon::LeafList::isEmpty() const
{
	return mHead == NULL;
}

inline bool CpPolygon::CellList::isEmpty() const
{
	return mHead == NULL;
}

#endif // CPPOLYGON_H
