#include "geo.h"

#include <QtGlobal>
#include <QRectF>

#include "math.h"
#include "math/primitives/point3.h"

using namespace geo;

const double			geo::EPS = 0.00000000001;
const double			geo::DTOR = 0.017453;
const double			geo::RTOD = 57.2957795;

const int geo::iroundf(const double x)
{
	return (int)((x > 0.0) ? floorf(x + 0.5) : ceilf(x - 0.5));
}

double geo::slope(const double x0, const double y0, const double x1, const double y1)
{
	return (y1 - y0) / (x1 - x0);
}

double geo::perpendicularSlope(const double x0, const double y0, const double x1, const double y1) {
	return -((x1-x0) / (y1-y0));
}

double geo::dist(const double x0, const double y0, const double x1, const double y1) {
	return sqrt(((x0-x1)*(x0-x1)) + ((y0-y1)*(y0-y1)));
}

double geo::distSquared(const double x0, const double y0, const double x1, const double y1)
{
	return ((x0-x1)*(x0-x1)) + ((y0-y1)*(y0-y1));
}

void geo::lineIntersection(	const Point3& a1, const Point3& a2,
							const Point3& b1, const Point3& b2,
							Point3& out) {
	lineIntersection(a1.x, a1.y, a2.x, a2.y, b1.x, b1.y, b2.x, b2.y, out);
}

void geo::lineIntersection(	const double a1x, const double a1y,
							const double a2x, const double a2y,
							const double b1x, const double b1y,
							const double b2x, const double b2y,
							Point3& out) {
	const double			A1 = a2y - a1y,
							B1 = a1x - a2x;
	const double			C1 = (A1*a1x)+(B1*a1y);
	const double			A2 = b2y - b1y,
							B2 = b1x - b2x;
	const double			C2 = (A2*b1x)+(B2*b1y);
	// Do the algorithm
	const double			det = (A1*B2) - (A2*B1);
	if (det == 0) {
		// Parallel
		out.x = a2x;
		out.y = a2y;
	} else {
		out.x = (B2*C1 - B1*C2)/det;
		out.y = (A1*C2 - A2*C1)/det;
	}
}

bool geo::lineSegmentIntersection(	const Point3& a1, const Point3& a2,
									const Point3& b1, const Point3& b2,
									Point3* out) {
	return lineSegmentIntersection(a1.x, a1.y, a2.x, a2.y, b1.x, b1.y, b2.x, b2.y, out);
}

bool geo::lineSegmentIntersection(	const double x1, const double y1,
									const double x2, const double y2,
									const double x3, const double y3,
									const double x4, const double y4,
									Point3* out) {
	const double		bx = x2 - x1,
						by = y2 - y1,
						dx = x4 - x3,
						dy = y4 - y3;
	const double		b_dot_d_perp = bx * dy - by * dx;
	if (b_dot_d_perp == 0) {
		return false;
	}
	const double		cx = x3 - x1,
						cy = y3 - y1;
	const double		t = (cx * dy - cy * dx) / b_dot_d_perp;
	if (t < 0 || t > 1) {
		return false;
	}
	const double		u = (cx * by - cy * bx) / b_dot_d_perp;
	if (u < 0 || u > 1) {
		return false;
	}
	if (out != NULL) {
		out->x = x1+t*bx;
		out->y = y1+t*by;
	}
	return true;
}

void geo::getFrame(const Point3* pts, const int size, QRectF& f)
{
	if (size < 1) return;
	double			l = pts[0].x, t = pts[0].y, r = pts[0].x, b = pts[0].y;
	for (int k=1; k<size; k++) {
		if (pts[k].x < l) l = pts[k].x;
		else if (pts[k].x > r) r = pts[k].x;
		if (pts[k].y < t) t = pts[k].y;
		else if (pts[k].y > b) b = pts[k].y;
	}
	f.setCoords(l, t, r, b);
}

double geo::areaOfTri(	const double ax, const double ay,
						const double bx, const double by,
						const double cx, const double cy)
{
	const double		a = (ax - cx),
						b = (ay - cy),
						c = (bx - cx),
						d = (by - cy);
	return 0.5 * geo::abs((a*d) - (b*c));
}

// Technically a function on a triangle, if we had one.
// This works by taking the given point, and converting the
// triangle into three triangles with the new point, then
// comparing the area of that with the area of the triangle.
bool geo::isInsideTri(	const double ax, const double ay,
						const double bx, const double by,
						const double cx, const double cy,
						const double x, const double y)
{
	const double		t = areaOfTri(ax, ay, bx, by, cx, cy);
	double				sum = areaOfTri(ax, ay, bx, by, x, y);
	if (sum > t) return false;

	sum += areaOfTri(bx, by, cx, cy, x, y);
	if (sum > t) return false;

	sum += areaOfTri(ax, ay, cx, cy, x, y);
	// Oh man...  there's some sort of small error
	// to account for.  Where does that come from?
	// And WHOA, running under QT on a mac, things are
	// MUCH less precise then they were under android.
	// What's the right amount of fudge to check for?
	return sum-0.01 < t;
}

double geo::degree(const double x2, const double y2) {
	if (y2 == 0) {
		if (x2 > 0) return 0.0;
		else if (x2 < 0) return 180.0;
		else return 0;
	}
	if (x2 == 0) {
		if (y2 > 0) return 90.0;
		else if (y2 < 0) return 270.0;
	}

	const double		x1 = 0, y1 = 0;
	const double		m1 = 0;
	const double		m2 = geo::slope(x1, y1, x2, y2);
	const double		tangent = (m2 - m1) / (1 + (m1 * m2));
	const double		radians = atan(tangent);
	const double		degree = radians * RTOD;

	if (x2 < 0 && y2 > 0) return degree + 180;
	else if (x2 < 0 && y2 < 0) return degree + 180;
	else if (x2 > 0 && y2 < 0) return degree + 360;

	return degree;
}

void geo::newPtOnLine(	const double x, const double y, const double  m, const double dist,
						Point3& newPt) {
	if (m == 0) {
		newPt.x = x + dist;
		newPt.y = y;
		return;
	} else if (!isfinite(m)) {
//	} else if (Double.isInfinite(m)) {
		newPt.x = x;
		newPt.y = y + dist;
		return;
	}

	/* This is basically a reverse engineering of the distance formula
	 * 		dist = sqrt(SQR(x2 - x1) + SQR(y2 - 1))
	 * It works by taking the final result -- the requested new distance,
	 * and deconstructing it into the xPart (SQR(x2 - x1) and the yPart,
	 * the sqrt of which can be added to the x and y values.
	 *
	 * The only trick is finding the xPart, which is just solved via
	 * the equation x + (x + (slope * slope)) = newDistance.
	 */
	const double		xPart = ((dist*dist) * (1 / (m * m))) / ( (1 / (m * m)) + 1);
	const double		yPart = xPart * (m * m);

	// Hmmm... just guessing...
	if (m < 0) {
		if (dist < 0) {
			newPt.x = x + sqrt(xPart);
			newPt.y = y - sqrt(yPart);
		} else {
			newPt.x = x - sqrt(xPart);
			newPt.y = y + sqrt(yPart);
		}
	} else {
		if (dist < 0) {
			newPt.x = x - sqrt(xPart);
			newPt.y = y - sqrt(yPart);
		} else {
			newPt.x = x + sqrt(xPart);
			newPt.y = y + sqrt(yPart);
		}
	}
}

bool geo::polygonContains(	const Point3* pts, const int size,
							const double x, const double y)
{
	int					k, j;
	bool				ans = false;
	for (k = 0, j = size-1; k < size; j = k++) {
		const Point3&	pk = pts[k],
						pj = pts[j];
		if (( ((pk.y <= y) && (y < pj.y))
				|| ((pj.y <= y) && (y < pk.y)))
				&& (x < (pj.x - pk.x) * (y - pk.y) / (pj.y - pk.y) + pk.x))
			ans = !ans;
	}
	return ans;
}

void geo::getBarycenter(	const Point3* pts, const int size,
							Point3& bary)
{
	bary.x = 0;
	bary.y = 0;
	bary.z = 0;
	for (int k=0; k<size; k++) {
		bary.x += pts[k].x;
		bary.y += pts[k].y;
		bary.z += pts[k].z;
	}
	if (size > 1) {
		const double		d = size;
		bary.x /= d;
		bary.y /= d;
		bary.z /= d;
	}
}
