#ifndef GEO_H
#define GEO_H

class Point3;
class QRectF;

/* GEO
 * A collection of geometry functions.
 ********************************************/

// Convert an x,y pixel value into an index in a flat array
#define IDX_FOR_PIXEL(x, y, w)		(((y)*(w))+(x))

namespace geo {

extern const double		EPS;		// a very small number
extern const double		DTOR;		// degrees to radians
extern const double		RTOD;		// radians to degrees

// Simple int from float rounding (nearest)
const int iroundf(const double x);

// Custom abs func because god help me, for some reason the system library
// is truncating small values (actually, not-that-small-of-values) to 0.
inline const double abs(const double v) { return v < 0 ? -v : v; }

double slope(const double x0, const double y0, const double x1, const double y1);
double perpendicularSlope(const double x0, const double y0, const double x1, const double y1);

double dist(const double x0, const double y0, const double x1, const double y1);
// Ignore that slow square root when we're only comparing
// relative distances.
double distSquared(const double x0, const double y0, const double x1, const double y1);

void lineIntersection(	const Point3& a1, const Point3& a2,
						const Point3& b1, const Point3& b2,
						Point3& out);
void lineIntersection(	const double a1x, const double a1y,
						const double a2x, const double a2y,
						const double b1x, const double b1y,
						const double b2x, const double b2y,
						Point3& out);

bool lineSegmentIntersection(	const Point3& a1, const Point3& a2,
								const Point3& b1, const Point3& b2,
								Point3* out);
bool lineSegmentIntersection(	const double x1, const double y1,
								const double x2, const double y2,
								const double x3, const double y3,
								const double x4, const double y4,
								Point3* out);

// Given a point array, answer it's bounding rectangle.
void getFrame(const Point3* pts, const int size, QRectF& f);

double areaOfTri(	const double ax, const double ay,
					const double bx, const double by,
					const double cx, const double cy);
bool isInsideTri(	const double ax, const double ay,
					const double bx, const double by,
					const double cx, const double cy,
					const double x, const double y);

// 0 = right, clockwise (90 = down)
double degree(const double x2, const double y2);

// Given a line in point-slope form, find a new point on
// that line that's dist away.
void newPtOnLine(	const double x, const double y, const double  m, const double dist,
					Point3& newPt);

bool polygonContains(	const Point3* pts, const int size,
						const double x, const double y);

// Given a polygon defined by pts and size, answer the barycenter.
void getBarycenter(const Point3* pts, const int size, Point3& bary);

// Utility for standard sorting of two doubles, where < is -1, = is 0, > is 1
inline int sort_doubles(const double d1, const double d2) {
	return (d1 == d2 ? 0 : (d1 < d2 ? -1 : 1));
}

} // namespace geo

// ************************************************************************
// ******************************* SHAPING ********************************
// ************************************************************************

// Given an x of 0 to 1, answer a value of 0 to 1
// that's smoothed towards the edges -- a standard S curve.
#define GEO_S_CURVE(x)		((x) * (x) * (3.0 - 2.0 * (x)))

#endif // GEO_H
