#include "mathutils.h"

#include "math/geo.h"
#include "math/primitives/polygon.h"

using namespace math;

/* MATH::ORDER
 ******************************************************************/
Order::Order()
{
}

void Order::makeCcw(Point3* pts, const int size)
{
	if (!verify(pts, size)) return;
	cacheDegrees(pts, size);

	// Sort the degrees, highest to lowest
	Point3			pt;
	double			d;
	double*			da = mD.get(size);
	bool			more = true;
	while (more) {
		more = false;
		for (int k = 1; k < size; k++) {
			if (da[k] > da[k-1]) {
				d = da[k-1];	da[k-1] = da[k];	da[k] = d;
				pt = pts[k-1];	pts[k-1] = pts[k];	pts[k] = pt;
				more = true;
			}
		}
	}
}

bool Order::verify(const Point3* pts, const int size)
{
	if (pts == NULL || size < 2) return false;
	return 	mD.get(size) != NULL;
}

void Order::cacheDegrees(const Point3* pts, const int size)
{
	// Find the barycenter
	double			baryX = 0,
					baryY = 0;
	for (int k=0; k<size; k++) {
		baryX += pts[k].x;
		baryY += pts[k].y;
	}
	baryX /= (double)size;
	baryY /= (double)size;
	// Find each point's degree
	double*			d = mD.get(size);
	for (int k=0; k<size; k++) {
		d[k] = geo::degree(pts[k].x - baryX, pts[k].y - baryY);
	}
}

/* MATH::NORMAL
 ******************************************************************/
bool Normal::setOn(const Polygon& p, Point3& nrm)
{
	const qint32		size = p.getSize();
	const Point3*		pts = p.getPoints();
	if (size < 3) return false;
	// Probably should ensure not using parallel lines, although
	// that would be an odd polygon if two segments sharing a
	// point were parallel.
	const Point3		&a = pts[0], &b = pts[1], &c = pts[2];
	mA = b - a;
	mB = c - b;
	nrm.asCross3(mA, mB);
	nrm.normalize();
	return true;
}

/* MATH::EDGE-ITERATOR
 ******************************************************************/
EdgeIterator::EdgeIterator(const Polygon& p)
	: mA(NULL)
	, mB(NULL)
	, mSize(p.getSize())
	, mPts(p.getPoints())
	, mIdx(0)
{
	if (mSize > 0) {
		mA = &(mPts[mSize-1]);
		mB = &(mPts[mIdx]);
	}
}

void EdgeIterator::operator++()
{
	mIdx++;
	if (mIdx < mSize) {
		mA = mB;
		mB = &(mPts[mIdx]);
	}
}

bool EdgeIterator::hasValue() const
{
	return (mSize > 1 && mIdx < mSize);
}
