#include "boolcutout.h"

#include <math.h>
#include "math/poly/bool.h"
#include "math/primitives/plane.h"
#include "math/tests/lineplane.h"

using namespace math;

static int clip(const int idx, const Polygon&);

/* MATH::BOOL-CUT-OUT
 ******************************************************************/
BoolCutOut::BoolCutOut()
{
}

void BoolCutOut::on(	const Polygon& a, const Polygon& b,
						Bool& op, PolygonList& results, PolyFilter& pf)
{
	if (!getFrame(a, mFrame)) return;
	if (!getBisector(a, b, mBisector)) return;
	// Using the bisector, find a
}

static void add_low_z(	const Point3* pts, const qint32 size,
						const Point3& bary, const Point3& low,
						Point3& ans, int idx)
{
	// Limit the index to the polygon range.
	if (idx < 0) idx = size-1;
	else if (idx >= size-1) idx = 0;
	const Point3&		pt = pts[idx];
	// If the new point is within range of the low, and closer to the
	// bary then the current answer, replace the answer.
	if (pt.sameAs3(low, 0.00001) && pt.dist3Squared(bary) < ans.dist3Squared(bary)) {
		ans.set(pt);
	}
}

bool BoolCutOut::getBisector(const Polygon& a, const Polygon& b, Segment3& seg) const
{
	// The bisector is a line that goes through the barycenter of b to either
	// edge of the enclosing polygon, because why not?
	Point3			baryB;
	b.getBarycenter(baryB);
	// Find the point to use as the basis for the segment that does bisection.
	// I'll use the pt with the lowest Z.
	const Point3*	pts = a.getPoints();
	const qint32	size = a.getSize();
	if (size < 3) return false;
	int				lowIdx = 0;
	for (qint32 k=1; k<size; k++) {
		if (pts[k].z < pts[lowIdx].z) lowIdx = k;
	}
	// I have the lowest point, but it might be part of a segment(s) with
	// equally low z values.  If so, find the pt closest to the bary center.
	seg.mPt[0].set(pts[lowIdx]);
	add_low_z(pts, size, baryB, pts[lowIdx], seg.mPt[0], lowIdx-1);
	add_low_z(pts, size, baryB, pts[lowIdx], seg.mPt[0], lowIdx+1);
	// We've got part of our bisector, now extend it through b's bary center
	// and beyond the edge of poly a.
	return true;
}

bool BoolCutOut::getFrame(const Polygon& a, Polygon& f) const
{
	// Get a 3d frame for a.
	if (!a.getFrame(f) || f.getSize() != 8) return false;
	// Compress the frame to the plane of a.  This is a little dodgy -- we're
	// relying on the fact that the source material started off in 2d space and
	// was extruded along the z axis, and we're relying on the format of the cube
	// to have the low z pts (bottom side) in 0-3 and top in 4-7.  Should work
	// fine for the original purpose of running this in the beveling algorithm,
	// but it's not a general solution.
	LinePlane			lp;
	Plane				p;
	p.from(a);
	Point3*				pts = f.getPoints();
	for (int k=0; k<4; k++) {
		if (lp.on(pts[k], pts[k+4], p, &(pts[k])) < LinePlane::INTERSECT) return false;
	}
	f.setSize(4);
	return true;
}

bool BoolCutOut::on(	const Polygon& dst, const Polygon& src,
						Polygon& out0, Polygon& out1)
{
	out0.clear();
	out1.clear();
	// Find the starting and intersection points on each poly
	if (!getStarts(dst, src)) return false;
	// I've now split the combined poly into two, so add those
	// to the output.
	addPoly0(dst, src, out0);
	addPoly1(dst, src, out1);
	return true;
}

void BoolCutOut::addPoly0(	const Polygon& dst, const Polygon& src,
							Polygon& out)
{
	out.add(mHit[0].mPt);
	addPoly(dst, mHit[0].mDst, mHit[1].mDst, 1, out, true);
	out.add(mHit[1].mPt);
	addPoly(src, mHit[1].mSrc, mHit[0].mSrc, -1, out, true);
}

void BoolCutOut::addPoly1(	const Polygon& dst, const Polygon& src,
							Polygon& out)
{
	out.add(mHit[0].mPt);
	addPoly(dst, clip(mHit[0].mDst-1, dst), clip(mHit[1].mDst+0, dst), -1, out, false);
	out.add(mHit[1].mPt);
	addPoly(src, mHit[1].mSrc, mHit[0].mSrc, 1, out, true);
}

void BoolCutOut::addPoly(	const Polygon& p, const int start, const int end,
							const int delta, Polygon& out, const bool close)
{
	const Point3*	pts = p.getPoints();
	int				idx = start;
	do {
		out.add(pts[idx]);
		idx = clip(idx+delta, p);
	} while (idx != end);
	if (close && start != end) out.add(pts[end]);
}

bool BoolCutOut::getStarts(const Polygon& dst, const Polygon& src)
{
	const int				ssize = src.getSize();
	if (ssize < 2) return false;
	for (int k=0; k<HIT_SIZE; k++) mHit[k].clear();

	const Point3*			spts = src.getPoints();
	const int				dsize = dst.getSize();
	const Point3*			dpts = dst.getPoints();

	// Take one side of the src poly and find its intersections
	// with the dst.  This is not great -- any complex src polys
	// will self intersect and be all messed up, but it works for
	// what I'm using it for.
	const Point3&			a0 = spts[0];
	const Point3&			a1 = spts[1];
	const Point3&			prvB = dpts[dsize-1];
	for (int k=0; k<dsize; k++) {
		const Point3&		nxtB = dpts[k];
		double				d = mLL.on(a0, a1, prvB, nxtB);
		// Make sure the intersection point is on the line and
		// also on the actual segment.
		if (finite(d) && d < 0.1f && mPS.on(mLL.mNearest2, prvB, nxtB, mPS.mHit) < 0.1f) {
			// Get distance from intersection to src segment, and
			// replace the lowest stored value.
			d = mPS.on(mLL.mNearest2, a0, a1, mPS.mHit);
			const int		idx = highestD();
			if (idx >= 0 && (mHit[idx].mD < 0 || d < mHit[idx].mD)) {
				mHit[idx].mD = d;
				mHit[idx].mPt.set(mLL.mNearest2);
				mHit[idx].mDst = k;
			}
		}
	}
	for (int k=0; k<HIT_SIZE; k++) {
		if (mHit[k].mDst < 0) return false;
	}
	mHit[0].mSrc = 0;
	mHit[1].mSrc = 1;
	// I now have two intersection (hit) points, so select the
	// one closest to the relevent src point.
	const double			a0_d = mHit[0].mPt.dist3Squared(a0),
							a1_d = mHit[0].mPt.dist3Squared(a1);
	// If intersection at hit[0] is closest to src pt 0 then 1,
	// everything's fine, otherwise switch the hits.
	if (a1_d < a0_d) mHit[0].swap(mHit[1]);

	return true;
}

int BoolCutOut::highestD() const
{
	int			idx = -1;
	double		d = -1;
	for (int k=0; k<HIT_SIZE; k++) {
		if (mHit[k].mD < 0) return k;
		if (idx < 0 || mHit[k].mD > d) {
			idx = k;
			d = mHit[k].mD;
		}
	}
	return idx;
}

/* MATH::BOOL-CUT-OUT::HIT
 ******************************************************************/
void BoolCutOut::Hit::clear()
{
	mDst = -1;
	mSrc = -1;
	mD = -1;
}

void BoolCutOut::Hit::swap(Hit& o)
{
	if (this != &o) {
		const int		d = mDst;	mDst = o.mDst;	o.mDst = d;
		const int		s = mSrc;	mSrc = o.mSrc;	o.mSrc = s;
		mPt.swap(o.mPt);
		const double	td = mD;	mD = o.mD;		o.mD = td;
	}
}

/* misc
 ******************************************************************/
static int clip(const int idx, const Polygon& p)
{
	const int			psize = p.getSize();
	if (idx >= psize) return 0;
	if (idx < 0) return psize-1;
	return idx;
}

