#include "boolimg.h"

#include <math.h>
#include <QImage>
#include <QPainter>
#include "math/poly/steplist.h"

using namespace math;

static const float			PT_W = 8;
static const float			PT_H = 8;
static const float			ALIGN_W = PT_W + 4;
static const float			ALIGN_H = PT_H + 4;
static const float			ALIGN_OFF = 4;

/* BOOL-IMG
 ******************************************************************/
BoolImg::BoolImg()
{
	mBorderL = ALIGN_W + 2;
	mBorderT = ALIGN_H + 2;

	mBg_C.setRgb(200, 200, 200);
	mEdgeA_C.setRgb(0, 0, 0);
	mEdgeB_C.setRgb(0, 0, 0);
	mPtA_C.setRgb(0, 200, 0);
	mPtB_C.setRgb(0, 0, 200);
	mI_C.setRgb(200, 0, 0);
	mGood_C.setRgb(255, 255, 255);
	mEvil_C.setRgb(200, 150, 150);

	// Opacity range for edges
	mEdgeR.mMin = 0.75;
	mEdgeR.mMax = 0.25;
}

void BoolImg::dump(	const StepList& a, const StepList& b,
					const float scale, const QString& filename)
{
	QRectF			af, bf, f;
	getFrame(a, af);
	getFrame(b, bf);
	f = af.united(bf);
	const int		w = (int)((f.width() * scale) + mBorderL*2),
					h = (int)((f.height() * scale) + mBorderT*2);
	if (w < 1 || h < 1) return;

	mOrigin.setX(f.left());
	mOrigin.setY(f.top());
	mScale = scale;

	QImage			img(w, h, QImage::Format_ARGB32);
	{
		QPainter	p(&img);
		p.fillRect(0, 0, w, h, mBg_C);
		p.setRenderHint(QPainter::Antialiasing);
		drawEdges(a, mEdgeA_C, p);
		drawEdges(b, mEdgeB_C, p);
		drawPoints(a, mPtA_C, 5, p);
		drawPoints(b, mPtB_C, 2, p);
	}
	img.save(filename);
}

void BoolImg::drawEdges(	const StepList& a, const QColor& c,
							QPainter& p)
{
	mPen.setColor(c);
	mPen.setWidthF(1.5f);
	p.setBrush(Qt::NoBrush);
	p.setPen(mPen);

	Range				rng(0, a.mSize-1);
	QPointF				prv, nxt;
	translate(*(a.mStep[a.mSize-1].mPt), prv);
	p.setOpacity(rng.translateTo(a.mSize-1, mEdgeR));
	for (int k=0; k<a.mSize; k++) {
		translate(*(a.mStep[k].mPt), nxt);
		p.drawLine(prv, nxt);

		prv = nxt;
		p.setOpacity(rng.translateTo(k, mEdgeR));
	}
}

void BoolImg::drawPoints(	const StepList& a, const QColor& c,
							const float first, QPainter& p)
{
	mBrush.setStyle(Qt::SolidPattern);
	p.setOpacity(0.75f);

	QPointF				pt;
	QRectF				alignF;

	for (int k=0; k<a.mSize; k++) {
		const Step&		step = a.mStep[k];
		translate(*(step.mPt), pt);

		// Draw good/evil indicator
		const QColor*	alignment = NULL;
		if (step.mAlignment == Step::GOOD) {
			alignment = &mGood_C;
			alignF.setCoords(pt.x()-ALIGN_W, pt.y()-PT_H+ALIGN_OFF, pt.x()+ALIGN_W, pt.y()+PT_H-ALIGN_OFF);
		} else if (step.mAlignment == Step::EVIL) {
			alignment = &mEvil_C;
			alignF.setCoords(pt.x()-PT_W+ALIGN_OFF, pt.y()-ALIGN_H, pt.x()+PT_W-ALIGN_OFF, pt.y()+ALIGN_H);
		}
		if (alignment != NULL) {
			mBrush.setColor(*alignment);
			p.setBrush(mBrush);
			p.setPen(c);
			p.drawRoundedRect(alignF, 2, 2);
		}

		// Draw pt
		if (step.isIntersection()) mBrush.setColor(mI_C);
		else mBrush.setColor(c);
		p.setBrush(mBrush);
		p.setPen(Qt::NoPen);
		p.drawEllipse(pt, PT_W, PT_H);

		// Mark the first pt
		if (k == 0) {
			p.setBrush(Qt::NoBrush);
			p.setPen(c);
			p.drawEllipse(pt, PT_W+first, PT_H+first);
		}
	}
}

void BoolImg::getFrame(const StepList& a, QRectF& r) const
{
	r = QRectF();
	if (a.mSize < 1) return;
	r.setCoords(	(float)a.mStep[0].mPt->x, (float)a.mStep[0].mPt->y,
					(float)a.mStep[0].mPt->x, (float)a.mStep[0].mPt->y );
	for (int k=1; k<a.mSize; k++) {
		const float		x = (float)a.mStep[k].mPt->x,
						y = (float)a.mStep[k].mPt->y;
		if (x < r.left()) r.setLeft(x);
		else if (x > r.right()) r.setRight(x);
		if (y < r.bottom()) r.setBottom(y);
		else if (y > r.top()) r.setTop(y);
	}
}

void BoolImg::translate(const Point3& pt, QPointF& out) const
{
	out.setX(mBorderL + ((((float)pt.x) - mOrigin.x()) * mScale));
	out.setY(mBorderT + ((((float)pt.y) - mOrigin.y()) * mScale));
}

