#include "build.h"

#include "math/poly/edge.h"
#include "math/poly/polyfilter.h"
#include "math/poly/polygonlist.h"

using namespace math;

/* BUILD
 ******************************************************************/
Build::Build()
	: mA(&mA_Stack)
	, mB(&mB_Stack)
{
}

void Build::on(EdgeList& a, EdgeList& b, PolygonList& r, PolyFilter& f)
{
	if (PRINT) {
		qDebug() << "Build::on()";
		qDebug() << "Edges A";
		a.print();
		qDebug() << "Edges B";
		b.print();
	}

	r.clear();
	while (nextPoly(mTmpR, a, b)) {
		if (mTmpR.getSize() > 2 && f.accept(mTmpR)) {
if (PRINT_OUTPUT) qDebug() << "Build output=" << mTmpR;
			r.pushBack(&mTmpR);
		}
	}
}

bool Build::nextPoly(Polygon& p, EdgeList& a, EdgeList& b)
{
	p.clear();
	// Find the next available starting point.  To be a starting
	// point, you need an edge that's not bad, with an a pt that's good.
	int					aIdx = -1;
	for (int k=0; k<a.mSize; k++) {
		const Edge&		e = a.mEdge[k];
		if (e.mAlignment >= Step::NEUTRAL && e.mA->mAlignment >= Step::GOOD) {
			aIdx = k;
			break;
		}
	}
	if (aIdx < 0) return false;
	// Generate a polygon starting on the a pt.  It can't be
	// any larger then the size of both lists, so use that as
	// an error condition.  Otherwise stop when we hit a cycle.
	int					errorSize = a.mSize + b.mSize + 1;
	Edge&				cycle = a.mEdge[aIdx];
	Edge*				e;
	const Step*			cycleStep = cycle.mA;
	start(aIdx, a, b);
	p.add(*(cycle.mA->mPt));
	cycle.mA->mAlignment = Step::EVIL;
	while ((e = nextEdge()) != NULL) {
		if (cycleStep == e->mB) return true;
		// Indicates we ran out of unused points, not an error condition
		if (e->mAlignment == Step::EVIL) return true;
		if (errorSize-- <= 0) {
			if (PRINT) qDebug() << "ERROR Build::nextPoly() hit cycle";
			return false;
		}

		// Mark this edge as used
		if (e->mAlignment >= Step::GOOD) e->mAlignment = Step::EVIL;
		p.add(*(e->mB->mPt));
		if (PRINT) qDebug() << "build pt " << *(e->mB->mPt);
	}
	if (PRINT) qDebug() << "ERROR Build::nextPoly() answered a null step";
	return false;
}

// idxA must be a valid index into a mStep.
void Build::start(const int idxA, EdgeList& a, EdgeList& b)
{
	if (PRINT) qDebug() << "START BUILD idx=" << idxA;
	mA->start(idxA, a);
	mB->start(0, b);
}

Edge* Build::nextEdge()
{
	Edge*				nxt = mA->cur();
	mA->step();
	// If the endpoint is an intersection, need to decide where to go next.
	// By default I should switch to b, but if that's bad, stick with a.
	if (nxt->mB->isIntersection()) {
		if (!(mB->seek(nxt->mB->mIntersection))) return NULL;
		if (mB->cur()->mAlignment > Step::EVIL) {
			switchSource();
			if (PRINT) qDebug() << "\t\tintersection, swap, new=" + mA->mIdx;
		}
	}
	return nxt;
}

void Build::switchSource()
{
	Source*		s = mA;		mA = mB;	mB = s;
}

/* BUILD::SOURCE
 ******************************************************************/
void Build::Source::start(const int idx, EdgeList& a)
{
	mList = &a;
	mIdx = idx;
}

Edge* Build::Source::cur()
{
	return &(mList->mEdge[mIdx]);
}

void Build::Source::step()
{
	mIdx += 1;
	if (mIdx >= mList->mSize) mIdx = 0;
}

bool Build::Source::seek(const Intersection* i)
{
	mIdx = mList->findIndexForA(i);
	if (mIdx < 0 || mIdx >= mList->mSize) {
		qDebug() << "ERROR Build::Source() intersection missing";
		return false;
	}
	return true;
}
