#include "steplist.h"

#include "math/poly/flatten.h"

using namespace math;

/* STEP-LIST
 ******************************************************************/
StepList::StepList()
	: mStep(NULL)
	, mSize(0)
{
	// This should prevent me from ever needing to grow.
	mStep = get(32);
}

void StepList::flatten(const Polygon& p, Flatten& flatten)
{
//qDebug() << "FLATTEN";
	const qint32			psize = p.getSize();
	mSize = 0;
	if (psize < 2) return;

	const Point3*			pts = p.getPoints();
	const Point3*			prv = &(pts[psize-1]);
	for (int k=0; k<psize; k++) {
		const Point3&		nxt = pts[k];
		flatten.addIntersections(*prv, nxt, *this);
		add(nxt, flatten.mEps);
		prv = &nxt;
	}
	if (mSize < 1) return;
	// If the last and first steps are the same, remove the
	// last, but maintain any intersection.
	if (sameAsLast(*(mStep[0].mPt), flatten.mEps)) {
		if (mStep[mSize-1].isIntersection()) mStep[0].mIntersection = mStep[mSize-1].mIntersection;
		mSize--;
	}
}

void StepList::add(const Point3& pt, const double eps)
{
	if ((mStep = get(mSize+1)) == NULL) throw std::bad_alloc();

	if (sameAsLast(pt, eps)) return;

	mStep[mSize].set(pt);
	mSize++;
}

void StepList::add(const Intersection* inter, const double eps)
{
	if (inter == NULL) throw std::bad_alloc();
	if ((mStep = get(mSize+1)) == NULL) throw std::bad_alloc();

	if (sameAsLast(inter->mPt, eps)) {
		mStep[mSize-1].set(*inter);
	} else {
		mStep[mSize].set(*inter);
		mSize++;
	}
}

int StepList::intersectionCount() const
{
	int			ans = 0;
	for (int k=0; k<mSize; k++) {
		if (mStep[k].mIntersection != NULL) ans++;
	}
	return ans;
}

bool StepList::sameAsLast(const Point3& pt, const double eps) const
{
	if (mSize <= 0) return false;
	return pt.sameAs2(*(mStep[mSize-1].mPt), eps);
}

int StepList::findIndexFor(const Intersection* inter) const
{
	if (inter == NULL) return -1;
	for (int k=0; k<mSize; k++) {
		const Step&		s = mStep[k];
		if (s.mIntersection == inter) return k;
	}
	return -1;
}

void StepList::print()
{
	for (int k=0; k<mSize; k++) {
		qDebug() << "\t" << k << (mStep[k]);
	}
}
