#include "point3.h"

#include <math.h>

#include "math/geo.h"

//#define PRINT_INTEGERS		(1)

/* POINT-3
 ******************************************************************/
Point3::Point3()
	: x(0)
	, y(0)
	, z(0)
{
}

Point3::Point3(const Point3& o)
	: x(o.x)
	, y(o.y)
	, z(o.z)
{
}

Point3::Point3(const double X, const double Y, const double Z)
	: x(X)
	, y(Y)
	, z(Z)
{
}

void Point3::set(const Point3& o)
{
	x = o.x;
	y = o.y;
	z = o.z;
}

void Point3::set(const double X, const double Y, const double Z)
{
	x = X;
	y = Y;
	z = Z;
}

void Point3::swap(Point3& o)
{
	if (this != &o) {
		const double		sx = x;		x = o.x;	o.x = sx;
		const double		sy = y;		y = o.y;	o.y = sy;
		const double		sz = z;		z = o.z;	o.z = sz;
	}
}

bool Point3::sameAs2(const Point3& pt, const double precision) const
{
	if (x == pt.x && y == pt.y) return true;
	return geo::abs(x-pt.x) < precision && geo::abs(y-pt.y) < precision;
}

bool Point3::sameAs3(const Point3& pt, const double precision) const
{
	if (x == pt.x && y == pt.y && z == pt.z) return true;
	return geo::abs(x-pt.x) < precision && geo::abs(y-pt.y) < precision && geo::abs(z-pt.z) < precision;
}

bool Point3::sameAs2(const Point3& pt) const
{
	return sameAs2(pt, geo::EPS);
}

bool Point3::sameAs3(const Point3& pt) const
{
	return sameAs3(pt, geo::EPS);
}

void Point3::normalize()
{
	const double		nrml = sqrt((x*x) + (y*y) + (z*z));
	if (nrml != 0) *this /= nrml;
}

double Point3::length() const
{
	return sqrt(x*x + y*y + z*z);
}

double Point3::dist2(const Point3& b) const
{
	return sqrt( ((x-b.x)*(x-b.x)) + ((y-b.y)*(y-b.y)) );
}

double Point3::dist3(const Point3& b) const
{
	return sqrt( ((x-b.x)*(x-b.x)) + ((y-b.y)*(y-b.y)) + ((z-b.z)*(z-b.z)) );
}

double Point3::dist3Squared(const Point3& b) const
{
	return ((x-b.x)*(x-b.x)) + ((y-b.y)*(y-b.y)) + ((z-b.z)*(z-b.z));
}

double Point3::dot2(const Point3& v) const
{
	return x * v.x + y * v.y;
}

double Point3::dot3(const Point3& v) const
{
	return x * v.x + y * v.y + z * v.z;
}

void Point3::asAdd(const Point3& a, const Point3& b)
{
	x = a.x+b.x;
	y = a.y+b.y;
	z = a.z+b.z;
}

void Point3::asSub(const Point3& a, const Point3& b)
{
	x = a.x-b.x;
	y = a.y-b.y;
	z = a.z-b.z;
}

void Point3::asCross3(const Point3& a, const Point3& b)
{
	x = a.y*b.z - b.y*a.z;
	y = a.z*b.x - b.z*a.x;
	z = a.x*b.y - b.x*a.y;
}

void Point3::asBlend(const Point3& a, const Point3& b, const double pos)
{
	x = (a.x*(1-pos)) + (b.x*pos);
	y = (a.y*(1-pos)) + (b.y*pos);
	z = (a.z*(1-pos)) + (b.z*pos);
}

void Point3::asMid(const Point3& a, const Point3& b)
{
	x = (a.x+b.x) / 2;
	y = (a.y+b.y) / 2;
	z = (a.z+b.z) / 2;
}

/* debug
 ******************************************************************/
QDebug& operator<<(QDebug& d, const Point3& pt)
{
	(d << "(").nospace();
#ifdef PRINT_INTEGERS
	(d << qint64(pt.x)).nospace();
#else
	(d << pt.x).nospace();
#endif
	(d << ", ").nospace();
#ifdef PRINT_INTEGERS
	(d << qint64(pt.y)).nospace();
#else
	(d << pt.y).nospace();
#endif
	(d << ", ").nospace();
#ifdef PRINT_INTEGERS
	(d << qint64(pt.z)).nospace();
#else
	(d << pt.z).nospace();
#endif
	d << ")";
	return d;
}
