#include "polygon.h"

#include <stdlib.h>
#include <string.h>

#include <QPolygonF>
#include <QRectF>
#include "point3.h"
#include "math/geo.h"

/* POLYGON
 ******************************************************************/
Polygon::Polygon()
	: mPt(0), mSize(0), mAlloc(0)
{
}

Polygon::~Polygon()
{
	realloc(mPt, 0);
}

qint32 Polygon::getSize() const
{
	return mSize;
}

Point3* Polygon::getPoints() const
{
	return mPt;
}

bool Polygon::isEmpty() const
{
	return mSize < 1;
}

void Polygon::clear()
{
	mSize = 0;
}

bool Polygon::set(const Polygon& o)
{
	if (!growTo(o.mSize)) return false;
	memcpy(mPt, o.mPt, sizeof(Point3)*o.mSize);
	mSize = o.mSize;
	return true;
}

bool Polygon::set(const QPolygonF& o)
{
	const int			size = o.size();
	if (!setSize(size)) return false;

	for (int k=0; k<size; k++) {
		const QPointF&	pt = o[k];
		mPt[k].set(pt.x(), pt.y(), 0);
	}
	return true;
}

bool Polygon::setSize(const qint32 newSize)
{
	if (!growTo(newSize)) return false;
	mSize = newSize;
	return true;
}

bool Polygon::add(const Point3& pt)
{
	return add(pt.x, pt.y, pt.z);
}

bool Polygon::add(const double x, const double y, const double z)
{
	if (mSize >= mAlloc && !growTo(mSize + 10)) {
		return false;
	}
	mPt[mSize++].set(x, y, z);
	return true;
}

// Remove all adjacent points within the given tolerance, i.e.
// remove what we have loosely identified as "duplicates."
void Polygon::compress(const double eps)
{
	for (int k=1; k<mSize; k++) {
		while (mSize > k && mPt[k-1].sameAs2(mPt[k], eps)) {
			for (int j=k+1; j<mSize; j++) {
				mPt[j-1].set(mPt[j]);
			}
			mSize--;
		}
	}
	// Check first - last
	if (mSize > 2 && mPt[0].sameAs2(mPt[mSize-1], eps)) mSize--;
}

bool Polygon::contains(const Point3& pt) const
{
	return geo::polygonContains(mPt, mSize, pt.x, pt.y);
}

bool Polygon::isClockWise() const
{
	return isClockWise(geo::EPS);
}

// Original function by Paul Bourke
// http://debian.fmi.uni-sofia.bg/~sergei/cgsr/docs/clockwise.htm
bool Polygon::isClockWise(const double precision) const
{
	if (mSize < 2) return false;
	const bool			isClosed = mPt[0].sameAs2(mPt[mSize-1], precision);
	const qint32		size = mSize + (isClosed ? -1 : 0);
	if (size < 3) return false;

	int					count = 0;
	for (qint32 i=0;i<size;i++) {
		const qint32	j = (i + 1) % size;
		const qint32	k = (i + 2) % size;
		float			z  = (mPt[j].x - mPt[i].x) * (mPt[k].y - mPt[j].y);
		z -= (mPt[j].y - mPt[i].y) * (mPt[k].x - mPt[j].x);
		if (z < 0) count--;
		else if (z > 0) count++;
	}
	if (count > 0) return false;
	else if (count < 0) return true;
	else return true;		// technically undetermined
}

void Polygon::getFrame(math::Rect& f) const
{
	if (mSize < 1) {
		f.set(0, 0, -1, -1);
		return;
	}
	double				l = mPt[0].x, t = mPt[0].y;
	double				r = l, b = t;
	for (int k=1; k<mSize; k++) {
		const Point3&	pt = mPt[k];
		if (pt.x < l) l = pt.x;
		if (pt.y < t) t = pt.y;
		if (pt.x > r) r = pt.x;
		if (pt.y > b) b = pt.y;
	}
	f.set(l, t, r, b);
}

void Polygon::getFrame(QRectF& f) const
{
	math::Rect			r;
	getFrame(r);
	f.setCoords((float)r.mL, (float)r.mT, (float)r.mR, (float)r.mB);
}

static const int		SIDE = 4;
static const int		LOW_X[SIDE] = { 0, 3, 4, 7 };
static const int		HIGH_X[SIDE] = { 1, 2, 5, 6 };
static const int		LOW_Y[SIDE] = { 0, 1, 4, 5 };
static const int		HIGH_Y[SIDE] = { 2, 3, 6, 7 };
// NOTE:  Must always keep these sides assigned this way,
// some clients rely on it.
static const int		LOW_Z[SIDE] = { 0, 1, 2, 3 };
static const int		HIGH_Z[SIDE] = { 4, 5, 6, 7 };

bool Polygon::getFrame(Polygon& f) const
{
	if (mSize < 1) return false;
	if (!f.setSize(SIDE+SIDE)) return false;
	Point3*			fpts = f.mPt;
	for (int k=0; k<f.mSize; k++) f.mPt[k].set(mPt[0]);
	for (int k=1; k<mSize; k++) {
		const Point3&	pt = mPt[k];
		for (int j=0; j<SIDE; j++) {
			if (pt.x < fpts[LOW_X[j]].x) fpts[LOW_X[j]].x = pt.x;
			if (pt.x > fpts[HIGH_X[j]].x) fpts[HIGH_X[j]].x = pt.x;
			if (pt.y < fpts[LOW_Y[j]].y) fpts[LOW_Y[j]].y = pt.y;
			if (pt.y > fpts[HIGH_Y[j]].y) fpts[HIGH_Y[j]].y = pt.y;
			if (pt.z < fpts[LOW_Z[j]].z) fpts[LOW_Z[j]].z = pt.z;
			if (pt.z > fpts[HIGH_Z[j]].z) fpts[HIGH_Z[j]].z = pt.z;
		}
	}
	return true;
}

void Polygon::getBarycenter(Point3& bary) const
{
	geo::getBarycenter(mPt, mSize, bary);
}

bool Polygon::growTo(const qint32 newAlloc)
{
	if (newAlloc <= mAlloc) return true;

	Point3*				newPt = (Point3*)realloc(mPt, newAlloc*sizeof(Point3));
	if (!newPt) return false;

	mPt = newPt;
	mAlloc = newAlloc;
	return true;
}

QDebug& operator<<(QDebug& d, const Polygon& p)
{
	const Point3*		pts = p.getPoints();
	const int			size = p.getSize();
	(d << "Polygon size=" << size).nospace();
	for (int k=0; k<size; k++) {
		(d << "\n\t").nospace();
		(d << pts[k]).nospace();
	}
	return d;
}
