#ifndef LINEARCONGRUENCE_H
#define LINEARCONGRUENCE_H

#include <Qt>

/* LINEAR-CONGRUENCE
 * Generate random numbers using linear congruence:
 *     R1 = ((A*R0) + C) mod M
 * This class is intended to be light and fast.  It's not as pleasing as
 * things like pink noise, but it's similar and has better performance.
 * Reference:  http://mathworld.wolfram.com/LinearCongruenceMethod.html
 ******************************************************************/
class LinearCongruence
{
public:					// static

	// A lot of work has gone into determining ideal default values,
	// these are good choices but you can override in the constructor.
	static const qint32		DEFAULT_A = 1366;
	static const qint32		DEFAULT_C = 150889;
	static const qint32		DEFAULT_M = 714025;

public:
	LinearCongruence(	const qint32 seed = 0, const qint32 a = DEFAULT_A,
						const qint32 c = DEFAULT_C, const qint32 m = DEFAULT_M);
	LinearCongruence(const LinearCongruence&);

	LinearCongruence& operator=(const LinearCongruence&);

	// Seeding with < zero is just a random number
	void				Seed(const qint32 seed = 0);

	// Absolute answers 0 to 1, relative answers -1 to 1.
	float				NextAbs() const;
	float				NextRel() const;

private:
	mutable qint32		mSeed;
	const qint32		mA, mC, mM;
};


/*----------------------------------------------------------------*/
/*----- inline definitions ---------------------------------------*/
float inline LinearCongruence::NextAbs() const
{
	mSeed = ((mA * mSeed) + mC) % mM;
	return float(mSeed) / float(mM);
}

float inline LinearCongruence::NextRel() const
{
	mSeed = ((mA * mSeed) + mC) % mM;
	return (float(mSeed) / float(mM)) * 2 - 1;
}

#endif // LINEARCONGRUENCE_H
