#ifndef POLYISEXTERNAL_H
#define POLYISEXTERNAL_H

#include "collections/recyclearray.h"
#include "math/primitives/polygon.h"
#include "math/tests/lineiscolinear.h"
#include "math/tests/ptinpoly.h"
#include "math/tests/pttopoly.h"
#include "math/tests/segseg.h"
#include "math/tests/segpoly.h"

namespace math {

/* MATH::POLY-IS-EXTERNAL
 * Given polygons a and b, answer true if a is external
 * to b.  This is essentially an adjancy test, used to
 * find two polygons that may share an edge, but don't
 * intrude on each other.
 *
 * Algorithm works by trying to identify conditions where b is
 * internal to a, considered a failure.  Do this by finding all
 * intersections of b against a, then applying a couple rules:
 * If intersecting segment from b has an endpoint inside a, then
 * fail.  If a single segment from b intersects across more than
 * one non-colinear segment from a, and the intersection points
 * don't lie on the colinear (if any), then fail.
 ******************************************************************/
class PolyIsExternal
{
public:
	double				mEps;
	bool				mPrint;

public:
    PolyIsExternal();

	bool				on(const Polygon& a, const Polygon& b);

private:
	/* COLINEAR
	 * A colinear segment
	 **************************************************************/
	struct Colinear {
		const Point3	*mA0, *mA1;

		bool			sameAs(const Point3&, const double eps) const;
	};
	typedef RecycleArray<Colinear>	ColinearBlock;

private:
	LineIsColinear		mLC;
	PtInPoly			mPiP;
	PtToPoly			mPtP;
	SegSeg				mSS;
	SegPoly				mSP;

	Point3				mPt;
	Point3				mHit[2];

	ColinearBlock		mColinear;

	bool				isExternalTo(const Polygon& a, const Polygon& b);

	// Determine if pt is on the interior (not outside, not edge) of b.
	bool				isInternal(const Point3& pt, const Polygon& b);
	// Same as isInternal() but fail automatically if the pt is on
	// the supplied endpoints (which are known to be on edges of the poly).
	bool				isInternal(	const Point3& pt, const Polygon& b,
									const Point3& edgeA, const Point3& edgeB, const Point3& edgeC);

	// Identify whether any of the mHit points are the same as any
	// endpoint in the colinear list.
	bool				looseHit(const Colinear*, const int size) const;
};

} // namespace math

#endif // POLYISEXTERNAL_H
