#include "ptinpoly.h"

#include "math/geo.h"

using namespace math;

static const Point3& furthest(const Point3& pt, const Polygon& poly);

/* MATH::PT-IN-POLY
 ******************************************************************/
PtInPoly::PtInPoly()
	: mEps(0.001)
	, mPrint(false)
{
	mSS.mEps = 0.001;
}

bool PtInPoly::on(const Point3& pt, const Polygon& poly)
{
	const qint32			psize = poly.getSize();
	if (!start(psize + 2)) return false;
	constructSegment(pt, poly);
if (mPrint) {
	qDebug() << "PtInPoly on " << pt;
	qDebug() << "seg " << mA0 << ", " << mA1;
}
	mSortOrigin.set(mA0);
	pushSrc(pt);
	// Add all the intersections
	const Point3*			pts = poly.getPoints();
	const Point3*			prv = &(pts[psize-1]);
	// Take the largest distance, so we have the most "room"
	// and are less influenced by
	for (int k = 0; k<psize; k++) {
		const Point3&		nxt = pts[k];
if (mPrint) qDebug() << "\tcmp " << *prv << ", " << nxt;
		if (mSS.on2(mA0, mA1, *prv, nxt)) {
if (mPrint) qDebug() << "\tADD A PT";
			// If the source pt is within tolerance of a hit point on
			// the poly, then I am automatically true.
			if (pt.sameAs3(mSS.mHit, mEps)) return true;
			pushIntersection(mSS.mHit);
		} else {
if (mPrint) qDebug() << "\tdist=" << mSS.mDistanceToHit << " vs eps=" << mSS.mEps;
		}
		prv = &nxt;
	}
if (mPrint) {
//	qDebug() << "SORT size=" << mSort.size();
	qDebug() << "SORT size=" << mList.size();
}

	// ERROR CONDITION, actually.
	if (mList.isEmpty()) return false;
if (mPrint) {
	ManagedListIterator<Hit*>		it(mList);
	while (it.hasValue()) {
		const Hit&					hit(**it);
		if (hit.mSrc != NULL) qDebug() << "\tSRC=" << hit.mHit << " d=" << hit.mD;
		else qDebug() << "\thit=" << hit.mHit << " d=" << hit.mD;
		++it;
	}
}
	// Count the number of hits surrounding the source, odd
	// for interior, even for exterior
	// NOTE:  Skip any duplicate points.
	prv = NULL;
	ManagedListIterator<Hit*>		i(mList);
	bool							exterior = true;
	while (i.hasValue()) {
		const Hit&					hit(**i);
//		const Point3				&nxt = (*i)->mHit;
if (mPrint) qDebug() << "nxt=" << hit.mHit;
		if (prv == NULL || hit.mSrc != NULL || !prv->sameAs3(hit.mHit, mEps)) {
if (mPrint) qDebug() << "switch exterior";
			exterior = !exterior;
		} else if (mPrint) qDebug() << "skip dup";
		if (hit.mSrc != NULL) return exterior;
		prv = &(hit.mHit);
		++i;
	}
	if (mPrint) qDebug() << "NO SRC";

	// ERROR CONDITION, actually.
	return false;
}

bool PtInPoly::start(const int size)
{
	mList.clear();
	if (size < 3) return false;
	return true;
}

void PtInPoly::constructSegment(const Point3& pt, const Polygon& poly)
{
	mA0.set(pt);
	mA1.set(furthest(pt, poly));
	mDelta0.asSub(mA0, mA1);
	mDelta1.asSub(mA1, mA0);
	// Expand the deltas so we're guaranteed to encompass the poly.
	mDelta0 *= 4;
	mDelta1 *= 4;
	// Extend the line according to the delta
	mA0.asAdd(mA0, mDelta0);
	mA1.asAdd(mA1, mDelta1);
}

void PtInPoly::pushSrc(const Point3& pt)
{
	mHit.mSrc = &pt;
	mHit.mHit.set(pt);
//	mHit.mD = mSortOrigin.dist3Squared(pt);
	mHit.mD = mSortOrigin.dist3(pt);
	mList.pushSorted(&mHit);
}

void PtInPoly::pushIntersection(const Point3& pt)
{
	mHit.mSrc = NULL;
	mHit.mHit.set(pt);
//	mHit.mD = mSortOrigin.dist3Squared(pt);
	mHit.mD = mSortOrigin.dist3(pt);
	mList.pushSorted(&mHit);
}

/* PT-IN-POLY::HIT
 ******************************************************************/
void PtInPoly::Hit::clear()
{
	mSrc = NULL;
}

void PtInPoly::Hit::set(const Hit& o)
{
	if (this != &o) {
		mSrc = o.mSrc;
		mHit.set(o.mHit);
		mD = o.mD;
	}
}

/* PT-IN-POLY::HIT-LIST
 ******************************************************************/
class PtInPoly::Factory : public ManagedFactory<Hit*>
{
public:
	Factory() : ManagedFactory<Hit*>(true) { }

	virtual Hit*		newObject() const {
		Hit*			ans = new Hit();
		if (ans == NULL) throw std::exception();
		return ans;
	}

	virtual void		deleteObject(Hit* o) const { delete o; }
	virtual void		initializeObject(Hit* o) const { o->clear(); }
	virtual void		setObject(Hit* dst, Hit* src) const { if (src != NULL) dst->set(*src); }
	virtual int			sortObject(Hit* o1, Hit* o2) const { return geo::sort_doubles(o1->mD, o2->mD); }
};
static PtInPoly::Factory	FACTORY;

PtInPoly::HitList::HitList() : inherited(FACTORY) { }

/* misc
 ******************************************************************/
static const Point3& furthest(const Point3& pt, const Polygon& poly)
{
	const qint32		size = poly.getSize();
	const Point3*		pts = poly.getPoints();
	const Point3*		ans = &(pts[0]);
	double				d = pt.dist3Squared(*ans);
	for (int k=1; k<size; k++) {
		const double	nextD = pt.dist3Squared(pts[k]);
		if (nextD > d) {
			ans = &(pts[k]);
			d = nextD;
		}
	}
	return *ans;
}
