#ifndef PTINPOLY_H
#define PTINPOLY_H

#include "collections/managedlist.h"
#include "math/primitives/polygon.h"
#include "math/tests/segseg.h"

namespace math {

/* MATH::PT-IN-POLY
 * Determine if the point lies in the polygon, in 3D.  It's
 * assumed that the point lies on the same plane as the poly.
 *
 * Do this by finding a line from the pt to the poly, then
 * extending that line beyond the bounds, finding all
 * intersections, and determining what lines enclose the point.
 * NOTE:  Pts on an edge are considered inside.
 * NOTE:  If the polygon is a point or line, this test will
 * always fail, even though the nice thing to do would be to
 * check if it falls on the point or line.
 ******************************************************************/
class PtInPoly
{
public:
	double				mEps;
	bool				mPrint;

public:
    PtInPoly();

	bool				on(const Point3&, const Polygon&);

private:
	Point3				mA0, mA1, mDelta0, mDelta1, mSsHit;
	SegSeg				mSS;

	// Guarantee that there are enough hits.
	bool				start(const int size);
	// Construct a line that will intersect the poly somewhere
	// and extend through it fully.
	void				constructSegment(const Point3&, const Polygon&);

private:
	/* HIT
	 * An intersection (or the original, test point)
	 **************************************************************/
	struct Hit {
		// Only the original source pt will have this set
		const Point3*	mSrc;
		Point3			mHit;
		double			mD;

		void			clear();
		void			set(const Hit&);
	};

private:
	/* HIT-LIST
	 * Memory manage the hit points.
	 **************************************************************/
	class HitList : public ManagedList<Hit*>
	{
	public:
		HitList();

	private:
		typedef ManagedList<Hit*>	inherited;
	};
	HitList				mList;

private:
	Hit					mHit;
	Point3				mSortOrigin;

	void				pushSrc(const Point3&);
	void				pushIntersection(const Point3&);

// internal
public:
	class Factory;
};

} // namespace math

#endif // PTINPOLY_H
