#include "activeproject.h"

#include <QtDebug>
#include <QMutexLocker>
#include "app/environment.h"
#include "io/json_parser.h"
#include "io/json_copier.h"
#include "fx/chain.h"

// There are no layers right now, so skip to the root.
static const int	INIT_DEPTH = fx::VisitParams::ROOT_DEPTH-1;

/* ACTIVE-PROJECT
 ******************************************************************/
ActiveProject::ActiveProject(	const ArcList& arcs, const ColorSrcList& cs,
								const Environment& env, QObject *parent)
	: QObject(parent)
	, mFxUniqueId(0)
	, mArcs(arcs)
	, mClrSrc(cs)
	, mEnv(env)
	, mTimeStamp(0)
	, mLock(QMutex::Recursive)
{
	mPrj.mTimeStamp = mTimeStamp;
}

ActiveProject::~ActiveProject()
{
}

int ActiveProject::nextFxUniqueId()
{
	return mFxUniqueId++;
}

QString ActiveProject::getInstallPath(const QString& sub) const
{
	return mEnv.getInstallPath(sub);
}

bool ActiveProject::getArc(ArcHead& head, const int idx) const
{
	return mArcs.getArc(idx, head);
}

bool ActiveProject::getColorSrc(ColorSrcHead& h, const ColorSrcMatch& m) const
{
	return mClrSrc.find(h, m);
}

void ActiveProject::clear()
{
	mFxUniqueId = 0;
	mTimeStamp = 0;

	Project			p;
	p.swap(mPrj);
	mPrj.mTimeStamp = mTimeStamp;
}

QSize ActiveProject::getCanvasSize() const
{
	QMutexLocker		lock(&mLock);
	return mPrj.mCanvasSize;
}

void ActiveProject::setCanvasSize(const QSize& s)
{
	{
		QMutexLocker		lock(&mLock);
		if (mPrj.mCanvasSize == s) return;
		if (s.width() < 1 || s.height() < 1) return;
		mPrj.mCanvasSize = s;
	}
	emit projectChanged(STRUCTURE_CHANGED);
}

int ActiveProject::getFxId(const int layer) const
{
	int					ans = -1;
	{
		QMutexLocker	lock(&mLock);
		const Fx*		fx = mPrj.mHead.fx();
		// XXX No layers right now, as soon as there are, figure out how that changes things
		if (fx != NULL) ans = fx->getId();
	}
	return ans;
}

bool ActiveProject::getFxUrl(const int fxId, param::Url& url)
{
	QMutexLocker		lock(&mLock);
	fx::FindVisitor		find(fxId);
	fx::VisitParams		vp;
	mPrj.mHead.forEach(find, vp);
	if (find.mFx == NULL) return false;
	param::Parent::makeUrl(find.mFx, url);
	return true;
}

void ActiveProject::verify()
{
	QMutexLocker		lock(&mLock);
	if (mPrj.mHead.fx() == NULL) {
		fx::Chain*	chain = new fx::Chain();
		if (chain == NULL) return;
		chain->init(*this);
		mPrj.mHead.setFx(chain);
	}
}

void ActiveProject::setFx(Fx* fx)
{
	{
		QMutexLocker		lock(&mLock);
		mPrj.mHead.setFx(fx);
	}
	emit projectChanged(STRUCTURE_CHANGED);
}

void ActiveProject::edit(fx::EditOp& edit)
{
	{
		QMutexLocker		lock(&mLock);
		mPrj.mHead.edit(edit);
	}
	if (edit.success()) emit projectChanged(STRUCTURE_CHANGED);
}

bool ActiveProject::copyProject(Project& dst) const
{
	QMutexLocker		lock(&mLock);
	return dst.copyFrom(mPrj);
}

bool ActiveProject::getExports(ExportList& e) const
{
	QMutexLocker		lock(&mLock);
	json::Copier		c;
	if (!c.from(mPrj.mExports)) return false;
	e.clear();
	return c.to(e);
}

bool ActiveProject::setExports(const ExportList& e)
{
	QMutexLocker		lock(&mLock);
	json::Copier		c;
	if (!c.from(e)) return false;
	mPrj.mExports.clear();
	return c.to(mPrj.mExports);
}

bool ActiveProject::contains(const param::Url& url) const
{
	QMutexLocker			lock(&mLock);
	return mPrj.mHead.findParent(url) != NULL;
}

void ActiveProject::forEach(fx::Visitor& v)
{
	fx::VisitParams			vp;
	vp.mDepth = INIT_DEPTH;

	QMutexLocker			lock(&mLock);
	mPrj.mHead.forEach(v, vp);
}

void ActiveProject::getParams(param::List& list, param::GetArgs& get) const
{
	get.mError = 0;
	QMutexLocker		lock(&mLock);
	if (get.mUrl.size() < 1) {
		mPrj.mHead.getParams(list, get);
	} else {
		// If users supply a URL, then track down the parent and limit
		// the params to that.  If no parent, then do nothing (as opposed
		// to, say, defaulting to getting every param in the project,
		// which would be odd).
		const param::Parent*	p = mPrj.mHead.findParent(get.mUrl);
		if (p != NULL) {
			p->getParams(list, get);
			list.prepend(get.mUrl);
		}
	}
}

void ActiveProject::setParamF(const param::Url& url, const float v)
{
	SetParam		sp(*this);
	sp.setParamF(url, v);
}

void ActiveProject::setParamStr(const param::Url& url, const QString& v)
{
	SetParam		sp(*this);
	sp.setParamStr(url, v);
}

void ActiveProject::setParamArc(const param::Url& url, const ArcHead& h)
{
	SetParam		sp(*this);
	sp.setParamArc(url, h);
}

void ActiveProject::setParamClr(const param::Url& url, const ColorSrcHead& h)
{
	SetParam		sp(*this);
	sp.setParamClr(url, h);
}

void ActiveProject::setFlag(const param::Url& url, const int bit, const bool on)
{
	bool					changed = false;
	param::UrlIterator		it(url);
	if (!it.hasValue()) return;
	{
		QMutexLocker		lock(&mLock);
		Fx*					fx = mPrj.mHead.fx();
		// Locate the param.  Need to match at the top level, or else DEATH.
		if (fx == NULL || !it.matches(fx->getId())) return;
		++it;
		param::Parent*		p = fx->findParent(it);
		if (p == NULL) return;
		changed = p->setFlag(bit, on);
	}
	if (changed) {
		emit projectChanged(STRUCTURE_CHANGED);
	}
}

/* ACTIVE-PROJECT::SET-PARAM
 ******************************************************************/
ActiveProject::SetParam::SetParam(ActiveProject& prj)
	: mPrj(prj)
	, mChanged(false)
{
}

ActiveProject::SetParam::~SetParam()
{
	if (mChanged) emit mPrj.paramBatchFinished();
}

void ActiveProject::SetParam::setParamF(const param::Url& url, const float v)
{
	mSetParam.clear();
	mSetParam.mV = v;
	setParam(url, mSetParam, false);
}

void ActiveProject::SetParam::setParamStr(const param::Url& url, const QString& v)
{
	mSetParam.clear();
	mSetParam.mStr = v;
	setParam(url, mSetParam, false);
}

void ActiveProject::SetParam::setParamArc(const param::Url& url, const ArcHead& h)
{
	mSetParam.clear();
	mSetParam.mArc.setTo(h);
	setParam(url, mSetParam, true);
}

void ActiveProject::SetParam::setParamClr(const param::Url& url, const ColorSrcHead& h)
{
	mSetParam.clear();
	mSetParam.mClr = h;
	setParam(url, mSetParam, true);
}

void ActiveProject::SetParam::setParam(const param::Url& url, param::SetArgs& sp, const bool structure)
{
	sp.mChanged = false;
	param::UrlIterator		it(url);
	if (!it.hasValue()) return;
	{
		// XXX Ideally, clients have a token that caches the
		// found parent, so I don't have to look for it every time.
		// Currently the expense is trivial, and for this app, that
		// will always be the case, but if the param framework is
		// ever used in a realtime environment, it would be nice.
		QMutexLocker		lock(&mPrj.mLock);
		Fx*					fx = mPrj.mPrj.mHead.fx();
		// Locate the param.  Need to match at the top level, or else DEATH.
		if (fx == NULL || !it.matches(fx->getId())) return;
		++it;
		param::Parent*		p = fx->findParent(it);
		if (p == NULL) return;
		p->setParam(it, sp);
	}
	if (sp.mChanged) {
		mChanged = true;
		if (structure) emit mPrj.projectChanged(STRUCTURE_CHANGED);
		else {
			Param			p;
			p.mUrl = url;
			sp.writeTo(p);
			emit mPrj.paramChanged(p);
		}
	}
}


/* ACTIVE-PROJECT::LAST-FX-ID
 * Find the highest FX ID in use.
 ******************************************************************/
class ActiveProject::LastFxId : public fx::Visitor
{
public:
	LastFxId() : mId(-1) { }
	int			mId;

	virtual int	visit(Fx& fx, fx::VisitParams&) {
		const int		id = fx.getId();
		if (id > mId) mId = id;
		return fx::CONTINUE;
	}
};

/* IO
 ******************************************************************/
void ActiveProject::writeJson(json::Writer& w) const
{
	QMutexLocker		lock(&mLock);
	mPrj.writeJson(w);
}

bool ActiveProject::readJson(QIODevice* d)
{
	if (d == NULL) return false;

	{
		QMutexLocker		lock(&mLock);
		clear();

		json::Parser		p;
		p.parse(d, &mPrj);
		// Make sure everyone has valid values.
		// This won't override anything read in.
		// (unless an FX has a bug)
		Fx*					fx = mPrj.mHead.fx();
		if (fx != NULL) fx->init(*this);
		// Set the new-FX-ID value
		LastFxId			last;
		fx::VisitParams		vp;
		mPrj.mHead.forEach(last, vp);
		mFxUniqueId = last.mId + 1;
	}
	emit projectChanged(STRUCTURE_CHANGED);
	return true;
}
