#include "paramurl.h"

#include <assert.h>
#include <string.h>

using namespace param;

static const int		BLOCK_SIZE = 8;

/* PARAM::URL
 ******************************************************************/
Url::Url()
	: mStack(NULL)
	, mAlloc(0)
	, mSize(0)
{
}

Url::Url(const Url& o)
	: mStack(NULL)
	, mAlloc(0)
	, mSize(0)
{
	*this = o;
}

Url::~Url()
{
	realloc(mStack, 0);
}

Url& Url::operator=(const Url& o)
{
	if (this != &o) {
		if (o.mSize <= 0) mSize = 0;
		else if (mAlloc >= o.mSize || growTo(o.mSize)) {
			for (int k=0; k<o.mSize; k++) mStack[k] = o.mStack[k];
			mSize = o.mSize;
		}
	}
	return *this;
}

bool Url::operator==(const Url& o) const
{
	if (this == &o) return true;
	if (mSize != o.mSize) return false;
	for (int k=0; k<mSize; k++) {
		if (mStack[k] != o.mStack[k]) return false;
	}
	return true;
}

bool Url::prepend(const qint32 id)
{
	assert(id >= 0);
	if (id < 0) return false;
	if (mSize >= mAlloc && !growTo(mSize + BLOCK_SIZE)) return false;

	// Because I currently store items in reverse, prepend
	// is actually a simple append.
	mStack[mSize] = id;
	mSize++;
	return true;
}

bool Url::prepend(const Url& o)
{
	for (int k=0; k<o.mSize; k++) prepend(o.mStack[k]);
	return true;
}

bool Url::append(const qint32 id)
{
	assert(id >= 0);
	if (mSize >= mAlloc && !growTo(mSize + BLOCK_SIZE)) return false;

	// I store items in reverse so prepend is efficient, since that
	// occurs frequently, but that means append is inefficient.  Of
	// course, I could move to a list model, so both are efficient.
	// Whatev!  It would take extreme conditions for it to matter.
	mSize++;
	for (int k=1; k<mSize; k++) mStack[k] = mStack[k-1];
	mStack[0] = id;
	return true;
}

bool Url::growTo(const int newAlloc)
{
	if (newAlloc <= mAlloc) return true;

	id_type*		newStack = (id_type*)realloc(mStack, newAlloc*sizeof(id_type));
	if (newStack == NULL) return false;

	mStack = newStack;
	mAlloc = newAlloc;
	return true;
}

void Url::popBack()
{
	// Remember, currently the back is the front.
	if (mSize <= 1) {
		mSize = 0;
	} else {
		for (int k=1; k<mSize; k++) mStack[k-1] = mStack[k];
		mSize--;
	}
}

void Url::print(QDebug& d) const
{
	// Remember, I store in reverse
	for (int k=mSize-1; k>=0; k--) {
		(d.nospace() << mStack[k]).nospace();
		if (k > 0) (d.nospace() << ":").nospace();
	}
}

QDebug& operator<<(QDebug& d, const param::Url& o)
{
	o.print(d);
	return d;
}

/* PARAM::URL-ITERATOR
 ******************************************************************/
UrlIterator::UrlIterator()
	: mUrl(NULL)
	, mIdx(-1)
	, mCurId(-1)
{
}

UrlIterator::UrlIterator(const Url& url)
	: mUrl(NULL)
	, mIdx(-1)
	, mCurId(-1)
{
	start(url);
}

void UrlIterator::start(const Url& url)
{
	mUrl = &url;
	mIdx = mUrl->mSize - 1;
	mCurId = (mIdx >= 0 ? mUrl->mStack[mIdx] : -1);
}

void UrlIterator::operator++()
{
	// Remember the stack is currently stored in reverse order.
	mIdx--;
	mCurId = (mIdx >= 0 ? mUrl->mStack[mIdx] : -1);
}

#ifdef QT_DEBUG
void UrlIterator::printRemainder() const
{
	qDebug() << "idx=" << mIdx;
	int			idx = mIdx;
	while (idx >= 0) {
		qDebug() << "\t" << mUrl->mStack[idx];
		idx--;
	}
}
#endif
