#ifndef PARAMURL_H
#define PARAMURL_H

#include <Qt>
#include <QDebug>

namespace param {

class UrlIterator;

/* PARAM::URL
 * A generic, segmented reference to a single param.  Currently
 * the segments can only be ints, but that might change to strings
 * at some point, so anyone hacking on this should be very careful
 * to avoid disturbing the API in such a way that makes that more
 * difficult.
 ******************************************************************/
class Url
{
public:
	Url();
	Url(const Url&);
	~Url();

	Url&			operator=(const Url&);
	bool			operator==(const Url&) const;

	int				size() const;

	// Add to the beginning of my path.  ID must always be >= 0,
	// failure to do so will make it unmatcheable in the iterator.
	bool			prepend(const qint32 id);
	bool			prepend(const Url&);
	// Add to the end, same rules.
	bool			append(const qint32 id);

	// Remove the last item.
	void			popBack();

	void			print(QDebug&) const;

private:
	friend class UrlIterator;
	typedef qint32	id_type;

	// Current implementation is to store the segments in
	// reverse order, so [0] is the param ID, and mSize-1 is
	// the top level of the nested object containing the param.
	// This implementation might change if we move to a standard
	// string model, so keep this encapsulated.
	id_type*		mStack;
	int				mAlloc, mSize;

	bool			growTo(const int newAlloc);
};

/* PARAM::URL-ITERATOR
 ******************************************************************/
class UrlIterator
{
public:
	UrlIterator();
	UrlIterator(const Url&);

	void			start(const Url&);
	void			operator++();

	bool			hasValue() const;
	// Answer true if we're on the last value
	bool			isLeaf() const;

	// Answer true if the current stage matches the requested
	// item.  Not that ints must always be >= 0, or else this
	// fails automatically.
	bool			matches(const int id) const;

#ifdef QT_DEBUG
	void			printRemainder() const;
#endif

private:
	const Url*		mUrl;
	int				mIdx;
	int				mCurId;
};

/* inline
 ******************************************************************/
inline int Url::size() const
{
	return mSize;
}

inline bool UrlIterator::hasValue() const
{
	return mIdx >= 0;
}

inline bool UrlIterator::isLeaf() const
{
	return mIdx == 0;
}

inline bool UrlIterator::matches(const int id) const
{
	if (id < 0) return false;
	return mCurId == id;
}

} // namespace param


QDebug& operator<<(QDebug&, const param::Url&);

#endif // PARAMURL_H
