#include "chainplug.h"

#include <assert.h>

#include <QDebug>
#include <QPainter>

#include "model/activeproject.h"
#include "ui/look.h"
#include "ui/navigation.h"

static const QString		ADD_SZ("+");
static const QString		REMOVE_SZ("-");

static const int			COMMAND_ID = -191;

/* CHAIN-PLUG::FX-CACHE-OP
 ******************************************************************/
class ChainPlug::FxCacheOp : public fx::Visitor {
public:
	ChainPlug&		mPlug;

	FxCacheOp(ChainPlug& p) : mPlug(p) { }

	virtual int		visit(Fx& fx, fx::VisitParams& vp)
	{
		if (vp.mDepth == fx::VisitParams::ROOT_DEPTH) {
			FxE*	e = new FxE(mPlug.mNav, fx);
			if (e != NULL) mPlug.mE.append(e);
		}
		return fx::CONTINUE;
	}
};

/* CHAIN-PLUG
 ******************************************************************/
ChainPlug::ChainPlug(ActiveProject& prj, Navigation& nav)
	: mPrj(prj)
	, mNav(nav)
	, mNeedsCache(true)
{
}

ChainPlug::~ChainPlug()
{
	uncache();
}

void ChainPlug::setParent(FxUi::Parent* p, const Look& look)
{
	assert(p != NULL);
	inherited::setParent(p, look);

	if (mNeedsCache) {
		cache(p->entryH(), look);
		mNeedsCache = false;
	}

	for (int i=0; i<mE.size(); i++) {
		if (mE[i] == NULL) continue;
		p->add(mE[i]);
	}
}

int ChainPlug::onProjectChanged()
{
	uncache();
	return FxUi::UI_CHANGED;
}

void ChainPlug::cache(const int entryH, const Look& look)
{
	// Cache all the top-level FX
	FxCacheOp			op(*this);
	mPrj.forEach(op);

	FxE*	e;
	if ((e = new FxE(mNav, ADD_SZ)) != NULL) mE.append(e);
	if ((e = new FxE(mNav, REMOVE_SZ)) != NULL) mE.append(e);

	// Create images
	for (int i=0; i<mE.size(); i++) {
		if (mE[i] != NULL) mE[i]->cache(entryH, look);
	}
}

void ChainPlug::uncache()
{
	for (int i=0; i<mE.size(); i++) delete mE[i];
	mE.clear();
	mNeedsCache = true;
}

/* CHAIN-PLUG::FX-E
 *******************************/
ChainPlug::FxE::FxE(Navigation& nav, const Fx& fx)
	: mNav(nav)
	, mFxId(fx.getId())
	, mLabel(fx.getLabel())
{
}

ChainPlug::FxE::FxE(Navigation& nav, const QString& label)
	: mNav(nav)
	, mFxId(COMMAND_ID)
	, mLabel(label)
{
}

QSize ChainPlug::FxE::size() const
{
	return mImg.size();
}

void ChainPlug::FxE::onPaint(paint_args& pa)
{
	pa.mPainter->drawPixmap(	pa.mFrame.left(), pa.mFrame.top(),
								(mNav.mSelectedFx == mFxId ? mImgSelected : mImg) );
}

void ChainPlug::FxE::mousePressEvent(const mouse_args&)
{
}

void ChainPlug::FxE::mouseMoveEvent(const mouse_args&)
{
}

void ChainPlug::FxE::mouseReleaseEvent(const mouse_args& down)
{
	if (down.mEntry == NULL) return;

	if (mFxId == COMMAND_ID) {
		if (mLabel == ADD_SZ && down.mEvent != NULL) mNav.addFx(down.mEvent->globalPos());
		else if (mLabel == REMOVE_SZ) mNav.removeFx(mNav.mSelectedFx);
	} else {
		mNav.selectFx(mFxId);
	}
}

static void draw(	QPainter& painter, const int w, const int h,
					const Look& look, const QString& label, const bool bold)
{
	// Clear the background
	painter.setPen(Qt::NoPen);
	painter.setBrush(look.mDashBgB); //.mAlphaB);
	painter.drawRect(0, 0, w, h);
	// Draw a param panel
	painter.setBrush(look.mParamBgB);
	painter.setRenderHint(QPainter::Antialiasing);
	painter.drawRoundedRect(0, 0, w, h, look.mRound, look.mRound);
	// Draw the label
	painter.setPen(look.mParamFgC);
	if (bold) {
		QFont		font(painter.font());
		font.setWeight(QFont::Bold);
		painter.setFont(font);
	}
	int				x = 4, y = 10;
	painter.drawText(x, y, label);
}

void ChainPlug::FxE::cache(const int entryH, const Look& look)
{
	if (mLabel.size() < 1) return;
	int					w = entryH, h = entryH;
	if (w < 1 || h < 1) return;
	QImage*				img = new QImage(w, h, QImage::Format_RGB888);
	if (img == NULL) return;

	// CREATE NORMAL IMG
	{
		QPainter		painter(img);
		draw(painter, w, h, look, mLabel, false);
	}
	mImg = QPixmap::fromImage(*img);

	// CREATE SELECTED IMG
	{
		QPainter		painter(img);
		draw(painter, w, h, look, mLabel, true);
	}
	mImgSelected = QPixmap::fromImage(*img);

	delete img;
}
