#include "framecontrol.h"

#include <math.h>
#include <QMouseEvent>
#include "engine/engineparam.h"

static const int		L_EDGE = (1<<0);
static const int		T_EDGE = (1<<1);
static const int		R_EDGE = (1<<2);
static const int		B_EDGE = (1<<3);

static const int		LT_EDGE = (L_EDGE | T_EDGE);
static const int		RT_EDGE = (R_EDGE | T_EDGE);
static const int		RB_EDGE = (R_EDGE | B_EDGE);
static const int		LB_EDGE = (L_EDGE | B_EDGE);

/* FRAME-CONTROL
 ******************************************************************/
FrameControl::FrameControl(QWidget& parent, ActiveProject& prj,
							const Look& look)
	: mParent(parent)
	, mPrj(prj)
	, mLook(look)
	, mDownMask(0)
{
}

void FrameControl::clear()
{
	RawControl::clear();
	mFrame.setRect(0, 0, -1, -1);
	mDownMask = 0;
}

void FrameControl::setTo(const param::List& list)
{
	param::ListIterator		it(list);
	const Param*			p;
	while ((p = *it) != NULL) {
		if (p->mType == param::RAW_FLOAT_TYPE && p->mGroupName == param::FRAME_SZ) {
			mEnabled = true;
			if (p->mParamName == param::LEFT_SZ) {
				mFrame.setLeft(p->mFV);
				mUrlL = p->mUrl;
			} else if (p->mParamName == param::TOP_SZ) {
				mFrame.setTop(p->mFV);
				mUrlT = p->mUrl;
			} else if (p->mParamName == param::RIGHT_SZ) {
				mFrame.setRight(p->mFV);
				mUrlR = p->mUrl;
			} else if (p->mParamName == param::BOTTOM_SZ) {
				mFrame.setBottom(p->mFV);
				mUrlB = p->mUrl;
			}
		}
		++it;
	}
}

bool FrameControl::mousePressEvent(QMouseEvent* me)
{
	mDownMask = 0;

	const float			EXP = 4;
	const float			GRIP = 10;
	const QRectF		expF(mFrame.left()-EXP, mFrame.top()-EXP, mFrame.width()+(EXP*2), mFrame.height()+(EXP*2));
	const QPointF		pos(me->posF());
	if (!expF.contains(pos)) return false;

	if (pos.x() <= mFrame.left()+GRIP) mDownMask |= L_EDGE;
	if (pos.y() <= mFrame.top()+GRIP) mDownMask |= T_EDGE;
	if (pos.x() >= mFrame.right()-GRIP) mDownMask |= R_EDGE;
	if (pos.y() >= mFrame.bottom()-GRIP) mDownMask |= B_EDGE;
	if (mDownMask == 0) mDownMask = (LT_EDGE | RB_EDGE);
	mDownFrame = mFrame;
	mDownPt = pos;
	if (mFrame.width() > 0 && mFrame.height() > 0) mDownRatio = mFrame.width() / mFrame.height();
	else mDownRatio = 0;
	return true;
}

void FrameControl::mouseMoveEvent(QMouseEvent* me)
{
	if (mDownMask == 0) return;

	const QRectF		oldF(mFrame);
	const QPointF		pos(me->posF());
	const float			deltaX = pos.x()-mDownPt.x(),
						deltaY = pos.y()-mDownPt.y();
	// Corners always maintain the aspect ratio.  If you
	// want a weird shape, you can grab an edge
	if (mDownMask == RB_EDGE && mDownRatio > 0) {
		const float				fx = fabs(deltaX), fy = fabs(deltaY);
		const float				d = (fx >= fy ? deltaX : deltaY);
		if (mDownRatio >= 1) {
			mFrame.setRight(mDownFrame.right()+d);
			mFrame.setBottom(mFrame.width()*mDownRatio);
		} else {
			mFrame.setBottom(mDownFrame.bottom()+d);
			mFrame.setRight(mFrame.height()*mDownRatio);
		}
	} else {
		if ((mDownMask&L_EDGE) != 0) mFrame.setLeft(mDownFrame.left()+deltaX);
		if ((mDownMask&T_EDGE) != 0) mFrame.setTop(mDownFrame.top()+deltaY);
		if ((mDownMask&R_EDGE) != 0) mFrame.setRight(mDownFrame.right()+deltaX);
		if ((mDownMask&B_EDGE) != 0) mFrame.setBottom(mDownFrame.bottom()+deltaY);
	}
	if (oldF != mFrame) {
		{
			ActiveProject::SetParam					sp(mPrj);
			if (mFrame.left() != oldF.left())		sp.setParamF(mUrlL, mFrame.left());
			if (mFrame.top() != oldF.top())			sp.setParamF(mUrlT, mFrame.top());
			if (mFrame.right() != oldF.right())		sp.setParamF(mUrlR, mFrame.right());
			if (mFrame.bottom() != oldF.bottom())	sp.setParamF(mUrlB, mFrame.bottom());
		}
		mParent.update();
	}
}

void FrameControl::mouseReleaseEvent(QMouseEvent*)
{
	mDownMask = 0;
}

void FrameControl::onPaint(QPainter& p)
{
	p.fillRect(mFrame, mLook.mFrameFillC);
	p.setPen(mLook.mFrameStrokeC);
	p.setBrush(Qt::NoBrush);
	p.drawRect(mFrame);
}
