#include "renderthread.h"

#include <assert.h>

#include <QtDebug>
#include <QPixmap>

#include "engine/fx.h"

/* RENDER-THREAD
 ******************************************************************/
RenderThread::RenderThread(QObject *parent)
	: QThread(parent)
	, mAbort(false)
	, mRestart(false)
	, mOut(NULL)
{
}

RenderThread::~RenderThread()
{
	mMutex.lock();
	mAbort = true;
	mRestart = true;
	mWait.wakeOne();
	mMutex.unlock();

	wait();

	delete mOut;

	deleteAll(mIn);
	deleteAll(mFactoryIn);
	deleteAll(mUpdateIn);
}

bool RenderThread::addParam(const Param& p)
{
	RenderIn*			in = newIn();
	if (in == NULL) return false;
	QMutexLocker	locker(&mMutex);
	in->setParam(p);
	mIn.append(in);
	return true;
}

bool RenderThread::renderProject(Project& prj, const QSize& size)
{
	RenderIn*			in = newIn();
	if (in == NULL) return false;

	in->setProject(prj, size);
	render(in);
	return true;
}

void RenderThread::render()
{
	QMutexLocker	locker(&mMutex);
	onInputChanged();
}

void RenderThread::render(RenderIn* in)
{
	if (in == NULL) return;

	QMutexLocker	locker(&mMutex);
	mIn.append(in);
	onInputChanged();
}

bool RenderThread::getOutput(QPixmap& pix)
{
	QMutexLocker	locker(&mMutex);
	if (mOut == NULL) return false;
	// Oookaayy...  really don't see a way to reuse a pixmap's data.
	// As far as I know this allocates memory on every install.  Obviously
	// there's a better way to do this somehow.
	pix = QPixmap::fromImage(mOut->mOutput.mImg);
	mFactoryOut.append(mOut);
	mOut = NULL;
	return true;
}

void RenderThread::run()
{
	RenderOut*				out;
	forever {
		if (mAbort) return;

		// Update my state, run the renderer.
		begin();
		Fx*					fx = mPrj.mHead.fx();
		if (fx != NULL && (out = newOut(mSize)) != NULL && mWorkArgs.begin(mSize)) {
			// Prepare for rendering
			const int		wh = mSize.width() * mSize.height() * 4;
			memset(out->mOutput.mImg.bits(), 0, wh);
			// Render
			fx->work(mWorkArgs, out->mOutput);

mWorkArgs.finish(out->mOutput.mImg);
			finish(out);
		}

		mMutex.lock();
		if (!mRestart) mWait.wait(&mMutex);
		mRestart = false;
		mMutex.unlock();
	}
}

void RenderThread::onInputChanged()
{
	if (!isRunning()) {
		start(LowPriority);
	} else {
		mRestart = true;
		mWait.wakeOne();
	}
}

void RenderThread::begin()
{
	// Pop off all the incoming
	{
		QMutexLocker	locker(&mMutex);
		mUpdateIn = mIn;
		mIn.clear();
	}

	// Apply them
	for (int k=0; k<mUpdateIn.size(); k++) {
		RenderIn*		in = mUpdateIn[k];
		if (in == NULL) continue;
		if (in->mType == RenderIn::FX_TYPE) {
			mPrj.swap(in->mPrj);
			// If this is the same project, retain caches
			if (mPrj.mTimeStamp == in->mPrj.mTimeStamp) retainCache(in->mPrj);
			mSize = in->mSize;
		} else if (in->mType == RenderIn::PARAM_TYPE) {
			Fx*			fx = mPrj.mHead.fx();
			if (fx != NULL) {
				mSetParam.clear();
				mSetParam.readFrom(in->mParam);
				// Locate the param
				param::UrlIterator	it(in->mParam.mUrl);
				++it;
				param::Parent*		p = fx->findParent(it);
				if (p != NULL) p->setParam(it, mSetParam);
			}
		}
		// Clean up data.
		in->clear();
	}

	// Push the spent shells into the factory
	{
		QMutexLocker	locker(&mMutex);
		mFactoryIn << mUpdateIn;
	}
	mUpdateIn.clear();
}

void RenderThread::finish(RenderOut* out)
{
	assert(out != NULL);
	{
		QMutexLocker	locker(&mMutex);
		if (mOut != NULL) mFactoryOut.append(mOut);
		mOut = out;
	}
	emit renderFinished();
}

RenderOut* RenderThread::newOut(const QSize& size)
{
	{
		QMutexLocker	locker(&mMutex);
		while (!mFactoryOut.isEmpty()) {
			RenderOut*	out = mFactoryOut.takeFirst();
			if (out == NULL) continue;
			if (out->mOutput.mImg.size() == size) return out;
			delete out;
		}
	}
	return new RenderOut(size);
}

void RenderThread::pushOut(RenderOut* out)
{
	assert(out != NULL);
	QMutexLocker	locker(&mMutex);
	mFactoryOut.append(out);
}

RenderIn* RenderThread::newIn()
{
	RenderIn*			ans = NULL;
	{
		QMutexLocker	locker(&mMutex);
		if (!mFactoryIn.isEmpty()) ans = mFactoryIn.takeFirst();
	}
	if (ans == NULL) ans = new RenderIn();
	return ans;
}

void RenderThread::deleteAll(QList<RenderIn*>& list)
{
	while (!list.isEmpty()) {
		RenderIn*		in = list.takeFirst();
		delete in;
	}
}

void RenderThread::retainCache(Project&)
{
	// For now, data is cached in the frame.  Maybe that will always
	// be true, but if we ever have data cached in the FX, we can
	// preserve it here.
}
