#ifndef RENDERTHREAD_H
#define RENDERTHREAD_H

#include <QMutex>
#include <QThread>
#include <QList>
#include <QWaitCondition>

#include "renderio.h"

class Frame;

/* RENDER-THREAD
 * Handle rendering images.  Adapted from the Qt mandlebrot example:
 * http://doc.qt.nokia.com/4.6/threads-mandelbrot.html
 ******************************************************************/
class RenderThread : public QThread
{
    Q_OBJECT
public:
    explicit RenderThread(QObject *parent = 0);
	~RenderThread();

	// Add the param to be sync'd in the render project, but
	// DON'T cause a render.  (so params can be added in batches)
	bool					addParam(const Param&);

	// Either call causes a render to happen.  One exchanges
	// project settings with the incoming, the other sets a
	// param on the current project.
	bool					renderProject(Project&, const QSize&);
	// Cause a new render
	void					render();
	// Much less convenient.
	void					render(RenderIn*);

	// Pop the current output from the stack and set my pixmap
	// to the results.
	bool					getOutput(QPixmap&);

signals:
	void					renderFinished();

protected:
	void					run();

private:
	QMutex					mMutex;
	QWaitCondition			mWait;
	bool					mAbort;
	volatile bool			mRestart;
	// Only in the thread...
	fx::WorkArgs			mWorkArgs;
	Project					mPrj;
	QSize					mSize;
	param::SetArgs			mSetParam;
	// The current output
	RenderOut*				mOut;				// MUST BE LOCKED

	// The list of incoming render commands
	QList<RenderIn*>		mIn;				// MUST BE LOCKED
	// A place to cache currently out-of-service params
	QList<RenderIn*>		mFactoryIn;			// MUST BE LOCKED
	// A temporary list used during udpate.
	QList<RenderIn*>		mUpdateIn;

	// A place to cache currently out-of-service params
	QList<RenderOut*>		mFactoryOut;		// MUST BE LOCKED

	// Cause a new render -- client must have locked
	void					onInputChanged();

	// Update my data model with any incoming changes
	void					begin();
	// Finalize my output
	void					finish(RenderOut*);

	RenderIn*				newIn();
	RenderOut*				newOut(const QSize&);

	// Push the out onto the factory stack
	void					pushOut(RenderOut*);

	void					deleteAll(QList<RenderIn*>&);

	// New projects are installed even for just changing a single
	// FX.  If the new project is the same as the old, let FX retain
	// any caches they've generated.
	void					retainCache(Project& oldPrj);
};

#endif // RENDERTHREAD_H
