#include "renderview.h"

#include <QtDebug>
#include <QtGui>

#include "io/svgreader.h"
#include "model/activeproject.h"

/* RENDER-VIEW::PARAM-FILTER
 ******************************************************************/
class RenderView::ParamFilter : public param::Filter
{
public:
	virtual bool accept(const QString*, const QString*, const int type) const {
		return type == param::RAW_FLOAT_TYPE;
	}

	virtual bool supports(const int feature) const {
		return (feature&(param::ARC_FEATURE | param::COLOR_FEATURE)) != 0;
	}

};
static const RenderView::ParamFilter		gFilter;

/* RENDER-VIEW
 ******************************************************************/
RenderView::RenderView(ActiveProject& prj, const Look& look, QWidget *parent)
	: inherited(parent)
	, mPrj(prj)
	, mEmpty(true)
	, mFrameCtrl(*this, prj, look)
	, mMouseCtrl(NULL)
{
	mDataSize = mPrj.getCanvasSize();

	connect(&mThread, SIGNAL(renderFinished()),
				 this, SLOT(onRenderFinished()));

	// Kick off drawing on my current project.
	onProjectChanged(0);
}

QSize RenderView::minimumSizeHint() const
{
	return QSize(256, 256);
}

QSize RenderView::sizeHint() const
{
	return QSize(600, 400);
}

void RenderView::onSelectFx(const int id)
{
	const bool				anyEnabled = anyControlsEnabled();
	mFrameCtrl.clear();

	param::List				list;
	param::GetArgs			ga;
	ga.mFilter = &gFilter;
	mPrj.getFxUrl(id, ga.mUrl);
	if (ga.mUrl.size() < 1) return;
	mPrj.getParams(list, ga);

	mFrameCtrl.setTo(list);

	if (anyEnabled != anyControlsEnabled()) update();
}

void RenderView::writeExports(ActiveProject& prj)
{
	ExportList					exports;
	prj.getExports(exports);
	// OK, should be in a separate thread.  If we ever have any
	// official batch mechanism we'll pretty much have to at that point.
	const ExportList::Step*		e;
	for (int k=0; (e=exports.at(k)) != NULL; k++) {
		write(*e, mPixmap);
	}
}

void RenderView::onProjectChanged(const int)
{
	// Replace the entire FX structure
	Project					prj;
	if (!mPrj.copyProject(prj)) return;
	mDataSize = mPrj.getCanvasSize();
	if (mThread.renderProject(prj, mDataSize)) mEmpty = false;
}

void RenderView::onParamChanged(const Param& p)
{
	if (mEmpty) onProjectChanged(ActiveProject::STRUCTURE_CHANGED);
//	else mThread.renderParam(p);
	else mThread.addParam(p);
}

void RenderView::onParamBatchFinished()
{
	mThread.render();
}

void RenderView::mousePressEvent(QMouseEvent* me)
{
	mMouseCtrl = NULL;
	if (mFrameCtrl.mousePressEvent(me)) mMouseCtrl = &mFrameCtrl;
}

void RenderView::mouseMoveEvent(QMouseEvent* me)
{
	if (mMouseCtrl != NULL) mMouseCtrl->mouseMoveEvent(me);
}

void RenderView::mouseReleaseEvent(QMouseEvent* me)
{
	if (mMouseCtrl != NULL) mMouseCtrl->mouseReleaseEvent(me);
	mMouseCtrl = NULL;
}

void RenderView::paintEvent(QPaintEvent*)
{
	if (!mPixmap.isNull()) {
		QPainter painter(this);
		painter.drawPixmap(0, 0, mPixmap);
		if (mFrameCtrl.isEnabled()) mFrameCtrl.onPaint(painter);
	}
}

bool RenderView::anyControlsEnabled() const
{
	return mFrameCtrl.isEnabled();
}

void RenderView::write(const ExportList::Step& e, const QPixmap& pm) const
{
	if (e.mFilename.length() < 1) return;

	const QSize			prvSize = pm.size();
	QSize				nxtSize = prvSize;
	if (nxtSize == prvSize) pm.save(e.mFilename);
}

void RenderView::onRenderFinished()
{
	if (mThread.getOutput(mPixmap)) update();
}
